#include <stdio.h>
#include <stdlib.h>

#include <sys/time.h>
#include <unistd.h>

#include "novas.h"

#define REPTS 1

void timing(const char *what) {
  struct timeval t;

  if (REPTS < 2)
    return;

  gettimeofday(&t, /*timezone=>*/ NULL);
  printf("%s ---  %ld.%06ld\n", what, t.tv_sec, t.tv_usec);
}  /* timing */

int main (void)
{
  double deltat;
  double tjd, mobl, tobl, eq, dpsi, deps;
  double sunra, sundec, sundis;
  static body earthb = {
    0, /* type: 0==major planet */
    3, /* number: 3==Earth */
    "Earth"
  };
  static body sunb = {
    0, /* type: 0==major planet */
    10, /* number: 10==Sun */
    "Sun"
  };
#if 0
  static body moonb = {  /* works only with solsys2.c /JPL data??? */
    0, /* type: 0==major planet */
    11, /* number: 11==Moon */
    "Moon"
  };
#endif
  static site_info metsahovi = {
    ( 60.0 + (12.0 + 58.0/60.0)/60.0),  /* latitude 60d 12' 58" north */
    (  1.0 + (37.0 + 34.3/60.0)/60.0),  /* longitude 1h 37m 34.3s east */
    15.0, /* xxx:??? height (meters) */
    +2.0,  /* temperature, C */
    1011.0  /* atmospheric pressure in millibars */
  };
  double zd, az, rar, decr;
  int i;

  /* Fine-tune celestial pole offsets based on
     http://maia.usno.navy.mil IERS Bulletin A 'final.data' files. */
  cel_pole(-0.051881, -0.004329);  /* in arcseconds, IERS in milliarcsecs */
  deltat = 63.6642;  /* TT-UT1 in seconds, IERS */

  timing("--- start ---");
  for (i=0;i<REPTS;i++) {
  tjd = julian_date(1999, 11, 26, 9.0);
  }
  timing("julian_date()");

  for (i=0;i<REPTS;i++) {
  earthtilt(tjd,  /* TDB Julian date/time */
            &mobl, &tobl, &eq, &dpsi, &deps);
  }
  timing("earthtilt()  ");
  printf("TDB Julian date      %f\n", tjd);
  printf("Mean obliq of eclip  %f\n", mobl);
  printf("True obliq of eclip  %f\n", tobl);
  printf("Equation of equinox  %f\n", eq);
  printf("Nutation in long as  %f\n", dpsi);
  printf("Nutation in obliq as %f\n", deps);

  for (i=0;i<REPTS;i++) {
  app_planet(tjd, &sunb, &earthb, &sunra, &sundec, &sundis);
  }
  timing("app_planet() ");
  printf("Sun apparent RA        %.8f\n", sunra);
  printf("Sun apparent Dec       %.8f\n", sundec);
  printf("Sun apparent distance  %.8f\n", sundis);

  for (i=0;i<REPTS;i++) {
  topo_planet(tjd,  /* TT/TDT Julian date/time */
              &sunb, &earthb,
              deltat,  /* TT-UT1 in seconds, IERS */
              /*site=>*/ &metsahovi,
              &sunra, &sundec, &sundis);
  }
  timing("topo_planet()");
  printf("Sun topocentric RA        %.8f\n", sunra);
  printf("Sun topocentric Dec       %.8f\n", sundec);
  printf("Sun topocentric distance  %.8f\n", sundis);

  for (i=0;i<REPTS;i++) {
  equ2hor(tjd, deltat,  /* TT/TDT Julian date/time, TT-UT1 delta */
          /* celestial ephemeris pole x, y */ 0.0028436, 0.379468,
          /* site info */ &metsahovi,
          sunra, sundec,
          /* ref_option=> */ 2,  /* refr, use temp/press from site info */
          &zd, &az, &rar, &decr);
  }
  timing("equ2hor()    ");
  printf("Sun topocentric zenith distance degrees  %f\n", zd);
  printf("Sun           elevation degrees          %f\n", 90.0-zd);
  printf("Sun topocentric azimuth degrees          %f\n", az);
  printf("Sun topocentric RA,  true equator, equinox of date, /w refr  %f\n", rar);
  printf("Sun topocentric Dec, true equator, equinox of date, /w refr  %f\n", decr);

  for (i=0;i<24;i++) {
    double norefrzd;

    tjd = julian_date(1999, 11, 19, (double)i);
    topo_planet(tjd, &sunb, &earthb,
                deltat,  /* TT-UT1 in seconds, IERS */
                /*site=>*/ &metsahovi,
                &sunra, &sundec, &sundis);
    equ2hor(tjd, deltat,
            /* celestial ephemeris pole x, y */ 0.0028436, 0.379468,
            /* site info */ &metsahovi,
            sunra, sundec,
            /* ref_option=> */ 0,  /* no refraction */
            &norefrzd, &az, &rar, &decr);
    equ2hor(tjd, deltat,
            /* celestial ephemeris pole x, y */ 0.0028436, 0.379468,
            /* site info */ &metsahovi,
            sunra, sundec,
            /* ref_option=> */ 2,  /* refr, use temp/press from site info */
            &zd, &az, &rar, &decr);
    printf("Sun %02dh az=%11.6f el=%11.6f, norefr el=%11.6f  corr=%.6f\n",
           i, az, (90.0-zd), (90.0-norefrzd), norefrzd-zd);
  }  /* for each hour */

  return (EXIT_SUCCESS);
}
