#include "registration.h"
RCSID(antcon_cc, "$Id: antcon.cc,v 1.85 2005/10/05 12:35:07 lindi Exp $");

/* antcond.cc  --  Showing sensor readouts in "down" 'antcon' PC.  */

// $Log: antcon.cc,v $
// Revision 1.85  2005/10/05 12:35:07  lindi
// Use getopt to parse -U<user> and -P<pidfile>
//
// Revision 1.84  2002/11/03 09:59:12  amn
// Taking the new 'common/tstamp.c' into use.
//
// Revision 1.83  2002/04/05 15:02:30  amn
// clipLimitMax were inited to MIN values accidentally.
//
// Revision 1.82  2002/04/05 09:55:13  amn
// Added clip[min|max][az|el]limits.
//
// Revision 1.81  2001/10/08 06:53:23  amn
// SunAnt now "tracks" objects under Earth, i.e. negative El -> pos El.
//
// Revision 1.80  2001/04/12 10:06:48  amn
// REmoved 150GHz el limit (35/40deg) --> back to 100deg.
//
// Revision 1.79  2001/03/15 13:23:01  prisse
//
// Committing in new max elevation limit +35.0 -> +40.0.
// Modified Files:
//  antcon.cc
//
// Revision 1.78  2001/02/12 12:33:18  amn
// Changed max tracking elevation to +35deg.
//
// Revision 1.77  2001/02/12 11:56:57  amn
// Added waiting for antcon PP_ updates before antcli replying.
//
// Revision 1.76  2001/01/16 09:41:36  amn
// Added antcli --> antcon --> sh updates of af/met/cPointing.cc parameters.
//
// Revision 1.75  2000/12/30 11:18:33  amn
// Automatically recompile both regular and SunAnt, refraction fix.
//
// Revision 1.74  2000/11/03 07:39:43  amn
// Recompiled with 22GHz VLBI RX pointing; after much SUNANT development.
//
// Revision 1.73  2000/10/10 12:53:05  amn
// Made SunAnt elevation tracking limit higher, +1-->+2deg.
//
// Revision 1.72  2000/09/20 10:28:43  amn
// Added offset-free calculated azel to 'AZEL' and 'LOG=' reports.
//
// Revision 1.71  2000/09/20 10:15:05  amn
// Added min/ave/max azel error to LOG=.
//
// Revision 1.70  2000/08/15 06:47:02  amn
// Reduced both az and el tracking limits for SunAnt.
//
// Revision 1.69  2000/08/14 11:32:34  amn
// Added SunAnt "soft"/slow limits => can track Sun 24h.
// Az encoder native zero is at -179.15, so -163..+163 range is safe,
// no auto 0-->360 wraparound will occur.
//
// Revision 1.68  2000/08/10 09:42:49  amn
// First usable SunAnt version.
//
// Revision 1.67  2000/07/28 12:04:46  amn
// Clearing of end of screen lines after source name and log file name.
//
// Revision 1.66  2000/07/28 10:37:09  amn
// Wrong acceleration formula; removed centi- and millisecond time input.
//
// Revision 1.65  2000/07/28 09:52:10  amn
// Log file activation was forgotten; new screen layout in antcon.
//
// Revision 1.64  2000/07/28 09:28:18  amn
// Changed to using VC 'open'(1) and added logging 'log=' to antcli.
//
// Revision 1.63  2000/07/26 09:53:16  amn
// Ra-decl sweep status overwrote screen prompt.
//
// Revision 1.62  2000/07/26 09:48:18  amn
// Added 'radecsweep' plus its display on antcon screen.
//
// Revision 1.61  2000/07/25 12:27:30  amn
// azelsweep works about ok.
//
// Revision 1.60  2000/07/24 09:57:12  amn
// Added 'updateSweep()' and reading in azelsweep values in 'antcli'.
//
// Revision 1.59  2000/06/20 13:42:33  amn
// Added 'tSweep' structs to sh.h for both ra-dec and az-el.
// Still missing 'updateSweep()' and antcli:'readSweep()'.
//
// Revision 1.58  2000/03/13 09:21:38  amn
// Changed 'antcli'-less default for YES button to start VAX tracking.
//
// Revision 1.57  2000/03/10 11:27:25  amn
// BBC osc frequency calculation in LINEFREQ/SOURCEVEL.
//
// Revision 1.56  2000/02/20 14:46:20  amn
// Removed reading keyboard input as "EIO errors" continue.
//
// Revision 1.55  2000/02/20 14:22:33  amn
// Added "safe position" limits to all tracking modes.
// Antcli now reports the true encoder position with 'azel'.
//
// Revision 1.54  2000/02/17 13:46:23  amn
// Added line frequency Doppler shift calculation.
//
// Revision 1.53  2000/02/16 11:59:12  amn
// Added many commands to 'antcli'.
//
// Revision 1.52  2000/02/09 08:02:59  amn
// Changed "VAX tracking" from own coordinates back to VAX coordinates.
//
// Revision 1.51  2000/02/06 17:56:36  amn
// Added the basic commands what the FS needs for pointing measurements.
//
// Revision 1.50  2000/02/04 13:55:32  amn
// Added 'source=' command sh-> variables.
//
// Revision 1.49  2000/02/02 15:25:18  amn
// *** empty log message ***
//
// Revision 1.48  2000/02/02 13:52:52  amn
// Added shared memory support and the first variables
// for 'antcli' 'source=' command line.
//
// Revision 1.47  2000/01/30 13:51:08  amn
// Tracks 'dr21' as old VAX vlbi_tracker_p.
//
// Revision 1.46  2000/01/28 13:59:51  amn
// First self-calculated (Sun and DR21) tracking.
//
// Revision 1.45  2000/01/28 13:01:24  amn
// 'cPointing.cc' + Novas 2.0 library + vMyTarget added for own tracking.
//
// Revision 1.44  1999/12/20 10:40:10  amn
// Service mode Motor 1 knob was affecting both motor 1 and 2...
//
// Revision 1.43  1999/11/22 12:54:49  amn
// Added xdm wait delay, disable 'r' for ref mark search.
//
// Revision 1.42  1999/11/22 11:18:07  amn
// Fixed manual offset knobs; setting of man pos knob initial values.
//
// Revision 1.41  1999/11/20 19:47:43  amn
// Revised panel state machine extensively.
//
// Revision 1.40  1999/11/18 09:24:24  amn
// Added slewing/tracking LED lamp indicators based on tracking error.
//
// Revision 1.39  1999/11/11 14:02:38  amn
// Slewing/tracking LEDs.
//
// Revision 1.38  1999/11/05 08:40:00  amn
// Minor cleanups.
//
// Revision 1.37  1999/11/02 11:14:41  amn
// Added manual az-el-pwm with tighten gearbox uplink command.
//
// Revision 1.36  1999/11/01 17:57:24  amn
// New xntpd time and interlock type.
//
// Revision 1.35  1999/11/01 15:06:13  amn
// Added UTC/xntpd time from antcond in uplink to ant/dos/up/ant.cpp.
//
// Revision 1.34  1999/10/28 18:11:28  amn
// Fixing panel offsets correctly (added to what VAX requests).
//
// Revision 1.33  1999/10/28 08:29:20  amn
// Separated cTime class to a separate file.
//
// Revision 1.32  1999/10/22 12:54:41  amn
// Receiving VAX position coordinates and sending them upstairs.
//
// Revision 1.31  1999/10/21 12:38:48  amn
// Added loop timing.
//
// Revision 1.30  1999/10/21 10:37:03  amn
// Added time handling; vPanel towards a real state machine.
//
// Revision 1.29  1999/10/20 12:12:32  amn
// Prepared for manual PWM for az/el motors.
//
// Revision 1.28  1999/10/20 11:49:21  amn
// Changed knob handling into a 'cKnob' class.
//
// Revision 1.27  1999/10/18 12:54:40  amn
// Added interlock --> panel LEDs.
//
// Revision 1.26  1999/10/15 12:40:53  amn
// Changed knobs to be logarithmic (or **2, actually).
//
// Revision 1.25  1999/10/14 12:22:24  amn
// Added manual select knob.  Just before adding new down/uplink structs.
//
// Revision 1.24  1999/10/12 14:18:25  amn
// Tried prolonging "antcon" ok LED, no real success.
//
// Revision 1.23  1999/10/11 13:16:27  amn
// Added real serial downlink with long int 'length'.
//
// Revision 1.22  1999/10/07 12:26:43  amn
// Added serial up/downlink.
//
// Revision 1.21  1999/10/07 09:19:56  amn
// Added ANTBOX panel support.
//
// Revision 1.20  1999/10/07 06:24:28  vphan
// Tho's completed version.
//


// Configuration/debugging defines.
#define REAL_SERIAL_CONNECTION 1
#define REAL_VAX_CONNECTION 1
#define ENABLE_KEY_INPUT 1
// Trouble with SLsys_getkey returning "EIO error".

// Potential parameters, now constants.
#define SLOWDOWNWAIT 3.5  // seconds
#define AUTOSTANDBYWAIT 15.0  // seconds

// Antenna "safe" position limits (for tracking).
// xxx: (Now these are the same as computer-compared
//       slow limits in ant/dos/up/ant.cpp...)
#if SUNANT
 #define MINTRACKAZ (-160.0)
 #define MAXTRACKAZ (+160.0)
 #define MINTRACKEL (+2.0)
 #define MAXTRACKEL (+95.0)
#else
 #define MINTRACKAZ (-240.0)
 #define MAXTRACKAZ (+240.0)
 #define MINTRACKEL (+4.0)
 #define MAXTRACKEL (+100.0)
//// 150GHz antenna:
// #define MAXTRACKEL (+40.0)
#endif

#define MYNAME "antcon: "


#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <assert.h>
/* This is for terminal handling */ 
#include <slang.h>

// For shared memory.
#include <sys/ipc.h>
#include <sys/shm.h>

/* The direct I/O port reference macros reside here. */
/* They work _only_ if the source files using them are compiled */
/* with '-O' optimization flag on; they rely on 'inline' code generation. */
// Used to be: #include <asm/io.h>
// Now in Debian slink 2.1:
// (Disable 'volatile' in conjunction with 'asm' to get rid of extra warnings
//  which appear with g++ egcs-2.91.60:)
#define __volatile__
#include <sys/io.h>


/* -------------------------------------------------------------------- */
// Serial port connection to upstairs pedestal DOS PC.
#include "cSerialLine.hh"

#define SERIAL_PORT_NAME "/dev/ttyS1"
#define baudRate 115200L
// #define baudRate 9600L

class cSerialLine sl;


/* -------------------------------------------------------------------- */
// New VAXIF MACH211-based VAX interface board.
class cVAXIF
{
private:
  unsigned int baseAddr;
  virtual long int getOne(int offset);
  virtual long int getOneUntilStaysSame(int offset);
  virtual double getConverted(int offset);
  virtual void sendOne(int offset, long int value);
  virtual void sendConverted(int offset, double value);
public:
  virtual void init(int baseAddr);
  virtual void cleanup(void);
  virtual double getAz(void);
  virtual double getEl(void);
  virtual void sendAz(double az);
  virtual void sendEl(double el);
};  /* cVAXIF */

void
cVAXIF::init(int baseAddr)
{
  if (ioperm(
      /*from address=>*/         baseAddr,
      /*number of addresses=>*/  8,
      /*turn on=>*/              1  /* 0 to turn off */
  ) < 0) {
    fprintf(stderr, MYNAME "access to I/O port 0x%x denied\n", baseAddr);
    exit(EXIT_FAILURE);
  }
  this->baseAddr = baseAddr;
  /* xxx: Set outputs to VAX to ? */
}

void
cVAXIF::cleanup(void) 
{
}

long int cVAXIF::getOne(int offset)
{
  long int value;
#if REAL_VAX_CONNECTION
  value = inb(this->baseAddr+offset+2);  /* high */
  value = (value<<8) | inb(this->baseAddr+offset+1);  /* mid */
  value = (value<<8) | inb(this->baseAddr+offset);  /* low */
    /* At the moment the board doesn't have any special latching */
    /* based on the order in which bytes are being read. */
#else
  value = 1500000;
#endif
  return value;
}

long int
cVAXIF::getOneUntilStaysSame(int offset)
{
  long int prevValue;
  long int value;
  int retries = 0;

  value = this->getOne(offset);
  do 
    {
      prevValue = value;
      value = this->getOne(offset);
      retries++;
    } while ((prevValue != value) && (retries < 10));
  /* xxx: what to do if retries > 10? */
  return value;
}

double
cVAXIF::getConverted(int offset)
{
  long int value;

  value = this->getOneUntilStaysSame(offset);

  /* Extend MSB == 1  => negative in 2-complement. */
  if (value & 0x00200000)
    {
      /* MSB sign bit == 1, extend. */
      value |= 0xFFE00000;
    }
  return ((double)value * 360.0 / 2097152.0);
  /* getConverted */
}

void
cVAXIF::sendOne(int offset, long int value)
{
#if REAL_VAX_CONNECTION
  value <<= 2;  /* VAX aligns MSBs, so fill two LSBs with 0 */
  outb(value & 0x000000FF, this->baseAddr+offset);  /* low */
  outb((value & 0x0000FF00)>>8, this->baseAddr+offset+1);  /* mid */
  outb((value & 0x00FF0000)>>16, this->baseAddr+offset+2);  /* high */
  /* High byte latches all of 24(22/21) bits to VAX. */
#endif
}

void
cVAXIF::sendConverted(int offset, double value)
{
  long int signBit;

  if (value < 0.0)
    {
      value = -value;
      signBit = 0x00000000;
    } 
  else 
    {
      signBit = 0x00200000;  /* positive == sign bit is on; not 2-compl. */
    }
  this->sendOne(offset, (long int)(value * 2097152.0 / 360.0) | signBit);
}  /* sendConverted */

double
cVAXIF::getAz(void)
{
  return this->getConverted(/*offset=>*/0);
}  /* getAz */

double
cVAXIF::getEl(void)
{
  return this->getConverted(/*offset=>*/4);
}  /* getEl */

void
cVAXIF::sendAz(double az) {
  this->sendConverted(/*offset=>*/0, az);
}  /* sendEl */

void
cVAXIF::sendEl(double el)
{
  this->sendConverted(/*offset=>*/4, el);
}  /* sendEl */


/* ------------------------------------------------------------------ */
// Time handling.
#define LINUX_TIME_AVAIL 1
#include "ctime.cpp"

/* ------------------------------------------------------------------ */
// Interlock chain explanations.
#include "interloc.cpp"

/* ------------------------------------------------------------------ */
/* Definition of the struct of all sensor values on the serial line down. */
#include "downlink.h"

typedef union
{
  tDownlink downlink;
  unsigned char buff[sizeof(tDownlink)];
} tDLBuff;


int
getVal(char *prompt)
{
  char buff[255];
  int temp;

  SLang_reset_tty();
  SLsmg_printf("%s", prompt);
  SLsmg_refresh();
  // gets(buff);  // dangerous, says g++ with -Wall
  fgets(buff, sizeof(buff), stdin);
  SLang_init_tty(-1,0,1);
  (void)sscanf(buff, "%d", &temp);
  return (temp);
}

double
getDouble(char *prompt)
{
  char buff[255];
  double temp;

  SLang_reset_tty();
  SLsmg_printf("%s", prompt);
  SLsmg_refresh();
  fgets(buff, sizeof(buff), stdin);
  SLang_init_tty(-1,0,1);
  (void)sscanf(buff, "%lg", &temp);
  return (temp);
}


// DOS-->Linux I/O macros.
#define outportb(a,v) outb(v,a)
#define inportb(a) inb(a)
#define LINUX_IOPERM 1

/* ------------------------------------------------------------------ */
// ACL-712-style Opto-22 50-pin parallel interface (8255 chips) board.
#include "c8255.cpp"

/* ------------------------------------------------------------------ */
// QDR 4-connector quadrature encoder decoding board,
// connected to one connector on ACL-712 6x3x8bit digital I/O board.
#include "cqdr.cpp"

/* ------------------------------------------------------------------ */
// cKnob "remembers" the last raw encoder value (unsigned 16-bit int)
// and knows how to scale linearly or quadratically
// to a given range of values.
class cKnob {
private:
  // Per-encoder "constant" values.
  class cQDR *myQDR;
  int encoderNumber;  // on a given QDR board
  int countsPerTurn;  // encoder property

  // Encoder "tracking" values.
  unsigned int prevRawValue;  // what was seen last time in encoder
  long int relCounter;  // raw encoder counts since last reset

  // Scaling values.
  long int zeroTolerance;  // return 0.0 from -z..0..z of 'relCounter'
  int toPower;  // 1==linear, 2==squared, 3==third pwr, ...
  double turnsForFullRange;
  double lowLimit;
  double highLimit;
  // How much (0.0..1.0...) to take into account delta^2
  // i.e. the faster one turns, the faster the readout goes.
  double accel;
  // Calculated/derived scaling values.
  double range;  // (high - low)
  double coeff;  // range / (total number of encoder ticks)
  long int minRel;  // the allowed range of raw 'relCounter'
  long int maxRel;
  double minDiv;  // for zero-corrected 'relCounter' division
  double maxDiv;
  double multiples;  // >0.0, only exact multiples of this allowed as result

public:
  virtual void init(
    class cQDR *aQDR,
    int en,
    int cpt
  );
  virtual void cleanup(void);

  virtual void reset(void);  // relCounter to zero
  virtual void setScaling(
    double ll,
    double hl,
    double turns = 1.0,  // # of turns to achieve ll..hl range
    long int zt = 0,
    int tp = 1,
    double acc = 0.0,
    double mult = 0.0  // >0.0, this is the exact multiple allowed
  );

  virtual double get(void);

  virtual void set(double v);  // relCounter to desired value
};  // cKnob


void
cKnob::init(
  class cQDR *aQDR,
  int en,
  int cpt
) {
  // Per-encoder "constant" values.
  this->myQDR = aQDR;
  this->encoderNumber = en;
  this->countsPerTurn = cpt;

  // Init raw "tracking" counter values.
  this->reset();

  // Init scaling values (to defaults, one turn linearly in range -1..1.).
  this->setScaling(/*ll=>*/ -1.0, /*hl=>*/ 1.0);
}  // init

void
cKnob::cleanup(void) {
  this->myQDR = NULL;
}  // cleanup

// Establish the current raw encoder value as the new relative zero.
void
cKnob::reset(void) {
  this->prevRawValue = this->myQDR->get(this->encoderNumber);
  this->relCounter = 0;
}  // reset

// Set the result range and behavior of knob.
void
cKnob::setScaling(
  double ll,
  double hl,
  double turns = 1.0,  // # of turns to achieve ll..hl range
  long int zt = 0,
  int tp = 1,
  double acc = 0.0,
  double mult = 0.0
) {
  this->lowLimit = ll;
  this->highLimit = hl;
  assert( this->highLimit > this->lowLimit );
  this->turnsForFullRange = turns;
  this->zeroTolerance = zt;
  this->toPower = tp;
  this->accel = acc;
  this->multiples = mult;

  // Calculated/derived scaling values.
  this->range = (this->highLimit - this->lowLimit);
  this->coeff = this->turnsForFullRange * (double)this->countsPerTurn
    / this->range;
  {
    double mi = this->lowLimit * this->coeff;
    double ma = this->highLimit * this->coeff;

    this->minDiv = fabs(mi);  // for zero-corrected 'relCounter' division
    this->minRel = (long int)mi - this->zeroTolerance;
    this->maxDiv = fabs(ma);
    this->maxRel = (long int)ma + this->zeroTolerance;
  }
}  // setScaling

// Get a new reading from encoder and convert it to scaled, in range double.
double
cKnob::get(void) {
  {  // Encoder --> relCounter.
    unsigned int a;
    signed long int l;

    // Get a new raw encoder value.
    a = this->myQDR->get(this->encoderNumber);

    // Update relative long counter, delta is always -32768..32767.
    l = (long)(a) - (long)(this->prevRawValue);
    if (l > (long)(0x7fff)) {
      l -= (long)0x10000;
    } else if (l < -(long)(0x8000)) {
      l += (long)0x10000;
    }
    {  // Acceleration; delta squared with a coefficient.
      double l2 = l * l;

      if (l < 0) {
	l2 = -l2;
      }
      this->relCounter += l + (long int)(this->accel * l2);
    }

    // Now this new raw location is the reference to next readout.
    this->prevRawValue = a;
  }

  { // Keeping relCounter in range.
    signed long int l;
    double rc;
    double t;
    double div;
    double tdiv;
    int e;

    // Clipping back to allowed raw value range.
    l = this->relCounter;
    l = (l > this->maxRel) ? this->maxRel : l;
    l = (l < this->minRel) ? this->minRel : l;
    this->relCounter = l;

    // Provide a range of ("center pos") values where
    // the knob stays==0.0.
    if (l > this->zeroTolerance) {
      l -= this->zeroTolerance;
    } else if (l < -(this->zeroTolerance)) {
      l += this->zeroTolerance;
    } else {
      l = 0;
    }

    // Convert 'relCounter' to double for exponentiation.
    t = rc = fabs( (double)l );

    // Acceleration by exponentiation divided by min/max.
    // (I.e. (x^2)/(max^2).
    // Done in "abs()" mode for negative and positive values.
    tdiv = div = fabs( (this->relCounter < 0) ? this->minDiv : this->maxDiv );
    for (e=1; e < this->toPower; e++) {
      t *= rc;
      tdiv *= div;
    }  // for

    // Negative values follow lowLimit (and its sign).
    if (this->relCounter < 0) {
      t = (t / tdiv) * this->lowLimit;
    } else {
      t = (t / tdiv) * this->highLimit;
    }

    // Round to the nearest multiple.
    if (this->multiples > 0.0) {
      t = this->multiples * rint(t / this->multiples);
    }  // if rounding
    return (t);
  }
}  // get

// Reset the counter and calculate 'relCounter' so that the counter
// starts with the given value 'v'.
void
cKnob::set(double v) {
  // Cannot do this (yet) to power-corrected knob scales.
  assert( this->toPower == 1.0 );
  this->reset();  // sync up to the current raw value of encoder
  this->relCounter = (long int)(v * this->coeff);
  if (v < 0.0) {
    this->relCounter -= this->zeroTolerance;
  } else {
    this->relCounter += this->zeroTolerance;
  }
}  // set


/* ------------------------------------------------------------------ */
// LEDIM 16-bit off/dim/lit/bright LEDs plus 24 input switches.
// connected to one connector on ACL-712 6x3x8bit digital I/O board.
class cLEDIM : public c8255 {
private:
  unsigned int dimbits;  // latest bit values loaded onto LEDIM board&chip
  unsigned int bribits;
public:
  virtual void init(
	int baseAddr,
	int conn,
    int addressing
  );
  virtual void setLeds(unsigned int leds, int brightness);
  virtual void refreshLeds(void);
  virtual void cleanup(void);
};  // cLEDIM

void
cLEDIM::init(
  int baseAddr,  /* of ET1722B card (or any 8255-compatible parallel) */
  int conn,
  int addressing
) {
  c8255::init(baseAddr, conn, addressing);

  // All ports ABC are outputs.
  this->setPortABCMode(
	/*A=>*/ OUT, /*B=>*/ OUT, /*Cupper=>*/ OUT, /*Clower=>*/ OUT
  );
  this->dimbits = 0;
  this->bribits = 0;
}

void
cLEDIM::cleanup() {
  c8255::cleanup();
}

// Port PC definitions of LEDIM.
#define OUT_dim_bright_bit 0x01
#define OUT_load_bit       0x02

// Set LEDs in mask 'leds' at the same time to desired brightness.
void
cLEDIM::setLeds(
  unsigned int leds,
  int brightness
) {
  if ((brightness & 0x01)) {  // dim bit on
    this->dimbits |= leds;  // add dim bits
  } else {
    this->dimbits &= ~(leds);  // turn off dim bits
  }
  this->out(/*offset=>*/ 0, (this->dimbits & 0xff));  // LSB LED lines
  this->out(/*offset=>*/ 1, (this->dimbits >> 8));  // MSB LED lines
  this->out(/*offset=>*/ 2, 0x0);  // select dim bits
  this->out(/*offset=>*/ 2, OUT_load_bit);  // latch dim bits ("LOADCLK")
  // Delay turning load to 0 to keep "panel ok" LED longer on
  this->out(/*offset=>*/ 2, 0x0);  // back to default
  // Seemed to cause glitches in LEDIM, wrong LEDs on...

  if ((brightness & 0x02)) {  // bright bit on
    this->bribits |= leds;
  } else {
    this->bribits &= ~(leds);
  }
  this->out(/*offset=>*/ 0, (this->bribits & 0xff));  // LSB LED lines
  this->out(/*offset=>*/ 1, (this->bribits >> 8));  // MSB LED lines
  this->out(/*offset=>*/ 2, OUT_dim_bright_bit);  // select bright bits
  // and LOAD signal back to 0 here at the latest
  this->out(/*offset=>*/ 2, (OUT_load_bit | OUT_dim_bright_bit));  // latch bright bits
  // delay(1);  // to make latch clock strobe signal long enough to see in LED
  // usleep(1);  // minor, minor Linux sleep: will it be fast enough?
  // looks at least terrible, "flashing"...
  this->out(/*offset=>*/ 2, 0x0);  // LOAD clock signal back to default

  return;
}  // setLeds

// Set all LEDs again on LEDIM board (and flash "Panel ok" LED):
void
cLEDIM::refreshLeds(void) {
  this->setLeds(/*leds=>*/ 0 /*no change*/, 0);
}  // refreshLeds

// LEDIM board bit assignments used in ANTBOX panel system.
#define LED_stop           0x0001
#define SW_stop          0x000001
#define LED_standby        0x0002
#define SW_standby       0x000002
#define LED_auto_standby   0x0004
#define LED_slewing        0x0008
#define LED_tracking       0x0010
#define LED_spare1         0x0020
#define LED_az_off         0x0040
#define LED_el_off         0x0080
#define LED_yes_ok         0x0100
#define SW_yes_ok        0x000100
#define LED_no_cancel      0x0200
#define SW_no_cancel     0x000200
#define LED_manual         0x0400
#define SW_manual        0x000400
#define LED_service        0x0800
#define SW_service       0x000800
#define LED_reset_tower_pc 0x1000
#define LED_check          0x2000
#define SW_check         0x002000
#define LED_spare2         0x4000
#define LED_spare3         0x8000
// Port PC definitions of LEDIM.
// (two bits reserved for loading LED dim/bright bits)
// (Six extra +24VDC on/off outputs.)
#define OUT_spare1         0x04
#define OUT_spare2         0x08
#define OUT_spare3         0x10
#define OUT_spare4         0x20
#define OUT_spare5         0x40
#define OUT_spare6         0x80

#if SUNANT
// For SunAnt we have six LEDs connected to 'sunant' LPT1 parallel port
// (which is actually 8255-compatible).
// We just re-map LED bit numbers.
class cLPTLEDIM : public cLEDIM {
private:
  unsigned char bits;
public:
  virtual void init(
	int baseAddr,
	int conn,
    int addressing
  );
  virtual void setLeds(unsigned int leds, int brightness);
  virtual void cleanup(void);
};  // cLPTLEDIM

void
cLPTLEDIM::init(
  int baseAddr,  /* of ET1722B card (or any 8255-compatible parallel) */
  int conn,
  int addressing
) {
  cLEDIM::init(baseAddr, conn, addressing);

  this->bits = 0;
  this->out(/*offset=>*/ 0, this->bits);
}

void
cLPTLEDIM::cleanup() {
  this->bits = 0;
  this->out(/*offset=>*/ 0, this->bits);

  cLEDIM::cleanup();
}

// Set LEDs in mask 'leds' at the same time to desired brightness.
void
cLPTLEDIM::setLeds(
  unsigned int leds,
  int brightness
) {
  if (leds & LED_check) {
    if (brightness) {  // anything >0 means on
      this->bits |=   0x04;  // red severe failure; 0x10 is yellow failure
    } else {
      this->bits &= ~(0x04);
    }
  }
  if (leds & LED_slewing) {
    if (brightness) {
      this->bits |=   0x20;
    } else {
      this->bits &= ~(0x20);
    }
  }
  if (leds & LED_tracking) {
    if (brightness) {
      this->bits |=   0x08;
    } else {
      this->bits &= ~(0x08);
    }
  }
  if (leds & LED_auto_standby) {
    if (brightness) {
      this->bits |=   0x03;  // Az and El stop
    } else {
      this->bits &= ~(0x03);
    }
  }
  // Toggle the state of 0x10 (former yellow failure) "antcon ok" LED.
  this->bits ^= 0x10;  // xor with 1 inverts
  this->out(/*offset=>*/ 0, this->bits);
}  // setLeds
#endif


/* ------------------------------------------------------------------ */
// LEDSW 16-bit off/dim/lit/bright LEDs plus 24 input switches.
// connected to one connector on ACL-712 6x3x8bit digital I/O board.
class cLEDSW : public c24In {
public:
  virtual unsigned long int get(void);
};

// Get all 24 switch bits.
unsigned long int
cLEDSW::get(void) {
  // We invert the 24 switch bits; after this 1==switch on.
  return (~(c24In::get()) & 0x00ffffff);
}  // get


/* ------------------------------------------------------------------ */
// The combination of QDR and LEDIM boards together make an ANTBOX panel.
class cPanel {
private:
  // The panel consists of QDR encoders ("three knobs"),
  // a set of 16 LEDs and 24 input switches.
  class cQDR *myQDR;
  class cLEDIM *myLEDIM;
  class cLEDSW *myLEDSW;
  int azElOrder;  // 1==Az left, El right (0 vice versa)

  // Run mode / state of this panel.
  // (States borrowed from uplink commands from 'downlink.h'.)
  // Adjusted additional states, never seen on uplink:
#define CC_standby_slowdown 1
#define CC_servoing_standby 2
#define CC_manual_offsets 4
  int runMode;
  int prevRunMode;
  int runModeBeforeManualOffsets;
  int inManualMode;
  int inServiceMode;
  int yesNoSelectActive;

  // Standby timed state transitions.
  tTime startOfStandbySlowdown;
  tTime startOfStandbyServoing;

  // Manual az/el knobs (for both speed and position).
  class cKnob azKnob;
  class cKnob elKnob;
  // Service mode manual (az or el) motor1/motor2 PWM knobs.
  class cKnob motor1PWMKnob;
  class cKnob motor2PWMKnob;

public:
  virtual void init(
    class cQDR *aQDR,
    class cLEDIM *aLEDIM,
    class cLEDSW *aLEDSW
  );
  virtual void cleanup(void);
  virtual void changeAzElOrder(int order);

  virtual void activateYesNoSelect(void);
  virtual void deactivateYesNoSelect(void);

  virtual unsigned int getSelectKnob(void);
  virtual void resetSelectKnob(void);
  virtual double getAzKnob(void);
  virtual double getElKnob(void);
  virtual double getM1Knob(void);
  virtual double getM2Knob(void);

  virtual void forceState(
    unsigned long int sw,
    tInterlockBits interlockBits,
    tCoord pos,
    tCoord off
  );
  virtual void run(
    tInterlockBits interlockBits,
    tCoord pos,
    tCoord off
  );
  virtual void setStatus(
    int check,
    int slew,
    int track,
    int autoStandby
  );
  // State machine.
  virtual int whatRunMode(void);
  virtual void setRunMode(int rm);
  virtual int standbySlowdownComplete(void);
  virtual int servoStandbyForLongEnough(void);
};  // cPanel

void
cPanel::init(
  class cQDR *aQDR,
  class cLEDIM *aLEDIM,
  class cLEDSW *aLEDSW
) {
  this->myQDR = aQDR;
  this->myLEDIM = aLEDIM;
  this->myLEDSW = aLEDSW;

  // Default to new knob order, Az first (left), El then (right).
  // (Init knobs but leave scaling undetermined until first used.)
  this->changeAzElOrder(/*order=>*/ 1);

  // Motor PWM service/manual knobs.
  this->motor1PWMKnob.init(
    /*aQDR=>*/ this->myQDR,
    /*en=>*/ 0,
    /*countsPerTurn=>*/ 4*512
  );
  this->motor1PWMKnob.setScaling(
    /*ll=>*/ -1023.0,
    /*hl=>*/ +1023.0,
    /*turns=>*/ 1.5,
    /*zeroTolerance=>*/ 100,
    /*toPower=>*/ 1,
    /*accel=>*/ 0.0
  );
  this->motor2PWMKnob.init(
    /*aQDR=>*/ this->myQDR,
    /*en=>*/ 1,
    /*countsPerTurn=>*/ 4*512
  );
  this->motor2PWMKnob.setScaling(
    /*ll=>*/ -1023.0,
    /*hl=>*/ +1023.0,
    /*turns=>*/ 1.5,
    /*zeroTolerance=>*/ 100,
    /*toPower=>*/ 1,
    /*accel=>*/ 0.0
  );

  // All LEDs off.
  this->myLEDIM->setLeds(/*leds=>*/ 0xffff, /*brightness=>*/ 0);
  // this->runMode state unknown, can be stopped or running...
  // xxx: let's try:
  this->runMode = CC_brake_standby;
  this->prevRunMode = CC_brake_standby;
  this->inManualMode = 0;
  this->inServiceMode = 0;
  this->yesNoSelectActive = 0;
}  // init

void
cPanel::cleanup(void) {
}  // cleanup

void
cPanel::changeAzElOrder(int order) {
  this->azElOrder = order;

  this->azKnob.init(
    /*aQDR=>*/ this->myQDR,
    /*en=>*/ (this->azElOrder) ? 0 : 1,
    /*countsPerTurn=>*/ 4*512
  );
  this->elKnob.init(
    /*aQDR=>*/ this->myQDR,
    /*en=>*/ (this->azElOrder) ? 1 : 0,
    /*countsPerTurn=>*/ 4*512
  );
  // 'this->az/elKnob.setScaling()' deferred to the point when the knobs
  // are being activated for either manual speed or position use.
}  // changeAzElOrder

void
cPanel::activateYesNoSelect(void) {
  // Activate "Yes/No" (and "Select" knob).
  this->yesNoSelectActive = 1;

  // Az/El knobs cannot be reset here anymore because they are reseted
  // only in manual velocity mode, preset in pos/offset mode.
  // this->azKnob.reset();
  // this->elKnob.reset();
  // Establish relative zero for motor knobs.
  this->motor1PWMKnob.reset();
  this->motor2PWMKnob.reset();
}  // activateYesNoSelect

void
cPanel::deactivateYesNoSelect(void) {
  // Deactivate "Yes/No" (and "Select" knob).
  this->yesNoSelectActive = 0;
}  // deactivateYesNoSelect

unsigned int
cPanel::getSelectKnob(void) {
  return (this->myQDR->get(2));
}  // getSelectKnob

void
cPanel::resetSelectKnob(void) {
  return (this->myQDR->reset(2));
}  // resetSelectKnob

double
cPanel::getAzKnob(void) {
  return (this->azKnob.get());
}  // getAzKnob

double
cPanel::getElKnob(void) {
  return (this->elKnob.get());
}  // getElKnob

double
cPanel::getM1Knob(void) {
  return (this->motor1PWMKnob.get());
}  // getM1Knob

double
cPanel::getM2Knob(void) {
  return (this->motor2PWMKnob.get());
}  // getKnob

int
cPanel::whatRunMode(void) {
  return (this->runMode);
}  // whatRunMode

void
cPanel::setRunMode(int rm) {
  // The panel supports only setting different standby modes.
  assert(
    ((this->runMode == CC_standby_slowdown) && (rm == CC_servoing_standby))
    || ((this->runMode == CC_servoing_standby) && (rm == CC_brake_standby))
  );
  if (rm == CC_servoing_standby) {
    getNow(this->startOfStandbyServoing);
  }
  this->runMode = rm;
}  // setRunMode

int
cPanel::standbySlowdownComplete(void) {
  tTime now;
  tTime td;

  if (this->whatRunMode() != CC_standby_slowdown) {
    return 1;
  }
  // In slowdown mode, it is complete if enough time has
  // expired (xxx: really, if antenna has stopped).
  getNow(now);
  subTime(now, this->startOfStandbySlowdown, td);
  if (dsecs(td) > SLOWDOWNWAIT) {
    return 1;
  }
  return 0;
}  // standbySlowdownComplete

int
cPanel::servoStandbyForLongEnough(void) {
  tTime now;
  tTime td;

  if (this->whatRunMode() != CC_servoing_standby) {
    return 1;
  }
  // In servo standby mode, if enough time has expired
  // put brakes on.
  getNow(now);
  subTime(now, this->startOfStandbyServoing, td);
  if (dsecs(td) > AUTOSTANDBYWAIT) {
    return 1;
  }
  return 0;
}  // servoStandbyForLongEnough


// ----- Handle panel state machine.
void
cPanel::forceState(
  unsigned long int sw,
  tInterlockBits interlockBits,
  tCoord pos,
  tCoord off
 ) {
  // Enable detecting state changes.
  this->prevRunMode = this->runMode;

  // STOP overrides everything else, turns all other LEDs off
  // and lits only STOP brightly.
  if (sw & SW_stop) {
    this->deactivateYesNoSelect();
    this->myLEDIM->setLeds(/*leds=>*/ 0xffff, /*brightness=>*/ 0);
    this->myLEDIM->setLeds(/*leds=>*/ LED_stop, /*brightness=>*/ 3);

  } else {
    // Any other switch pressed but not STOP, turn STOP LED off.
    // Hmm, not any other, but standby, manual, or service.
    if (sw & (SW_standby | SW_manual | SW_service)) {
      this->myLEDIM->setLeds(/*leds=>*/ LED_stop, /*brightness=>*/ 0);
    }

    // STANDBY returns back to "normal" where the panel is functional.
    if (sw & SW_standby) {
      // Stationary LEDs in "normal" mode.
      this->myLEDIM->setLeds(
        /*leds=>*/ LED_manual | LED_service | LED_reset_tower_pc,
        /*brightness=>*/ 1
      );
      this->deactivateYesNoSelect();
      this->runMode = CC_standby_slowdown;
      getNow(this->startOfStandbySlowdown);
      this->inManualMode = 0;
      this->inServiceMode = 0;
      // State LEDs.
      this->myLEDIM->setLeds(/*leds=>*/ LED_standby, /*brightness=>*/ 3);
      // xxx: Capture the current antenna position
      // to be used in staying in the same position.

    // MANUAL.
    } else if (sw & SW_manual) {
      this->activateYesNoSelect();
      this->resetSelectKnob();  // to default to man panel offsets
      this->inManualMode = 1;
      this->inServiceMode = 0;
      // State LEDs.
      this->myLEDIM->setLeds(/*leds=>*/ LED_manual, /*brightness=>*/ 3);
      this->myLEDIM->setLeds(/*leds=>*/ LED_service, /*brightness=>*/ 1);
      this->myLEDIM->setLeds(/*leds=>*/ LED_standby, /*brightness=>*/ 1);

    // SERVICE.
    } else if (sw & SW_service) {
      this->activateYesNoSelect();
      this->inManualMode = 0;
      this->inServiceMode = 1;
      // State LEDs.
      this->myLEDIM->setLeds(/*leds=>*/ LED_service, /*brightness=>*/ 3);
      this->myLEDIM->setLeds(/*leds=>*/ LED_manual, /*brightness=>*/ 1);
      this->myLEDIM->setLeds(/*leds=>*/ LED_standby, /*brightness=>*/ 1);

    // xxx: Yes/no, check.

    // If nothing is pressed, still blink "Panel OK" LED.
    } else {
      this->myLEDIM->refreshLeds();
    }
  }  // end else not STOP

  // Update run mode from panel selection knob.
  if (this->inManualMode) {
    unsigned int s;

    s = this->getSelectKnob() % 1024;  // 1/2 turn to 3 functions
    if (s < 342) {
      if (this->prevRunMode != CC_manual_offsets) {
        this->runModeBeforeManualOffsets = this->runMode;
        this->runMode = CC_manual_offsets;
        this->azKnob.setScaling(
          /*ll=>*/ -10.24,
          /*hl=>*/ +10.24,
          /*turns=>*/ 200.0,
          /*zeroTolerance=>*/ 100,
          /*toPower=>*/ 1,
          /*accel=>*/ 0.0,  // xxx: 0.5 was super-fast
          /*mult=>*/ 0.001
        );
        this->elKnob.setScaling(
          /*ll=>*/ -10.24,
          /*hl=>*/ +10.24,
          /*turns=>*/ 200.0,
          /*zeroTolerance=>*/ 100,
          /*toPower=>*/ 1,
          /*accel=>*/ 0.0,
          /*mult=>*/ 0.001
        );
        // Initial values when changing to manual offset mode.
        this->azKnob.set(off.az);
        this->elKnob.set(off.el);
      }  // if changed
    } else if (s < 2*342) {
      if (this->prevRunMode != CC_manual_position) {
        this->runMode = CC_manual_position;
        this->azKnob.setScaling(
          /*ll=>*/ -240.0,
          /*hl=>*/ +240.0,
          /*turns=>*/ 6.0,
          /*zeroTolerance=>*/ 0,
          /*toPower=>*/ 1,
          /*accel=>*/ 0.0  // xxx: 0.5 was super-fast
        );
        this->elKnob.setScaling(
          /*ll=>*/ 0.5,
          /*hl=>*/ 110.0,
          /*turns=>*/ 2.0,
          /*zeroTolerance=>*/ 0,
          /*toPower=>*/ 1,
          /*accel=>*/ 0.0
        );
        // Initial values when changing to manual position mode.
        this->azKnob.set(pos.az - off.az);
        this->elKnob.set(pos.el - off.el);
      }  // if changed
    } else {
      if (this->prevRunMode != CC_manual_velocity) {
        this->runMode = CC_manual_velocity;
        this->azKnob.setScaling(
          /*ll=>*/ -1.2,
          /*hl=>*/ +1.2,
          /*turns=>*/ 1.5,
          /*zeroTolerance=>*/ 100,
          /*toPower=>*/ 2,
          /*accel=>*/ 0.0
        );
        this->elKnob.setScaling(
          /*ll=>*/ -1.2,
          /*hl=>*/ +1.2,
          /*turns=>*/ 1.5,
          /*zeroTolerance=>*/ 100,
          /*toPower=>*/ 2,
          /*accel=>*/ 0.0
        );
        // Initial values to 0.0 when changing to manual velocity mode.
        this->azKnob.reset();
        this->elKnob.reset();
      }  // if changed
    }
  } else if (this->inServiceMode) {
    unsigned int s;

    s = this->getSelectKnob() % 1024;  // 1/2 turn to 3 functions
    if (s < 342) {
      this->runMode = CC_manual_azeltighten_pwm;
    } else if (s < 2*342) {
      this->runMode = CC_manual_az_pwm;
    } else {
      this->runMode = CC_manual_el_pwm;
    }
  } else {  // not in special manual/service mode
    // We are in "normal" panel mode, determine if we can VAX-track.
    // xxx: temp test: secret 'yes' button to activate VAX tracking
    // xxx: should be eliminated and changed to 'antcliMode'-based change
    if (sw & SW_yes_ok) {
      this->runMode = CC_track_position;
      this->myLEDIM->setLeds(/*leds=>*/ LED_standby, /*brightness=>*/ 1);
    }
    // Notice expired "servo the same place" mode.
    if ((this->whatRunMode() == CC_servoing_standby)
        && this->servoStandbyForLongEnough()
    ) {
      this->setRunMode(CC_brake_standby);
    }
  }  // end else in normal mode

  // Handle manual az/el off override buttons.
  // Latched panel switches turn bits off in interlock chain.
  this->myLEDIM->setLeds(
    /*leds=>*/ LED_az_off,
    /*brightness=>*/ (interlockBits & IL_az_off) ? 0 : 3
  );
  this->myLEDIM->setLeds(
    /*leds=>*/ LED_el_off,
    /*brightness=>*/ (interlockBits & IL_el_off) ? 0 : 3
  );

  // Check Yes/no.
  if (this->yesNoSelectActive) {
    // Yes button.
    if (sw & SW_yes_ok) {
      this->myLEDIM->setLeds(
	/*leds=>*/ LED_yes_ok,
	/*brightness=>*/ 3
      );
      if (this->whatRunMode() == CC_manual_offsets) {
        // Accept manual offsets.
        this->runMode = this->runModeBeforeManualOffsets;
        this->inManualMode = 0;
        this->deactivateYesNoSelect();
        this->myLEDIM->setLeds(/*leds=>*/ LED_manual, /*brightness=>*/ 1);
      }
    } else {  // active, but nothing pressed
      this->myLEDIM->setLeds(
        /*leds=>*/ LED_yes_ok,
	/*brightness=>*/ 1
      );
    }

    // No button.
    if (sw & SW_no_cancel) {
      this->myLEDIM->setLeds(
	/*leds=>*/ LED_no_cancel,
	/*brightness=>*/ 3
      );
      if (this->whatRunMode() == CC_manual_offsets) {
        // Discard manual offsets.
        // xxx:
        this->runMode = this->runModeBeforeManualOffsets;
        this->inManualMode = 0;
        this->deactivateYesNoSelect();
        this->myLEDIM->setLeds(/*leds=>*/ LED_manual, /*brightness=>*/ 1);
      }
    } else {  // active, but nothing pressed
      this->myLEDIM->setLeds(
	/*leds=>*/ LED_no_cancel,
	/*brightness=>*/ 1
      );
    }
  } else {  // yes/no not active
    this->myLEDIM->setLeds(
      /*leds=>*/ (LED_yes_ok | LED_no_cancel),
      /*brightness=>*/ 0
    );
  }  // else yes/no not active
}  // forceState

void
cPanel::run(
  tInterlockBits interlockBits,
  tCoord pos,
  tCoord off
) {
  this->forceState(
    /*switches=>*/ this->myLEDSW->get(),
    interlockBits,
    pos,
    off
  );
}  // run

void
cPanel::setStatus(
  int check,
  int slew,
  int track,
  int autoStandby
) {
  this->myLEDIM->setLeds(
    /*leds=>*/ LED_check,
    /*brightness=>*/ check
  );
  this->myLEDIM->setLeds(
    /*leds=>*/ LED_slewing,
    /*brightness=>*/ slew
  );
  this->myLEDIM->setLeds(
    /*leds=>*/ LED_tracking,
    /*brightness=>*/ track
  );
  this->myLEDIM->setLeds(
    /*leds=>*/ LED_auto_standby,
    /*brightness=>*/ autoStandby
  );
}  // setStatus


/* ------------------------------------------------------------------ */
// Astronomy formulae / ephemeric conversions.
#include "../../af/met/cPointing.cc"


/* ------------------------------------------------------------------ */
// Auxiliary routines to 'main()'.
void sendUp(
  long int command,
  double az,
  double el
) {
  union {
    char line[81];
    tUplink uplink;
  };

  assert( gettimeofday(&(uplink.time), NULL) == 0 );
  // xxx: need to fix for PWMISA usecs
  uplink.command = command;
  uplink.cc.az = az;
  uplink.cc.el = el;
  uplink.length = sizeof(uplink);
  sl.flushPendingInput();
  sl.write("+", 1);
  sl.write(line, sizeof(uplink));
}  // sendUp


/*------------------------------------------------------------*/
// The complete and common shared struct
// across ant/dos/up/ant.cpp -- antcond (==antcon+antcli) -- xantcon.
//
// ant.cpp <--> antcond is transmitted field-by-field using COM2 serial,
// antcond <--> xantcon uses shared memory and the same struct layout.
#include "sh.h"

// Global shared struct.
// tSh vSh;  // no more, this is now shared memory...
// And a pointer to it, for shared memory references in Linux with 'sh->'.
// ptSh sh = &vSh;  // not initialized anymore with static struct address
ptSh sh = NULL;


/* ------------------------------------------------------------------ */
// Recalculate the current sweep (ra-dec or az-el) offsets
// based on current time.
void
updateSweep(ptSweep p, double now)
{
  double st = p->startTime;  // a convenience short name, needed often

  // If not active, no offsets.
  if (!p->isActive) {
    p->status = ' ';
    p->o.x = p->o.y = 0.0;
    return;
  }

  // Calculate sweep time (0.0 == start of real constant velocity sweep).
  p->t = now - st;
  // ('Remaining' is calculated as (p->dur - p->t).)

  if (now < (st - p->accdur)) {  // before acceleration starts
    p->status = 'W';
    p->o = p->accStart;  // goto start position
    return;
  }

  if (now < st) {  // before start of constant velocity: accelerate
    double t = p->t + p->accdur;  // 0.0 --> accdur

    p->status = 'A';
    t = t * t;
    p->o.x = p->accStart.x + p->acca.x * t * 0.5;
    p->o.y = p->accStart.y + p->acca.y * t * 0.5;
    return;
  }

  // xxx: p->startError = ???

  if (now < (st + p->dur)) {  // during sweep
    p->status = 'S';
    p->o.x = p->start.x + p->v.x * (now - st);
    p->o.y = p->start.y + p->v.y * (now - st);
    return;
  }

  if (now < (st + p->dur + p->decdur)) {  // decelerate after sweep
    double t = p->decdur - (p->t - p->dur);  // decdur --> 0.0

    p->status = 'D';
    t = t * t;
    p->o.x = p->decEnd.x + p->deca.x * t * 0.5;
    p->o.y = p->decEnd.y + p->deca.y * t * 0.5;
    return;
  }

  if (now < (st + p->dur + p->decdur + p->keepdur)) {  // keep end point
    p->status = 'K';
    p->o = p->decEnd;  // goto end position
    return;
  }

  // All time elapsed, deactivate sweep and return to zero offsets.
  p->status = ' ';
  p->o.x = p->o.y = 0.0;
  p->isActive = 0;
}  // updateSweep


// Timestamp ASCII formatting for log files.
#include "tstamp.c"



/* ------------------------------------------------------------------ */
int main(int argc, char **argv)
{
  class cVAXIF vVAXIF;

  class cLEDSW vLEDSW;
#if SUNANT
  class cLPTLEDIM vLEDIM;
#else
  class cLEDIM vLEDIM;
#endif
  class cQDR vQDR;
  class cPanel vPanel;

  tDLBuff dl;
  // Last time we got a message from upstairs in antcon.cc.
  class cTime vDownlinkTime;
  // (xxx: Note: classes cannot be easily put into shared struct 'sh->'
  // since their constructor functions won't probably be run.)

  cNovasPointing vNovas;

#if ENABLE_KEY_INPUT
  // (Obsolete) Keyboard az/el jump amount.
  double vJumpDegrees = 0.0;
  char ch = '\0';
#endif

  long int vSerialReadErrors = 0;

  // Shared memory unique identifier (key) and the returned reference id.
  key_t shKey;
  int shId = -1;

  // unpriviledged user name
  char *my_user_name = NULL;

  // pid file name
  char *my_pid_file = NULL;

  // Create and initialize 'antcon--antcli--xantcon' shared memory.
  shKey = fourCharLong('a','n','t','c');
  assert( (shId = shmget(shKey, sizeof(tSh), IPC_CREAT | 0777)) != -1 );
  assert( (sh = (ptSh)shmat(shId, NULL, 0)) != (void *)-1 );
  sh->utc = 0.0;
  sh->target.az = sh->target.el = 0.0;
  sh->clipLimitMin.az = MINTRACKAZ;
  sh->clipLimitMin.el = MINTRACKEL;
  sh->clipLimitMax.az = MAXTRACKAZ;
  sh->clipLimitMax.el = MAXTRACKEL;
  sh->slewLimit = 0.001;  // degrees, usually overridden from 'antcli'
  sh->panelOffset.az = sh->panelOffset.el = 0.0;
  sh->myOffset.az = sh->myOffset.el = 0.0;
  sh->myTarget.az = sh->myTarget.el = 0.0;
  sh->myFixedPos.az = sh->myFixedPos.el = 0.0;
  sh->VAXTarget.az = sh->VAXTarget.el = 0.0;
  sh->manualTarget.az = sh->manualTarget.el = 0.0;
  sh->manualSpeed.az = sh->manualSpeed.el = 0.0;
  sh->standbyPos.az = sh->standbyPos.el = 0.0;

  // The default we assume 'antcli' wants us to do is
  // VAX tracking; however, we do start up in 'standby' mode.
  // (This means "secret YES" button from panel will VAX track.)
  sh->antcliUpdates = 0;  // incremented by 'antcli', decr by 'antcon'
  sh->antcliMode = ACM_VAX_track;
  // Pointing parameter updates via shared memory.
  sh->getPP = 0;  // incremented by 'antcli', decr by 'antcon'
  sh->updatePP = 0;  // incremented by 'antcli', decr by 'antcon'

  // xxx: default source?
  sh->appRa = sh->appDec = 0.0;  // until calculated by Novas
  sh->raOffset = sh->decOffset = 0.0;

  // Init both ra-dec and az-el sweeps to be inactive:
  sh->radecsweep.isActive = sh->azelsweep.isActive = 0;
  sh->radecsweep.o.x = sh->azelsweep.o.x = 0.0;
  sh->radecsweep.o.y = sh->azelsweep.o.y = 0.0;

  // Init logging to disabled state.
  sh->log.isActive = 0;

  // parse arguments
  {
	  int option_char;
	  while ((option_char = getopt (argc, argv, "U:P:")) != EOF)
		  switch (option_char) {
		  case 'U': {
			  assert ((my_user_name = strdup(optarg)) != NULL);
			  // TODO: figure out where priviledges can be dropped
			  break;
		  }
		  case 'P': {
			  assert ((my_pid_file = strdup(optarg)) != NULL);
			  break;
		  }
		  case ':':  // missing parameter after a option letter
		  case '?': {
			  fprintf (stderr,
				   "usage: %s [-U<user>] [-P<pidfile>]\n", argv[0]);
			  exit(EXIT_FAILURE);
		  }
		  }
  }
  // Write pid file
  if (my_pid_file) {
	  FILE *fp;
	  assert((fp = fopen(my_pid_file, "w")) != NULL);
	  fprintf(fp, "%d\n", getpid());
	  fclose(fp);
  }

  /* New VAXIF MACH211-based VAX interface board. */
  /* Supports both getting VAX commanded az/el and */
  /* sending encoder az/el values back to VAX. */
  vVAXIF.init(
    /*baseAddr=>*/ 0x300
  );

  vQDR.init(
	/*baseAddr=>*/ 0x220,
	/*conn=>*/ 0,
    /*addressing=>*/ 0  // normal 4 separate registers mode
  );
#if SUNANT
  vLEDIM.init(
	/*baseAddr=>*/ 0x378,
	/*conn=>*/ 0,
    /*addressing=>*/ 0  // normal 4 separate registers mode
  );
#else
  vLEDIM.init(
	/*baseAddr=>*/ 0x220,
	/*conn=>*/ 1,
    /*addressing=>*/ 0  // normal 4 separate registers mode
  );
#endif
  vLEDSW.init(
	/*baseAddr=>*/ 0x220,
	/*conn=>*/ 2,
    /*addressing=>*/ 0  // normal 4 separate registers mode
  );
  vPanel.init(&vQDR, &vLEDIM, &vLEDSW);

  /* Serial up/downlink. */
  sl.open(SERIAL_PORT_NAME);
  // Can later use 'sl.whatPathname()' to access the stored serial device path.
  sl.serialSettings(/*baudRate=>*/ baudRate,
		    /*numberOfDataBits=>*/ 8,
		    /*numberOfStopBits=>*/ 1,
		    /*whichParity=>*/ eParityNone,
		    /*minBytes=>*/ 0,
		    /*maxTimeout=>*/ 2);  // 0.2 seconds
  // /*maxTimeout=>*/ 5);  // 0.5 seconds

  {  // Init downlink struct; if no serial conn, stays the same.
    memset(&dl, 0, sizeof(dl));
    dl.downlink.interlock = ~(0);  // "all ok"
    dl.downlink.length = sizeof(dl);
  }

  /* Initialize S-Lang routines and set in/out handling. */
#define COL0 4
#define COLP 7
#define COL1 26
#define COL2 45
#define COL3 64
  SLang_init_slang();
  // From 'slang.h':
   /* This function is mandatory and must be called by all applications that
    * use the interpreter */

  SLtt_get_terminfo();
  SLsmg_init_smg();
  SLang_init_tty(-1,0,1);
  // xxx: init_SLfiles(); ???
  SLang_init_slfile();  // "called if fputs, fgets, etc are need in S-Lang"

  SLsmg_gotorc(0, 1);
  SLsmg_printf("$Id: antcon.cc,v 1.85 2005/10/05 12:35:07 lindi Exp $");
  SLsmg_gotorc(1, COLP);
  SLsmg_printf("                Azimuth           Elevation  (degrees)");
  SLsmg_gotorc(2, COLP);
  SLsmg_printf("Absolute");
  SLsmg_gotorc(3, COLP);
  SLsmg_printf("Incremental");
  SLsmg_gotorc(4, COLP);
  SLsmg_printf("Target");
  SLsmg_gotorc(5, COLP);
  SLsmg_printf("Pos. error");

  SLsmg_gotorc(7, COLP);
  SLsmg_printf("Antcon target");
  SLsmg_gotorc(8, COLP);
  SLsmg_printf("VAX target");
  SLsmg_gotorc(9, COLP);
  SLsmg_printf("Panel offsets");
  SLsmg_gotorc(10, COLP);
  SLsmg_printf("Antcon offsets");
  SLsmg_gotorc(11, COLP);
  SLsmg_printf("Az-El Sweep");

  SLsmg_gotorc(13, COLP);
  SLsmg_printf("Apparent RA/Decl");
  SLsmg_gotorc(14, COLP);
  SLsmg_printf("RA/Decl offsets");
  SLsmg_gotorc(15, COLP);
  SLsmg_printf("RA-Decl Sweep");

  SLsmg_gotorc(18, COLP);
  SLsmg_printf("Manual pos.");
  SLsmg_gotorc(19, COLP);
  SLsmg_printf("Manual speed");
  SLsmg_gotorc(20, COLP);
  SLsmg_printf("Service motor");

#if ENABLE_KEY_INPUT
  SLsmg_gotorc(24, 1);
  SLsmg_printf("Press 'q' to stop.");
#endif

  // Start with the assumption of "normal" operation.
  {
    static const tCoord zeroCoord = {0.0, 0.0};

    vPanel.forceState(/*sw=>*/ SW_standby, /*interlock=>*/ ~0,
                      /*pos=>*/ zeroCoord, /*off=>*/ zeroCoord);
  }

  // A test source for Novas library:
#if 0
  vNovas.setSun();
#else
  {
    double ra =  20.0 + 37.0/60.0 + 14.3/60.0/60.0;
    double dec = 42.0 +  9.0/60.0 +  0.0/60.0/60.0;
    vNovas.setSource("dr21", ra, dec, /*epoch=>*/ 1950.0);
  }
#endif

  // The main event processing loop.
  for (;;) {

#if REAL_SERIAL_CONNECTION
    // Attempt getting fresh values; if fails, uses old values
    // for the next iteration of the main 'for' loop.

    /* The other end is not supposed to send any responses
       before receiving a command, make sure by: */
    sl.flushPendingInput();

    /* Ask for all sensor values by sending a command character. */
    sl.write("*",1);

    /* Try to read exactly the expected number of reply bytes. */
    {
      tDLBuff dltemp;
      size_t i;
      
      i = sl.read(dltemp.buff, sizeof(tDownlink));
      vDownlinkTime.now();
      // xxx: should compare dl.time and vDownlinkTime...
      if ( i != sizeof(tDownlink) ) {
	// Too few characters, ignore.
	vSerialReadErrors++;
      } else {
	/* Message must have correct length at end. */
	if (dltemp.downlink.length != sizeof(dltemp)) {
	  vSerialReadErrors++;
	} else {
	  // Successful read, copy to real use.
	  dl = dltemp;
	}
      }
    }
#else
    vDownlinkTime.now();
#endif

    // Copy downlink stuff into its proper place in 'sh->'.
    sh->interlock = dl.downlink.interlock;
    sh->abs = dl.downlink.abs;  // struct copy
    sh->incr = dl.downlink.incr;  // struct copy

    // Detect "interlock stopped" (and xxx: lost brake currents (fuse)).
    if ( (sh->interlock & 0x0fffff) == 0x0 ) {
      // Complete stop, all interlocks active, i.e. 'STOP' pressed.
      vPanel.forceState(/*sw=>*/ SW_stop, sh->interlock,
                        /*pos=>*/ sh->incr,
                        /*off=>*/ sh->panelOffset);
    }
    
    /* Update error display signals. */
    /* We compare the fresh 'incr' encoder readouts */
    /* to the previously-requested target coordinates. */
    {
      // Used to be analog control of ESSCO servo.
      /* We want -1.0--+1.0 degree error to match
	 the maximum D/A output range which will be
	 adjusted with resistors (at the moment 10V -> 1.2V). */
      /* +-1.0 degree comes from the original ESSCO 11-bit D/A
	 which took input in BCD millidegrees, -0.999--+0.999. */

      sh->posErr.az = sh->target.az - sh->incr.az;
      sh->posErr.el = sh->target.el - sh->incr.el;
    }

    /* Handling of position error log files. */
    if (sh->log.isActive) {
      // Active logs must have the log file open.
      if (sh->log.filep == NULL) {
        /* Open log file always in append mode. */
        sh->log.filep = fopen(sh->log.filename, "at");  // append in text mode
        if (sh->log.filep == NULL) {
          // If cannot open, deactivate and put error message back to filename.
          snprintf(sh->log.filename, sizeof(sh->log.filename),
                   "fopen(): %s (%d)",
                   strerror(errno),
                   errno);
          sh->log.isActive = 0;
        }  // if open error
        // Zero min/ave/max azel error fields.
        sh->log.minErr.az = sh->log.minErr.el = 999.0;
        sh->log.nErr = 0;
        sh->log.sumErr.az = sh->log.sumErr.el = 0.0;
        sh->log.maxErr.az = sh->log.maxErr.el = -999.0;
      }  // if not already open

      // Accumulate every loop azel error.
        if (sh->posErr.az < sh->log.minErr.az)
          sh->log.minErr.az = sh->posErr.az;
        if (sh->posErr.el < sh->log.minErr.el)
          sh->log.minErr.el = sh->posErr.el;
        sh->log.nErr++;
        sh->log.sumErr.az += sh->posErr.az;
        sh->log.sumErr.el += sh->posErr.el;
        if (sh->posErr.az > sh->log.maxErr.az)
          sh->log.maxErr.az = sh->posErr.az;
        if (sh->posErr.el > sh->log.maxErr.el)
          sh->log.maxErr.el = sh->posErr.el;

      // If log line writing interval has elapsed, write one line.
      if ((sh->utc > sh->log.nextTime) && (sh->log.filep)) {
        fprintf(sh->log.filep,
                "%s\t%11.6f\t%11.6f\t%11.6f\t%11.6f\t%11.6f\t%11.6f\t%11.6f\t%11.6f\t%d\t%11.6f\t%11.6f\t%11.6f\t%11.6f\t%11.6f\t%11.6f\n",
                toTimestamp(sh->utc, NULL),  // the time from prev. loop
                sh->incr.az,
                sh->incr.el,
                sh->target.az,
                sh->target.el,
                sh->appRa,
                sh->appDec,
                sh->myTarget.az,
                sh->myTarget.el,
                sh->log.nErr,
                sh->log.minErr.az,
                sh->log.minErr.el,
                (sh->log.nErr > 0) ? sh->log.sumErr.az / (double)sh->log.nErr : 0.0,
                (sh->log.nErr > 0) ? sh->log.sumErr.el / (double)sh->log.nErr : 0.0,
                sh->log.maxErr.az,
                sh->log.maxErr.el
                );
        // Next line not sooner than after 'incr' seconds.
        sh->log.nextTime = sh->utc + sh->log.incr;
        // Zero min/ave/max azel error fields.
        sh->log.minErr.az = sh->log.minErr.el = 999.0;
        sh->log.nErr = 0;
        sh->log.sumErr.az = sh->log.sumErr.el = 0.0;
        sh->log.maxErr.az = sh->log.maxErr.el = -999.0;
      }  // if needs to write a log line

      // If the whole duration has been written (or premature termination
      // is requested, close and stop.
      if (((sh->utc > sh->log.endTime) || (sh->log.isActive > 1))
          && (sh->log.filep)) {
        (void)fclose(sh->log.filep);
        sh->log.filep = NULL;
        sh->log.isActive = 0;
      }  // if the file is ready and complete
    }  // if log active

    /* Input handling, max. one key per large 'for' loop */
#if ENABLE_KEY_INPUT
    if (SLang_input_pending(0)) {

       ch = SLang_getkey();

      switch (ch) {
      case 'q':
      case 'Q':
	goto lEnd;
#if 0
      case 'r': {
	sl.write("r",1);
	break;
      }
#endif
      case 'l': {
	sl.write("l",1);
	break;
      }
      case 'a': {
	if (vPanel.whatRunMode() == CC_track_position) {
	  sh->panelOffset.az = getDouble(" Azimuth offset (deg) ->");
	} else if (vPanel.whatRunMode() == CC_manual_position) {
	  sh->manualTarget.az = getDouble(" Azimuth deg   ->");
	}
	break;
      }
      case 'e': {
	if (vPanel.whatRunMode() == CC_track_position) {
	  sh->panelOffset.el = getDouble(" Elevation offset (deg) ->");
	} else if (vPanel.whatRunMode() == CC_manual_position) {
	  sh->manualTarget.el = getDouble(" Elevation deg ->");
	}
	break;
      }
      case 'j': {
	vJumpDegrees = getDouble(" Jump amount deg ->");
	break;
      }
      case '8': {
	sh->manualTarget.el += vJumpDegrees;
	break;
      }
      case '2': {
	sh->manualTarget.el -= vJumpDegrees;
	break;
      }
      case '4': {
	sh->manualTarget.az -= vJumpDegrees;
	break;
      }
      case '6': {
	sh->manualTarget.az += vJumpDegrees;
	break;
      }
      }  // switch
    }  // if keystrokes available
#endif

    // The following is what needs to be done in "background".

    // Run panel keys, LEDs, and knobs.
    // (Inform panel about the state of interlock chain,
    //  current antenna position and panel offsets and
    //  let it update knob-based values.)
    vPanel.run(sh->interlock,
               sh->incr,
               sh->panelOffset);

    // See if we've got command line based changes from 'antcli'
    // into shared memory.
    // (For as long as the counter stays positive, we'll
    // do one source update per each loop pass.)
    // First 'pp_*' pointing parameters.
    if (sh->updatePP > 0) {
      sh->foundPP = vNovas.updateParam(sh->ppname, sh->ppvalue);
      sh->updatePP--;
    }
    if (sh->getPP > 0) {
      sh->foundPP = vNovas.getParam(sh->ppname, &(sh->ppvalue));
      sh->getPP--;
    }
    if (sh->antcliUpdates > 0) {
      // What 'antcli' wants us to do?
      switch (sh->antcliMode) {
        // Does 'antcli' want a standby stop?
        case ACM_standby: {
          vPanel.forceState(/*sw=>*/ SW_standby, sh->interlock,
                            /*pos=>*/ sh->incr,
                            /*off=>*/ sh->panelOffset);
          break;
        }
        // Does 'antcli' want a fixed az-el position?
        case ACM_azel: 
        case ACM_VAX_track: {
          // Start tracking az-el.
          vPanel.forceState(/*sw=>*/ SW_yes_ok, sh->interlock,
                            /*pos=>*/ sh->incr,
                            /*off=>*/ sh->panelOffset);
          break;
        }
        // Does 'antcli' want source changed?
        case ACM_source: {
          if (sh->isPlanet) {
            // Is a solar system planet.
            vNovas.setPlanet(sh->sourcePlanetNumber);
            // xxx: maybe one 'updateApparent()' to get the following:
            // xxx: copy calculated ra, dec, and today's epoch
            //      to 'sh->source*'
          } else {
            // Is a regular object.
            vNovas.setSource(sh->sourceName,
                             sh->sourceRa,
                             sh->sourceDec,
                             sh->sourceEpoch,
                             sh->sourceRadialvel,
                             sh->sourcePromora,
                             sh->sourcePromodec,
                             sh->sourceParallax);
          }
          // Start tracking this source.
          vPanel.forceState(/*sw=>*/ SW_yes_ok, sh->interlock,
                            /*pos=>*/ sh->incr,
                            /*off=>*/ sh->panelOffset);
          break;
        }
        default: {
          // xxx: nothing, no known 'sh->antcliMode'
          break;
        }
      }  // switch
      sh->antcliUpdates--;
    }  // if updates

    // Get the time when to calculate the coordinates.
    assert( gettimeofday(&(sh->coord_calc_utc_tv), /*tz=>*/ NULL) == 0 );
    sh->utc = dsecs(sh->coord_calc_utc_tv);
    sh->tdtJD = vNovas.tdt_jd_utc(sh->coord_calc_utc_tv);
    // xxx: other time calculations to 'sh->' (and to display)?
    // sh->utcJD ???
    // sh->utcMJD ???
    // sh->fractional year ("epoch") ???
    // sh->DOY ???
    // sh->eq_eq ??? via vNovas.testEEST() or its variant
    // sh->gmst, gast, lst

    // Calculate our own Novas 2.0 library -based target coordinates. 
    if (sh->antcliMode == ACM_source) {
      // Feed in ra-dec sweep increments and/or ra-dec offsets (from antcli).
      updateSweep(&(sh->radecsweep), sh->utc);
      vNovas.setRadecOff(
        (sh->radecsweep.o.x + sh->raOffset),
        (sh->radecsweep.o.y + sh->decOffset)
      );

      vNovas.updateApparent(sh->tdtJD);

      // Copy (especially) planet ra/dec to 'sh->app*'.
      vNovas.getApp(&(sh->appRa), &(sh->appDec));

      //// Override az/el with old VAX ephemeric formulae:
      //vNovas.convertUTC(tv);  // old VAX routine name...

      vNovas.applyCorrections();
      sh->myTarget.az = vNovas.getFinalAz();
      sh->myTarget.el = vNovas.getFinalEl();
#if SUNANT
      // "Track" the Sun under Earth during nights...
      if (sh->myTarget.el < 0.0) {
        sh->myTarget.el = -(sh->myTarget.el);
      }
#endif
    } else {
      // xxx: ??? sh->appRa = sh->appDec = 0.0;
    }

    // Get VAX-requested position.
    sh->VAXTarget.az = vVAXIF.getAz();
    sh->VAXTarget.el = vVAXIF.getEl();

    // xxx:
    // Recognize that VAX has stopped / commands the same az-el position
    // and change the run mode to CC_servo_standby
    // (which then timeouts to CC_brake_standby, "AUTO STANDBY").

    /* Update VAX parallel encoder output. */
    {
      double vAzAsEssco = sh->incr.az - sh->azelsweep.o.x - sh->panelOffset.az - sh->myOffset.az;
      double vElAsEssco = sh->incr.el - sh->azelsweep.o.y - sh->panelOffset.el - sh->myOffset.el;
      vVAXIF.sendAz(vAzAsEssco);
      vVAXIF.sendEl(vElAsEssco);
    }

    // Get panel offset updates from vPanel.
    if (vPanel.whatRunMode() == CC_manual_offsets) {
      // Update panel offsets from the knobs.
      sh->panelOffset.az = vPanel.getAzKnob();
      sh->panelOffset.el = vPanel.getElKnob();
    }
    // Get manual position updates from vPanel.
    if (vPanel.whatRunMode() == CC_manual_position) {
      // Update manual position from the knobs.
      sh->manualTarget.az = vPanel.getAzKnob();
      sh->manualTarget.el = vPanel.getElKnob();
    }
 
    /* If we are tracking, update target position. */
    if ((vPanel.whatRunMode() == CC_track_position)
        || (vPanel.whatRunMode() == CC_manual_offsets)) {
      tCoord t;;

      // Where should we servo the antenna?
      switch (sh->antcliMode) {
        case ACM_azel: {
          t = sh->myFixedPos;
          break;
        }
        case ACM_source: {
          t = sh->myTarget;
          break;
        }
        case ACM_VAX_track: {
          t = sh->VAXTarget;
          break;
        }
        default: {
          // If we don't know, use fixed position
          // (including: we shouldn't get here with ACM_standby).
          t = sh->myFixedPos;
          break;
        }
      }  // switch

      // Feed in az-el sweep increments.
      updateSweep(&(sh->azelsweep), sh->utc);

      // Apply sweep, antcon, and panel offsets.
      t.az = t.az + sh->azelsweep.o.x + sh->myOffset.az + sh->panelOffset.az;
      t.el = t.el + sh->azelsweep.o.y + sh->myOffset.el + sh->panelOffset.el;

      // Enforce minimum/maximum limits for calculated
      // (non-manual) antenna positions.
      // Temporary, user-settable.
      if (t.az < sh->clipLimitMin.az) {
        t.az = sh->clipLimitMin.az;
      }
      if (t.az > sh->clipLimitMax.az) {
        t.az = sh->clipLimitMax.az;
      }
      if (t.el < sh->clipLimitMin.el) {
        t.el = sh->clipLimitMin.el;
      }
      if (t.el > sh->clipLimitMax.el) {
        t.el = sh->clipLimitMax.el;
      }
      // Hard-coded safety limits.
      if (t.az < MINTRACKAZ) {
        t.az = MINTRACKAZ;
      }
      if (t.az > MAXTRACKAZ) {
        t.az = MAXTRACKAZ;
      }
      if (t.el < MINTRACKEL) {
        t.el = MINTRACKEL;
      }
      if (t.el > MAXTRACKEL) {
        t.el = MAXTRACKEL;
      }
      sh->target.az = t.az;
      sh->target.el = t.el;

    } else if (vPanel.whatRunMode() == CC_manual_position) {
      // Or update target to be the manual (knob-based) position.
      sh->target.az = sh->manualTarget.az + sh->panelOffset.az;
      sh->target.el = sh->manualTarget.el + sh->panelOffset.el;

    } else if (vPanel.whatRunMode() == CC_servoing_standby) {
      // Or update target to be the position where antenna stopped.
      sh->target.az = sh->standbyPos.az;
      sh->target.el = sh->standbyPos.el;
    }

    // Update the new target position in position/servo modes.
    if ( (vPanel.whatRunMode() == CC_track_position)
           || (vPanel.whatRunMode() == CC_manual_offsets)
           || (vPanel.whatRunMode() == CC_servoing_standby)
           || (vPanel.whatRunMode() == CC_manual_position) )
    {
      sendUp(
        /*command=>*/ CC_track_position,
        /*az=>*/ sh->target.az,
        /*el=>*/ sh->target.el
      );

      /* Update panel LED status. */
      {
        double vErr;
        int vSlew, vTrack;

        vErr = sqrt(sh->posErr.az*sh->posErr.az + sh->posErr.el*sh->posErr.el);
        // xxx: may need to consider using the same limits as 'antcli'
        // xxx: what about cosine correction for high elevations?
        // xxx: should perhaps take into account beam width, especially
        // with high frequencies
#if SUNANT
        vTrack = (vErr < 0.030) ? 3 : ((vErr < 0.035) ? 2 : ((vErr < 0.040) ? 1 : 0));
        vSlew  = (vErr < 0.030) ? 0 : ((vErr < 0.035) ? 1 : ((vErr < 0.040) ? 1 : 3));
#else
        vTrack = (vErr < 0.001) ? 3 : ((vErr < 0.002) ? 2 : ((vErr < 0.005) ? 1 : 0));
        vSlew  = (vErr < 0.001) ? 0 : ((vErr < 0.002) ? 1 : ((vErr < 0.005) ? 1 : 3));
#endif
        sh->vectPosErr = vErr;
        vPanel.setStatus(
          /* check brightness => */ 0,
          /* slewing brightness => */ vSlew,
          /* tracking brightness => */ vTrack,
          /* auto-standby brightness => */ 0
        );
      }  // update LED status display
    } else {
      // Non-servo (direct motor control) modes.
      // Update LED status for non-servoing modes.
      vPanel.setStatus(
        /* check brightness => */ 0,
        /* slewing brightness => */ 0,
        /* tracking brightness => */ 0,
        /* auto-standby brightness => */
          (vPanel.whatRunMode() == CC_brake_standby) ? 3 : 0
      );
      if (vPanel.whatRunMode() == CC_standby_slowdown) {
        sendUp(/*command=>*/ CC_manual_velocity,
               /*az=>*/ 0.0,
               /*el=>*/ 0.0);
        sh->manualSpeed.az = sh->manualSpeed.el = 0.0;
        sh->knobMot1pwm = sh->knobMot2pwm = 0.0;
        if (vPanel.standbySlowdownComplete()) {
          // Update the latest position seen while slowing down
          // so that we can later servo on that.
          sh->standbyPos.az = sh->incr.az;
          sh->standbyPos.el = sh->incr.el;
          // Change to servo-standby mode.
          vPanel.setRunMode(CC_servoing_standby);
        }
      } else if (vPanel.whatRunMode() == CC_brake_standby) {
        sendUp(/*command=>*/ CC_brake_standby,
               /*az=>*/ 0.0,
               /*el=>*/ 0.0);
        sh->manualSpeed.az = sh->manualSpeed.el = 0.0;
        sh->knobMot1pwm = sh->knobMot2pwm = 0.0;
      } else if (vPanel.whatRunMode() == CC_manual_velocity) {
        // If we are running at manual velocity, get it from panel.
        sendUp(/*command=>*/ CC_manual_velocity,
               /*az=>*/ (sh->manualSpeed.az = vPanel.getAzKnob()),
               /*el=>*/ (sh->manualSpeed.el = vPanel.getElKnob()));
        sh->knobMot1pwm = sh->knobMot2pwm = 0.0;
      } else if (
        (vPanel.whatRunMode() == CC_manual_azeltighten_pwm)
          || (vPanel.whatRunMode() == CC_manual_az_pwm)
          || (vPanel.whatRunMode() == CC_manual_el_pwm)
        ) {
        sendUp(/*command=>*/ vPanel.whatRunMode(),
               /*az=>*/ (sh->knobMot1pwm = vPanel.getM1Knob()),
               /*el=>*/ (sh->knobMot2pwm = vPanel.getM2Knob()));
        sh->manualSpeed.az = sh->manualSpeed.el = 0.0;
      } else {
      }  // else cannot handle run mode, do nothing
    }  // else a non-servo mode

    /* Update display values. */
    {
      SLsmg_gotorc(2, COL1);
      SLsmg_printf("%6.1f", sh->abs.az);
      SLsmg_gotorc(2, COL2);
      SLsmg_printf("%6.1f", sh->abs.el);

      /* Remote IK121 values. */
      SLsmg_gotorc(3, COL1);
      SLsmg_printf("%11.6f", sh->incr.az);
      SLsmg_gotorc(3, COL2);
      SLsmg_printf("%11.6f", sh->incr.el);

      SLsmg_gotorc(4, COL1);
      SLsmg_printf("%11.6f", sh->target.az);
      SLsmg_gotorc(4, COL2);
      SLsmg_printf("%11.6f", sh->target.el);

      SLsmg_gotorc(5, COL1);
      SLsmg_printf("%11.6f", sh->posErr.az);
      SLsmg_gotorc(5, COL2);
      SLsmg_printf("%11.6f", sh->posErr.el);
      SLsmg_gotorc(5, COL3);
      if (sh->log.isActive) {
        SLsmg_printf("Log %-11s", sh->log.filename);
      } else {
        SLsmg_printf("(no logging)");
      }
      SLsmg_erase_eol();

      /* xxx: vDiffVoltsEl, vDiffVoltsAz */

      SLsmg_gotorc(7, COL0);
      SLsmg_printf("%c",
        ((vPanel.whatRunMode() == CC_track_position)
           && (sh->antcliMode == ACM_source))
          ? '>' : ' '
      );
      SLsmg_gotorc(7, COL1);
      SLsmg_printf("%11.6f", sh->myTarget.az);
      SLsmg_gotorc(7, COL2);
      SLsmg_printf("%11.6f", sh->myTarget.el);

      SLsmg_gotorc(8, COL0);
      SLsmg_printf("%c",
        ((vPanel.whatRunMode() == CC_track_position)
           && (sh->antcliMode == ACM_VAX_track))
          ? '>' : ' '
      );
      SLsmg_gotorc(8, COL1);
      SLsmg_printf("%11.6f", sh->VAXTarget.az);
      SLsmg_gotorc(8, COL2);
      SLsmg_printf("%11.6f", sh->VAXTarget.el);

      SLsmg_gotorc(9, COL0);
      SLsmg_printf("%c", (vPanel.whatRunMode() == CC_manual_offsets) ? '>' : ' ');
      SLsmg_gotorc(9, COL1);
      SLsmg_printf("%11.6f", sh->panelOffset.az);
      SLsmg_gotorc(9, COL2);
      SLsmg_printf("%11.6f", sh->panelOffset.el);

      SLsmg_gotorc(10, COL1);
      SLsmg_printf("%11.6f", sh->myOffset.az);
      SLsmg_gotorc(10, COL2);
      SLsmg_printf("%11.6f", sh->myOffset.el);

      SLsmg_gotorc(11, (COL1-1) );
      SLsmg_printf("%c", sh->azelsweep.status);
      SLsmg_gotorc(11, COL1);
      SLsmg_printf("%11.6f", sh->azelsweep.o.x);
      SLsmg_gotorc(11, COL2);
      SLsmg_printf("%11.6f", sh->azelsweep.o.y);


      if (sh->antcliMode == ACM_source) {
        SLsmg_gotorc(13, COL1);
        SLsmg_printf("%11.6f", sh->appRa);
        SLsmg_gotorc(13, COL2);
        SLsmg_printf("%11.6f", sh->appDec);
        SLsmg_gotorc(13, COL3);
        SLsmg_printf("%-15s", sh->sourceName);
      } else {
        SLsmg_gotorc(13, COL1);
        SLsmg_printf("%11.6f", 0.0);
        SLsmg_gotorc(13, COL2);
        SLsmg_printf("%11.6f", 0.0);
        SLsmg_gotorc(13, 64);
        SLsmg_printf("%-15s", "(no source)");
      }
      SLsmg_erase_eol();

      SLsmg_gotorc(14, COL1);
      SLsmg_printf("%11.6f", sh->raOffset);
      SLsmg_gotorc(14, COL2);
      SLsmg_printf("%11.6f", sh->decOffset);

      SLsmg_gotorc(15, (COL1-1) );
      SLsmg_printf("%c", sh->radecsweep.status);
      SLsmg_gotorc(15, COL1);
      SLsmg_printf("%11.6f", sh->radecsweep.o.x);
      SLsmg_gotorc(15, COL2);
      SLsmg_printf("%11.6f", sh->radecsweep.o.y);

      SLsmg_gotorc(18, COL0);
      SLsmg_printf("%c", (vPanel.whatRunMode() == CC_manual_position) ? '>' : ' ');
      SLsmg_gotorc(18, COL1);
      SLsmg_printf("%11.6f", sh->manualTarget.az);
      SLsmg_gotorc(18, COL2);
      SLsmg_printf("%11.6f", sh->manualTarget.el);

      SLsmg_gotorc(19, COL0);
      SLsmg_printf("%c", (vPanel.whatRunMode() == CC_manual_velocity) ? '>' : ' ');
      SLsmg_gotorc(19, COL1);
      SLsmg_printf("%11.6f", sh->manualSpeed.az);
      SLsmg_gotorc(19, COL2);
      SLsmg_printf("%11.6f", sh->manualSpeed.el);

      SLsmg_gotorc(20, COL0);
      SLsmg_printf("%c",
        ((vPanel.whatRunMode() == CC_manual_azeltighten_pwm)
           || (vPanel.whatRunMode() == CC_manual_az_pwm)
           || (vPanel.whatRunMode() == CC_manual_el_pwm)) ? '>' : ' ');
      SLsmg_gotorc(20, COL1-2);
      SLsmg_printf("%c1: %+05d  2: %+05d%c",
        ((vPanel.whatRunMode() == CC_manual_azeltighten_pwm) 
          || (vPanel.whatRunMode() == CC_manual_az_pwm)) ? '>' : ' ',
        ((vPanel.whatRunMode() == CC_manual_azeltighten_pwm) 
          || (vPanel.whatRunMode() == CC_manual_az_pwm)) ? (int)sh->knobMot1pwm : 0,
        (vPanel.whatRunMode() == CC_manual_az_pwm) ? (int)sh->knobMot2pwm : 0,
        (vPanel.whatRunMode() == CC_manual_az_pwm) ? '<' : ' '
      );
      SLsmg_gotorc(20, COL2-2);
      SLsmg_printf("%c1: %+05d  2: %+05d%c",
        ((vPanel.whatRunMode() == CC_manual_azeltighten_pwm) 
          || (vPanel.whatRunMode() == CC_manual_el_pwm)) ? '>' : ' ',
        (vPanel.whatRunMode() == CC_manual_el_pwm) ? (int)sh->knobMot1pwm : 0,
        ((vPanel.whatRunMode() == CC_manual_azeltighten_pwm) 
          || (vPanel.whatRunMode() == CC_manual_el_pwm)) ? (int)sh->knobMot2pwm : 0,
        (vPanel.whatRunMode() == CC_manual_el_pwm) ? '<' : ' '
      );

      SLsmg_gotorc(22, 1);
      SLsmg_printf(
	"Interlock  %06lX, %s",
	sh->interlock,
	explainInterlock(sh->interlock)
      );
      SLsmg_erase_eol();
      SLsmg_gotorc(23, 1);
      SLsmg_printf("Sel = %5u,  PrevLoop = %7d us,  SerErr = %9ld",
        vPanel.getSelectKnob(),
        vDownlinkTime.delta(),
        vSerialReadErrors
      );
#if 0
      SLsmg_gotorc(17, 1);
      SLsmg_printf("A/D tilt-y-az: %9.4f", dl.downlink.tilt.az);
      SLsmg_printf("  tilt-x-el: %9.4f (millidegrees)", dl.downlink.tilt.el);

      SLsmg_printf("    tilt-Temp:  %8.4f", dl.downlink.tiltTemp);
      SLsmg_printf("   air-Temp:  %8.4f C", dl.downlink.extTemp);

#endif

      SLsmg_gotorc(21, 1);
      if (vPanel.whatRunMode() == CC_standby_slowdown) {
	SLsmg_printf("STOPPING...");
      } else if (vPanel.whatRunMode() == CC_servoing_standby) {
	SLsmg_printf("STANDBY");
      } else if (vPanel.whatRunMode() == CC_brake_standby) {
	SLsmg_printf("AUTO-STANDBY");
      } else if (vPanel.whatRunMode() == CC_track_position) {
	SLsmg_printf("TRACKING");
      } else if (vPanel.whatRunMode() == CC_manual_offsets) {
	SLsmg_printf("MANUAL OFFSETS");
      } else if (vPanel.whatRunMode() == CC_manual_position) {
	SLsmg_printf("MANUAL POS");
      } else if (vPanel.whatRunMode() == CC_manual_velocity) {
	SLsmg_printf("MANUAL VEL");
      } else if (vPanel.whatRunMode() == CC_manual_azeltighten_pwm) {
	SLsmg_printf("SVC AZ-EL-TIGHTEN");
      } else if (vPanel.whatRunMode() == CC_manual_az_pwm) {
	SLsmg_printf("SVC AZ MOTORS");
      } else if (vPanel.whatRunMode() == CC_manual_el_pwm) {
	SLsmg_printf("SVC EL MOTORS");
      } else {
	SLsmg_printf("???");
      }
      // To clear interactive az/el prompts.
      SLsmg_erase_eol();
    }  /* of block update display values */

    // Refresh changed values on-screen.
    SLsmg_refresh();
  }  /* for ever */

lEnd:
  /* Cleanups. */
  SLsmg_printf("\n\n\n");

  sl.close();

  vPanel.cleanup();
  vLEDSW.cleanup();
  vLEDIM.cleanup();
  vQDR.cleanup();

  vVAXIF.cleanup();

  SLsmg_refresh();
  SLsmg_reset_smg();
  SLang_reset_tty();
#if 0
  printf("dl = %u, ul = %u\n", sizeof(tDownlink), sizeof(tUplink));
#endif

#if 0
  printf("shKey=0x%08X, shId=%d, sh=%08X\n",
         shKey, shId, (unsigned int)sh);
#endif
  if ((shId != -1) && (sh != (ptSh)-1) && (sh != NULL)) {
    assert( shmctl(shId, IPC_RMID, NULL) == 0 );
    assert( shmdt(sh) == 0 );
  }  // if shared memory was allocated

  return (EXIT_SUCCESS);
}  // main
