// qdrtst.cpp  --  Quadrature Board Test Program.

// 20-May-1999  amn  Original version.


#define BASIC_LAMP_TEST 0


#include <conio.h>
#include <stdio.h>
#include <dos.h>


#include "c8255.cpp"


/* ------------------------------------------------------------------ */
// QDR 4-connector quadrature encoder decoding board,
// connected to one connector on ACL-712 6x3x8bit digital I/O board.
#include "cQDR.cpp"

/* ------------------------------------------------------------------ */
// LEDIM 16-bit off/dim/lit/bright LEDs plus 24 input switches.
// connected to one connector on ACL-712 6x3x8bit digital I/O board.
class cLEDIM : public c8255 {
private:
  unsigned int dimbits;  // latest bit values loaded onto LEDIM board&chip
  unsigned int bribits;
public:
  virtual void init(
	int baseAddr,
	int conn,
    int addressing
  );
  virtual void setLeds(unsigned int leds, int brightness);
  virtual void refreshLeds(void);
  virtual void cleanup(void);
};

void cLEDIM::init(
  int baseAddr,  /* of ET1722B card (or any 8255-compatible parallel) */
  int conn,
  int addressing
) {
  c8255::init(baseAddr, conn, addressing);

  // All ports ABC are outputs.
  this->setPortABCMode(
	/*A=>*/ OUT, /*B=>*/ OUT, /*Cupper=>*/ OUT, /*Clower=>*/ OUT
  );
  this->dimbits = 0;
  this->bribits = 0;
}

void cLEDIM::cleanup() {
  c8255::cleanup();
}

// Set LEDs in mask 'leds' at the same time to desired brightness.
void
cLEDIM::setLeds(unsigned int leds, int brightness) {
  unsigned char v;
  unsigned int t;

  if ((brightness & 0x01)) {  // dim bit on
    this->dimbits |= leds;  // add dim bits
  } else {
    this->dimbits &= ~(leds);  // turn off dim bits
  }
  this->out(/*offset=>*/ 0, (this->dimbits & 0xff));  // LSB LED lines
  this->out(/*offset=>*/ 1, (this->dimbits >> 8));  // MSB LED lines
  this->out(/*offset=>*/ 2, 0x0);  // select dim bits
  this->out(/*offset=>*/ 2, 0x2);  // latch dim bits ("LOADCLK")
  this->out(/*offset=>*/ 2, 0x0);  // back to default

  if ((brightness & 0x02)) {  // bright bit on
    this->bribits |= leds;
  } else {
    this->bribits &= ~(leds);
  }
  this->out(/*offset=>*/ 0, (this->bribits & 0xff));  // LSB LED lines
  this->out(/*offset=>*/ 1, (this->bribits >> 8));  // MSB LED lines
  this->out(/*offset=>*/ 2, 0x1);  // select bright bits
  this->out(/*offset=>*/ 2, 0x3);  // latch bright bits
  delay(1);  // to make latch clock strobe signal long enough to see in LED
  this->out(/*offset=>*/ 2, 0x0);  // back to default

  return;
}  // setLeds

// Set all LEDs again on LEDIM board (and flash "Panel ok" LED):
void
cLEDIM::refreshLeds(void) {
  this->setLeds(/*leds=>*/ 0 /*no change*/, 0);
}  // refreshLeds


/* ------------------------------------------------------------------ */
// LEDSW 16-bit off/dim/lit/bright LEDs plus 24 input switches.
// connected to one connector on ACL-712 6x3x8bit digital I/O board.
class cLEDSW : public c24In {
public:
  virtual unsigned long int get(void);
};

// Get all 24 switch bits.
unsigned long int
cLEDSW::get(void) {
  // We invert the 24 switch bits; afther this 1==switch on.
  return (~(c24In::get()) & 0x00ffffff);
}  // get


/* ------------------------------------------------------------------ */
// The combination of QDR and LEDIM boards together make an ANTBOX panel.
class cPanel {
private:
  // The panel consists of QDR encoders ("three knobs"),
  // a set of 16 LEDs and 24 input switches.
  class cQDR *myQDR;
  class cLEDIM *myLEDIM;
  class cLEDSW *myLEDSW;
  int yesNoSelectActive;
public:
  virtual void init(
    class cQDR *aQDR,
    class cLEDIM *aLEDIM,
    class cLEDSW *aLEDSW
  );
  virtual void cleanup(void);
  virtual void activateYesNoSelect(void);
  virtual void deactivateYesNoSelect(void);
  virtual void run(void);
};  // class cPanel

void
cPanel::init(
  class cQDR *aQDR,
  class cLEDIM *aLEDIM,
  class cLEDSW *aLEDSW
) {
  this->myQDR = aQDR;
  this->myLEDIM = aLEDIM;
  this->myLEDSW = aLEDSW;

  // All LEDs off.
  this->myLEDIM->setLeds(/*leds=>*/ 0xffff, /*brightness=>*/ 0);
  this->yesNoSelectActive = 0;
}  // init

void
cPanel::cleanup(void) {
}  // cleanup

// LEDIM board bit assignments used in ANTBOX panel system.
#define LED_stop           0x0001
#define SW_stop          0x000001
#define LED_standby        0x0002
#define SW_standby       0x000002
#define LED_auto_standby   0x0004
#define LED_slewing        0x0008
#define LED_tracking       0x0010
#define LED_spare1         0x0020
#define LED_az_off         0x0040
#define LED_el_off         0x0080
#define LED_yes_ok         0x0100
#define SW_yes_ok        0x000100
#define LED_no_cancel      0x0200
#define SW_no_cancel     0x000200
#define LED_manual         0x0400
#define SW_manual        0x000400
#define LED_service        0x0800
#define SW_service       0x000800
#define LED_reset_tower_pc 0x1000
#define LED_check          0x2000
#define SW_check         0x002000
#define LED_spare2         0x4000
#define LED_spare3         0x8000
// Port PC definitions of LEDIM.
#define OUT_dim_bright_bit 0x01
#define OUT_load_bit       0x02
// (Six extra +24VDC on/off outputs.)
#define OUT_spare1         0x04
#define OUT_spare2         0x08
#define OUT_spare3         0x10
#define OUT_spare4         0x20
#define OUT_spare5         0x40
#define OUT_spare6         0x80

void
cPanel::activateYesNoSelect(void) {
  // Activate "Yes/No" (and "Select" knob).
#if 0
  this->myLEDIM->setLeds(
    /*leds=>*/ LED_yes_ok | LED_no_cancel,
    /*brightness=>*/ 1
  );
#endif
  this->yesNoSelectActive = 1;
}  // activateYesNoSelect

void
cPanel::deactivateYesNoSelect(void) {
  // Deactivate "Yes/No" (and "Select" knob).
#if 0
  this->myLEDIM->setLeds(
    /*leds=>*/ LED_yes_ok | LED_no_cancel,
    /*brightness=>*/ 0
  );
#endif
  this->yesNoSelectActive = 0;
}  // deactivateYesNoSelect

void
cPanel::run(void) {
  unsigned long int sw = this->myLEDSW->get();

  printf("Select   = %5u,   Az = %5u,   El = %5u\n\n",
    this->myQDR->get(2),
    this->myQDR->get(0),
    this->myQDR->get(1)
  );

#define DEBUG_SWITCHES 1
#if DEBUG_SWITCHES
  printf("Switches = %03X\n", sw);
#endif

  // STOP overrides everything else, turns all other LEDs off
  // and lits only STOP brightly.
  if (sw & SW_stop) {
#if DEBUG_SWITCHES
    printf("Stopped \n");
#endif
    this->deactivateYesNoSelect();
    this->myLEDIM->setLeds(/*leds=>*/ 0xffff, /*brightness=>*/ 0);
    this->myLEDIM->setLeds(/*leds=>*/ LED_stop, /*brightness=>*/ 3);

  } else {
    // Any other switch pressed but not STOP, turn STOP LED off.
    // Hmm, not any other, but standby, manual, or service.
    if (sw & (SW_standby | SW_manual | SW_service)) {
      this->myLEDIM->setLeds(/*leds=>*/ LED_stop, /*brightness=>*/ 0);
    }

    // STANDBY returns back to "normal" where the panel is functional.
    if (sw & SW_standby) {
#if DEBUG_SWITCHES
      printf("Normal  \n");
#endif
      // Stationary LEDs in "normal" mode.
      this->myLEDIM->setLeds(
        /*leds=>*/ LED_manual | LED_service | LED_reset_tower_pc,
        /*brightness=>*/ 1
      );
      this->deactivateYesNoSelect();
      // State LEDs.
      this->myLEDIM->setLeds(/*leds=>*/ LED_standby, /*brightness=>*/ 3);

    // MANUAL.
    } else if (sw & SW_manual) {
#if DEBUG_SWITCHES
      printf("Manual  \n");
#endif
      this->activateYesNoSelect();
      // State LEDs.
      this->myLEDIM->setLeds(/*leds=>*/ LED_manual, /*brightness=>*/ 3);
      this->myLEDIM->setLeds(/*leds=>*/ LED_service, /*brightness=>*/ 1);
      this->myLEDIM->setLeds(/*leds=>*/ LED_standby, /*brightness=>*/ 3);

    // SERVICE.
    } else if (sw & SW_service) {
#if DEBUG_SWITCHES
      printf("Service \n");
#endif
      this->activateYesNoSelect();
      // State LEDs.
      this->myLEDIM->setLeds(/*leds=>*/ LED_service, /*brightness=>*/ 3);
      this->myLEDIM->setLeds(/*leds=>*/ LED_manual, /*brightness=>*/ 1);
      this->myLEDIM->setLeds(/*leds=>*/ LED_standby, /*brightness=>*/ 3);

    // Yes/no.
    // If nothing is pressed, still blink "Panel OK" LED.
    } else {
      this->myLEDIM->refreshLeds();
    }
  }  // end else not STOP

    // Check Yes/no.
    if (this->yesNoSelectActive) {
      if (sw & SW_yes_ok) {
        this->myLEDIM->setLeds(
          /*leds=>*/ LED_yes_ok,
          /*brightness=>*/ 3
        );
      } else {  // active, but nothing pressed
        this->myLEDIM->setLeds(
          /*leds=>*/ LED_yes_ok,
          /*brightness=>*/ 1
        );
      }
      if (sw & SW_no_cancel) {
        this->myLEDIM->setLeds(
          /*leds=>*/ LED_no_cancel,
          /*brightness=>*/ 3
        );
      } else {  // active, but nothing pressed
        this->myLEDIM->setLeds(
          /*leds=>*/ LED_no_cancel,
          /*brightness=>*/ 1
        );
      }
    } else {  // not active
      this->myLEDIM->setLeds(
        /*leds=>*/ (LED_yes_ok | LED_no_cancel),
        /*brightness=>*/ 0
      );
    }

}  // run


/* ------------------------------------------------------------------ */
int
main(int argc, char *argv[]) {
  class cLEDSW vLEDSW;
  class cLEDIM vLEDIM;
  class cQDR vQDR;
  class cPanel vPanel;

  vQDR.init(
	/*baseAddr=>*/ 0x220,
	/*conn=>*/ 0,
    /*addressing=>*/ 0  // normal 4 separate registers mode
  );
  vLEDIM.init(
	/*baseAddr=>*/ 0x220,
	/*conn=>*/ 1,
    /*addressing=>*/ 0  // normal 4 separate registers mode
  );
  vLEDSW.init(
	/*baseAddr=>*/ 0x220,
	/*conn=>*/ 2,
    /*addressing=>*/ 0  // normal 4 separate registers mode
  );
  vPanel.init(&vQDR, &vLEDIM, &vLEDSW);

  clrscr();
#if BASIC_LAMP_TEST
  for (;;) {
    gotoxy(1, 1);
    printf("Off...\n");
    vLEDIM.setLeds(0xffff, 0);
    delay(500);
    printf("Dim...\n");
    vLEDIM.setLeds(0xffff, 1);
    delay(500);
    printf("Lit...\n");
    vLEDIM.setLeds(0xffff, 2);
    delay(500);
    printf("Bright...\n");
    vLEDIM.setLeds(0xffff, 3);
    delay(500);
  }  // for ever
#endif

  gotoxy(1, 1);

  for (;;) {
    gotoxy(1, 1);
    vPanel.run();
  }  // for ever

  vPanel.cleanup();
  vLEDSW.cleanup();
  vLEDIM.cleanup();
  vQDR.cleanup();

  return (0);
}  // main
