/* m100m.cpp  --  100MHz MAC211 time difference counter test program.  */

/* When         Who  What */
/* 05-Feb-1997  amn  Original version. */

#include <conio.h>
#include <stdio.h>
#include <dos.h>

int getVal(char *prompt) {
  char buff[255];
  int temp;

  printf("%s", prompt);
  gets(buff);
  (void)sscanf(buff, "%x", &temp);
  return (temp);
}


/* ------------------------------------------------------------------ */
class cTctr {
private:
  int baseAddr;
// baseAddr + 0 write: data out for transmit to MACH211 chip

// baseAddr + 1 when read from == LPT status port,
//                                reading back data from chip

// (baseAddr +2 == control port, extra output bits, incl. strobe)

  // The value currently in LPT interface data output register;
  // allows changing individual bits.
  unsigned char currentDataOutByte;
  // Same for control port.
  unsigned char currentControlByte;

protected:
  virtual void setDataOut(unsigned char byte);
  virtual void setDataOutBitTo(unsigned char bit, int one);
  virtual int getStatusByte(void);
  virtual void setControl(unsigned char byte);
  virtual void setControlBitTo(unsigned char bit, int one);

public:
  virtual void init(int baseAddr);
  // virtual void setPolarities(...);
  virtual void cleanup(void);
  virtual void trig(void);
  virtual int isBusy(void);
  virtual long int get(void);  // value as integer counts
};  /* cTctr */

// Sets LPT port data output register.
void cTctr::setDataOut(unsigned char byte) {
  outportb(this->baseAddr, byte);
  this->currentDataOutByte = byte;
}

// Sets a given LPT port data output register bit to 0/1.
void cTctr::setDataOutBitTo(unsigned char bit, int one) {
  this->setDataOut(
    ((one) ?
      (this->currentDataOutByte | bit)
      : (this->currentDataOutByte & (~bit))
    )
  );
}

// Gets the LPT port status byte (5 MSBs really from the connector).
int cTctr::getStatusByte(void) {
  return (inportb(this->baseAddr + 1));
}

// Sets LPT port control output register.
void cTctr::setControl(unsigned char byte) {
  outportb(this->baseAddr + 2, byte);
  this->currentControlByte = byte;
}

// Sets a given LPT port control output register bit to 0/1.
void cTctr::setControlBitTo(unsigned char bit, int one) {
  this->setControl(
    ((one) ?
      (this->currentControlByte | bit)
      : (this->currentControlByte & (~bit))
    )
  );
}

void cTctr::init(int baseAddr) {
  this->baseAddr = baseAddr;

  // Init output data lines.
  this->setDataOut(0xff);
  this->setControl(0xff);
  // xxx: Strobe to zero.
}

void cTctr::cleanup(void) {
  // Nothing yet.
}

void cTctr::trig(void) {
  // Data out = 7 and Strobe...
  this->setDataOut(0xff);
  this->setControl(0x00);
  this->setControl(0xff);
}

int cTctr::isBusy(void) {
  // MSB on?  Chip is busy counting:
  return ( (this->getStatusByte() & 0x80) == 0x00 );  // xxx
}

long int cTctr::get(void) {
  int i;
  long int temp = 0;

  // Wait until one start--stop has been counted.
#if 1
  while (this->isBusy()) ;
#endif

  // Get counter value nibble at a time, MSN first.
  for (i=6; i>=0; i--) {
    // Set chip data output mux.
    this->setDataOut(i);
    // Get one nibble.
    temp <<= 4;
    temp |= ((this->getStatusByte() & 0x78) >> 3);
  }  /* for each nibble */

  return temp;
}


/* ------------------------------------------------------------------ */
int main(int argc, char **argv) {
  class cTctr vTctr;

  vTctr.init(/*baseAddr=>*/ 0x378);

  clrscr();
  printf("Press 't' to trig, 'q' to stop.\n\n");
  for(;;) {
    // Screen output.
#if 1
    gotoxy(2, 4);
#endif
    printf("Ctr = %9ld\n", vTctr.get());

    // "User interface."
 	if (kbhit()) {
	  switch (getch()) {
		case 't': {
          vTctr.trig();
          break;
        }
		case 'q':
		case 'Q':
		  goto lEnd;
      }  /* switch */
    }  /* if key */
  }  /* forever */
lEnd:
  vTctr.cleanup();
  return (0);
}  /* main */