#define POSITION_DISPLAY 1

/* ET-1728 D/A 2-channel D/A converter, 12-bit each. */
#define DA_INCLUDED 1
#define DA_CH1 0x2c0
#define DA_CH2 0x2c2

typedef int Boolean;
#define FALSE 0
#define TRUE 1

#define inpw inport

#include <conio.h>
#include <stdio.h>
#include <dos.h>
#include <time.h>
#include <math.h>  /* for pow() */


#define int_enable	0xF3		/* interrupt 10 & 11 enable */
#define int_disable	0x0C		/* interrupt 10 & 11 disable */

#define INTCON1		0x21		/* Address of primary 8259  */
#define ICW1		0x20		/* Address of primary 8259 ICW1. */
#define INTCON2		0xA1		/* Address of secondary 8259 . */
#define ICW2		0xA0		/* Address of secondary 8259 ICW1. */

#define INT10		0x72		/* Interrupt 10 number */
#define INT11		0x73		/* Interrupt 11 number */
#define EOI10		0x62		/* Specific EOI for int10 */
#define EOI11		0x63		/* Specific EOI for int11 */
#define NS_EOI		0x20		/* Non-specific EOI */


#define	IOBASE		0x300		/* Card I/O base address */

#define	TIMER0		IOBASE		/* Timer 0 base address */
#define	COUNTER00	TIMER0		/* Timer 0 counter 0    */
#define	COUNTER01	TIMER0 + 1	/* Timer 0 counter 1    */
#define	COUNTER02	TIMER0 + 2	/* Timer 0 counter 2    */
#define	CONTROL0		TIMER0 + 3	/* Timer 0 control      */

#define	TIMER1		IOBASE + 4	/* Timer 1 base address */
#define	COUNTER10	TIMER1		/* Timer 1 counter 0    */
#define	COUNTER11	TIMER1 + 1	/* Timer 1 counter 1    */
#define	COUNTER12	TIMER1 + 2	/* Timer 1 counter 2    */
#define	CONTROL1		TIMER1 + 3	/* Timer 1 control      */

#define	TIMER2		IOBASE + 8	/* Timer 2 base address */
#define	COUNTER20	TIMER2		/* Timer 2 counter 0    */
#define	COUNTER21	TIMER2 + 1	/* Timer 2 counter 1    */
#define	COUNTER22	TIMER2 + 2	/* Timer 2 counter 2    */
#define	CONTROL2		TIMER2 + 3	/* Timer 2 control      */


#define	TSOURCE		IOBASE + 16	/* 1 = internal 5M osc. */
#define	SPARE1		IOBASE + 18	/* spare out bit	*/
#define	AZMENA		IOBASE + 20	/* Azimuth enable	*/
#define	ELEVENA		IOBASE + 22	/* Elevation enable	*/
#define AZDIR1		IOBASE + 24	/* Az motor 1 direction */
#define AZDIR2		IOBASE + 26	/* Az motor 2 direction */
#define ELDIR1		IOBASE + 28	/* El motor 1 direction */
#define ELDIR2		IOBASE + 30	/* El motor 2 direction */

#define UP              1               /* Motor dir UP         */
#define DOWN            0               /* Motor dir DOWN       */

#define	AZLOW		IOBASE + 16	/* Az low 8 bits	*/
#define	AZHIGH		IOBASE + 18	/* Az high 16 bits	*/
#define	ELLOW		IOBASE + 20	/* El low 8 bits	*/
#define	ELHIGH		IOBASE + 22	/* El high 16 bits	*/

#define VAXIN		IOBASE + 24	/* VAX control port	*/
#define STATUS		IOBASE + 26	/* status port		*/
#define HEARTBEAT	IOBASE + 28	/* Heartbeat port	*/
#define HUNT		IOBASE + 30	/* Hunt for 1 PPS	*/



#define HEARTBEAT_LEN	10000		/* = 200 ms */
#define	BASE_FREQ	250		/* 10 kHz PWM base freq	*/
#define IDLE_SPEED	2               /* 800 ns idle pulse	*/
#define SLOW_SPEED	1		/* in PWM 1..BASE_FREQ scale */
  /* ??? '1' is PTF :-) for GAL bug, actually 20-30 might be reasonable */

/*	Variables					*/

int	currentMode = 0x3;		/* Default mode */
int	prescaler = 100;			/* Prescaler for timer int*/
int	timerInterval = 5000;		/* Timer = 100 ms */
int	timer_ms = 100;			/* Time in milliseconds */
int	sampleInterval = 100;		/* Sample interval 20 us */


/* Command from VAX							*/

unsigned int elevSetLow;                 /* Elevation low 8 bits         */
unsigned int elevSetHigh;                /* Elevation high 16 bits       */
long elevSet;  /* Elevation 21 (+sign) bits */

unsigned int azimSetLow;                 /* Azimuth low 8 bits           */
unsigned int azimSetHigh;                /* Azimuth high 16 bits         */
long azimSet;  /* Azimuth 22 (+sign) bits */




void (interrupt far * intsave)();
void interrupt far timerInt();

void (interrupt far * vaxintsave)();
void interrupt far vaxInt();

void
synchronize(void)
{
outp(HUNT,0);		/* wait for 1 pps pulse */

			/* Start real-time clock */

outp(CONTROL0,0xB4);	/* Counter 02, mode 2, LSB + MSB write */
outp(COUNTER01, timerInterval & 0xFF);
outp(COUNTER01, (timerInterval >>8) & 0xFF);
}



void
initServo(void)
{
outp(TSOURCE,0x1);		/* Internal clock		*/

				/* TIMER 0 CH 0 = HEARTBEAT	*/
outp(CONTROL0,0x32);	/* Counter 00, mode 1, LSB + MSB write */
outp(COUNTER00, HEARTBEAT_LEN & 0xFF);
outp(COUNTER00, (HEARTBEAT_LEN >>8) & 0xFF);

				/* TIMER 0 CH 1&2 = REALTIME CLK */
				/* This prescaler provides clock for both */
				/*  - rt clock (timer 1 ctr 2) */
				/*  - heartbeat (timer 1 ctr 0) */
outp(CONTROL0,0x74);	/* Counter 01, mode 2, LSB + MSB write */
outp(COUNTER01, prescaler & 0xFF);
outp(COUNTER01, (prescaler >>8) & 0xFF);

outp(CONTROL0,0xB4);	/* Counter 02, mode 2, LSB + MSB write */
outp(COUNTER02, timerInterval & 0xFF);
outp(COUNTER02, (timerInterval >>8) & 0xFF);


				/* TIMER 1 = AZIMUTH CONTROL	*/

				/* PWM base frequency */
outp(CONTROL1,0x34);	/* Counter 10, mode 2, LSB + MSB write */
outp(COUNTER10, BASE_FREQ & 0xFF);
outp(COUNTER10, (BASE_FREQ >>8) & 0xFF);


outp(CONTROL1,0x72);	/* Counter 11, mode 1, LSB + MSB write */
outp(COUNTER11, IDLE_SPEED & 0xFF);
outp(COUNTER11, (IDLE_SPEED >>8) & 0xFF);

outp(CONTROL1,0xB2);	/* Counter 12, mode 1, LSB + MSB write */
outp(COUNTER12, IDLE_SPEED & 0xFF);
outp(COUNTER12, (IDLE_SPEED >>8) & 0xFF);




				/* TIMER 2 = ELEVATION CONTROL	*/

				/* PWM (hardware enabled) "slow speed". */
outp(CONTROL2,0x32);	/* Counter 00, mode 1, LSB + MSB write */
outp(COUNTER20, SLOW_SPEED & 0xFF);
outp(COUNTER20, (SLOW_SPEED >>8) & 0xFF);

outp(CONTROL2,0x72);	/* Counter 11, mode 1, LSB + MSB write */
outp(COUNTER21, IDLE_SPEED & 0xFF);
outp(COUNTER21, (IDLE_SPEED >>8) & 0xFF);

outp(CONTROL2,0xB2);	/* Counter 12, mode 1, LSB + MSB write */
outp(COUNTER22, IDLE_SPEED & 0xFF);
outp(COUNTER22, (IDLE_SPEED >>8) & 0xFF);



outp(AZDIR1,DOWN);			/* Motor directions		*/
outp(AZDIR2,UP);			/* 				*/
outp(ELDIR1,DOWN);			/* 				*/
outp(ELDIR2,UP);			/* 				*/


outp(AZMENA,0);				/* Disable azimuth enable	*/
outp(ELEVENA,0);			/* Disable elevation enable	*/
}


unsigned int
pwmValueTo12bitDA(int pwm, int dir, int enable) {
  /* Converts from 0..245 to 0..2047 (negative, 0==-10V) */
  /* or 2048..4095 (positive, 4095==+10V). */
  return (
	(((long int)pwm * 2047) / (BASE_FREQ-5) * (dir == UP ? 1 : -1) * enable) + 2048
  );
}


/* Variables shared between setPulse/DA routines, interrupt routines, */
/* and main(). */
int vAzEnabled;
int vAzDirection;

long elevBitsSincePreviousTimerInterrupt;
long azimBitsSincePreviousTimerInterrupt;


void
setPulseAz1(unsigned int speed)  	/* Pulse ratio 1... BASE_FREQ	*/
{
unsigned int xxx;

outp(CONTROL1,0x72);	/* Counter 11, mode 1, LSB + MSB write */
outp(COUNTER11, speed & 0xFF);
outp(COUNTER11, (speed >>8) & 0xFF);
#if DA_INCLUDED
xxx = pwmValueTo12bitDA(speed, vAzDirection, vAzEnabled);
outportb(DA_CH1, (unsigned char)((xxx >> 8) & 0x0F));  /* MSB (12-bit, upper nibble zero) */
outportb(DA_CH1+1, (unsigned char)(xxx & 0xFF));  /* LSB; it latches new value to D/A chip */
#endif
}

void
setPulseAz2(unsigned int speed)  	/* Pulse ratio 1... BASE_FREQ	*/
{
unsigned int xxx;

outp(CONTROL1,0xB2);	/* Counter 12, mode 1, LSB + MSB write */
outp(COUNTER12, speed & 0xFF);
outp(COUNTER12, (speed >>8) & 0xFF);
#if DA_INCLUDED
xxx = pwmValueTo12bitDA(speed, vAzDirection, vAzEnabled);
#if 0
printf("\n  D/A channel 2 = %04X\n\n", xxx);
#endif
outportb(DA_CH2, (unsigned char)((xxx >> 8) & 0x0F));  /* MSB (12-bit, upper nibble zero) */
outportb(DA_CH2+1, (unsigned char)(xxx & 0xFF));  /* LSB; it latches new value to D/A chip */
#endif
}

void
setPulseEl1(unsigned int speed)  	/* Pulse ratio 1... BASE_FREQ	*/
{
outp(CONTROL2,0x72);	/* Counter 11, mode 1, LSB + MSB write */
outp(COUNTER21, speed & 0xFF);
outp(COUNTER21, (speed >>8) & 0xFF);
}

void
setPulseEl2(unsigned int speed)  	/* Pulse ratio 1... BASE_FREQ	*/
{
outp(CONTROL2,0xB2);	/* Counter 12, mode 1, LSB + MSB write */
outp(COUNTER22, speed & 0xFF);
outp(COUNTER22, (speed >>8) & 0xFF);
}

heartbeat()
{
inpw(HEARTBEAT);
}


void getPosition(
  long *elevPos,
  long *azimPos
) {
  unsigned int high;
  unsigned int low;
  Boolean isPositive;
	/* We always have 21 significant unsigned bits. */
    /* The 22. bit is not realized in hardware in Essco box for elevation, */
    /* thus it must be masked out.  Elevation 21, azimuth 22 unsigned bits. */
    /* Sign is carried in a separate 23. bit (1=+). */

  low = inpw(ELLOW);
  high = inpw(ELHIGH);
  isPositive = (high & 0x8000);
  high &= 0x3fff;
  *elevPos = (((long)high) << 7) | (low >> 9);
  *elevPos = (isPositive) ? (*elevPos) : (-(*elevPos));

  low = inpw(AZLOW);
  high = inpw(AZHIGH);
  isPositive = (high & 0x8000);
  high &= 0x7fff;
  *azimPos = (((long)high) << 7) | (low >> 9);
  *azimPos = (isPositive) ? (*azimPos) : (-(*azimPos));
}  /* getPosition */



long	ltime;
int	milliseconds = 0;
long ms_since_init = 0;

realtimeInstall()

{
	/* Get current local time and start incrementing from there */
	/* with the timer interrupt subroutine. */
	time(&ltime);
	intsave =  getvect(INT11);
	setvect(INT11, timerInt);

	vaxintsave =  getvect(INT10);
	setvect(INT10, vaxInt);

	outp(INTCON2, (inp(INTCON2) & int_enable));
				/* Set up 8259 interrupt controller */

}

disableInterrupt()

{
	outp(INTCON2, (inp(INTCON2) | int_disable));
				/* Set up 8259 interrupt controller */

}


realtimeRestore()

{
	outp(INTCON2, (inp(INTCON2) | int_disable));
				/* Set up 8259 interrupt controller */
	setvect(INT10, vaxintsave);
	setvect(INT11, intsave);

}


void interrupt far timerInt()
{
  static long previousElev = 0;
  static long previousAzim = 0;
  static long thisElev;
  static long thisAzim;

  ms_since_init += timer_ms;

  milliseconds += timer_ms;
  if (milliseconds >= 1000) {	/* Full second gone by ???*/
    milliseconds -= 1000;
    ltime++;			/* increment GMT time 1 second */
  }

  /* Now the speed measurement. */
  getPosition(&thisElev, &thisAzim);
  elevBitsSincePreviousTimerInterrupt = thisElev - previousElev;
  azimBitsSincePreviousTimerInterrupt = thisAzim - previousAzim;
  previousElev = thisElev;
  previousAzim = thisAzim;

  /* Security check: speed must not exceed about 1 degree / second. */
#define MAX_CHANGE_IN_BITS_PER_INTERRUPT 100
  if ( (elevBitsSincePreviousTimerInterrupt > MAX_CHANGE_IN_BITS_PER_INTERRUPT)
       || (azimBitsSincePreviousTimerInterrupt > MAX_CHANGE_IN_BITS_PER_INTERRUPT) ) {
    /* vAzEnabled = 0; ??? Sorry, random reading prevent this. */
    outp(AZMENA, vAzEnabled);
  }

  outp(ICW1,NS_EOI);			/* Send End-of-Interrupt to 8259. */
  outp(ICW2,EOI11);			/* Send End-of-Interrupt to 8259. */
}  /* timerInt */


void interrupt far vaxInt()
{
  static unsigned int vaxdata,vaxtype;
  static unsigned int elevationLow;	/* Elevation low 11 bits	*/
  static unsigned int elevationHigh;	/* Elevation high 11 bits	*/
  static unsigned int azimuthLow;		/* Azimuth low 11 bits		*/
  static unsigned int azimuthHigh;	/* Azimuth high 11 bits		*/

  vaxdata = inpw(VAXIN);
  vaxtype = (vaxdata & 0xf000) >> 12;

  if ( vaxtype == 0) elevationLow = vaxdata;
  if ( vaxtype == 1) {
    elevationHigh = vaxdata;
    elevSet = (((long)elevationHigh & 0x03ff) << 11) | (elevationLow & 0x07ff);
  }
  if ( vaxtype == 4) azimuthLow = vaxdata;
  if ( vaxtype == 5) {
    azimuthHigh = vaxdata;
    azimSet = (((long)azimuthHigh & 0x07ff) << 11) | (azimuthLow & 0x07ff);
  }
  outp(ICW1,NS_EOI);			/* Send End-of-Interrupt to 8259. */
  outp(ICW2,EOI10);			/* Send End-of-Interrupt to 8259. */
}  /* vaxInt */


void printBinary( unsigned int tmp)

{
int i;

for(i=0;i<16;i++) {
    if(tmp & 0x8000)
        printf("1");
    else
        printf("0");
	tmp *= 2;
    }
}

double
toDegrees(
  long actual23BitValue
) {
  return ((double)actual23BitValue
    * (360.0 / 4194304.0 /* == (pow(2.0, 22.0))*/)
  );
}  /* toDegrees */


void
turnAzDirection(
  int vDir
) {
  outp(AZMENA, 0);
  vAzDirection = vDir;

  /* ??? At first a motor pair is forced to turn in opposite directions. */
  outp(AZDIR1, vAzDirection);
  outp(AZDIR2, (vAzDirection==UP ? DOWN : UP));

  outp(AZMENA, vAzEnabled);
}


main()
{
  struct tm *owntime;

  long elevPos;  /* Elevation 21 (+sign) bits */
  long azimPos;  /* Azimuth 22 (+sign) bits */

  int vAzPulse[3];  /* Too lazy to use 0..n-1; using 1..n and leaving 0 unused. */
  int vWhichAzPulse;

  Boolean vStressTestOn;
  long vSTNextChangeTime;
  int vSTPhase;
  typedef enum {eST1up, eST1down, eST2up, eST2down} tSTPhase;
#define cST1Max 235
#define cST1Min 41
#define cST2Max 235
#define cST2Min 41
#define cSTInc 10
#define cSTDec 15
#define cSTInterval 100  /* milliseconds */

  realtimeInstall();

  initServo();


  vAzDirection = DOWN;
  vAzPulse[1] = vAzPulse[2] = 1;
  vWhichAzPulse = 1;
  vAzEnabled = 0;
  vStressTestOn = FALSE;
  vSTNextChangeTime = ms_since_init;
  vSTPhase = eST1up;
  clrscr();
  for(;;) {

	if (kbhit()) {
	  switch (getch()) {
	    case '1':
	      vWhichAzPulse = 1;
		  break;
	    case '2':
	      vWhichAzPulse = 2;
	      break;
	    case '+':
	      if (vAzPulse[vWhichAzPulse] <
		BASE_FREQ - 5  /* leave time for FET control condensator */
	      ) {
		vAzPulse[vWhichAzPulse]++;
	      }
	      break;
	    case '-':
	      if (vAzPulse[vWhichAzPulse] > 1) vAzPulse[vWhichAzPulse]--;
		  break;
	    case 'p':
	    case 'P':
	      vAzEnabled = (vAzEnabled ? 0 : 1);
	      outp(AZMENA, vAzEnabled);
	      break;
	    case 'u':
	    case 'U':
	      turnAzDirection(UP);
	      break;
	    case 'd':
	    case 'D':
	      turnAzDirection(DOWN);
		  break;
	    case 't':
	    case 'T':
	      vStressTestOn = (vStressTestOn ? FALSE : TRUE);
	      break;
	    case 'q':
	    case 'Q':
	      goto lExit;
	      break;
	    default:
	      break;
	  }
	}
	gotoxy(1,1);
	printf( " p)      Az enabled   %d\n", vAzEnabled);
	printf( " u,d)    Az direction %d  (%s)  \n",
	  vAzDirection,
	  (vAzDirection == UP ? "up" : "down")
	);
	printf("%c1:+,-)  Az1 pulse    %03d\n",
	  (vWhichAzPulse == 1 ? '>' : ' '),
	  vAzPulse[1]
	);
	printf("%c2:+,-)  Az2 pulse    %03d\n",
	  (vWhichAzPulse == 2 ? '>' : ' '),
	  vAzPulse[2]
	);
	printf( " t)      Stress test  %d (%d)\n", vStressTestOn, vSTPhase);

	/* Stress test. */
	if (vStressTestOn) {
	  if (ms_since_init > vSTNextChangeTime) {
	    switch (vSTPhase) {
	      case eST1up: {
		if ((vAzPulse[1] += cSTInc) >= cST1Max) {
		  vSTPhase = eST1down;
		}
		break;
	      }
		  case eST1down: {
		if ((vAzPulse[1] -= cSTDec) <= cST1Min) {
		  vSTPhase = eST2up;
		  vWhichAzPulse = 2;
		}
		break;
	      }
	      case eST2up: {
		if ((vAzPulse[2] += cSTInc) >= cST2Max) {
		  vSTPhase = eST2down;
		}
		break;
	      }
		  case eST2down: {
		if ((vAzPulse[2] -= cSTDec) <= cST2Min) {
		  vSTPhase = eST1up;
		  vWhichAzPulse = 1;
		}
		break;
	      }
	      default: {
		vSTPhase = eST1up;
		break;
	      }
	    }

		vSTNextChangeTime = ms_since_init + cSTInterval;
	  }
	}

	/* Security checks: */
	{
	  int a1 = vAzPulse[1];
	  int a2 = vAzPulse[2];

	  /* In range 1..(BASE_FREQ-5): */
	  a1 = (a1 < 1) ? 1 : a1;
	  vAzPulse[1] = a1 = (a1 > BASE_FREQ-5) ? BASE_FREQ-5 : a1;
	  a2 = (a2 < 1) ? 1 : a2;
	  vAzPulse[2] = a2 = (a2 > BASE_FREQ-5) ? BASE_FREQ-5 : a2;

#define MAX_OTHER_DRIVE 100
	  if ( (a1>a2 ? a1 : a2) > MAX_OTHER_DRIVE /* 4,5 A */) {
		/* Ease the "other" motor to maximum "safe" drive. */
		int ap;

		ap = vAzPulse[(vWhichAzPulse==1 ? 2 : 1)];
		vAzPulse[(vWhichAzPulse==1 ? 2 : 1)] = (ap > MAX_OTHER_DRIVE ? MAX_OTHER_DRIVE : ap);
	  }
	}

	setPulseAz1(vAzPulse[1]);
	setPulseAz2(vAzPulse[2]);

#if POSITION_DISPLAY

	getPosition(&elevPos, &azimPos);

	gotoxy(1,8);
#if 0
	printf(" VAX azimuth       %X %X \n", azimuthHigh, azimuthLow);
	printf(" VAX elevation     %X %X \n", elevationHigh, elevationLow);

        printBinary(azimPosHigh);
        printBinary(azimPosLow & 0xff00);
        printf("\n");

        printBinary(elevPosHigh);
        printBinary(elevPosLow & 0xff00);
	printf("\n");
#endif

	printf(" Antenna azimuth   %08lX \n", azimPos);
	printf(" Antenna elevation %08lX \n", elevPos);
	printf(" VAX azimuth   %08lX \n", azimSet);
	printf(" VAX elevation %08lX \n\n", elevSet);

	printf(" Azimuth   (deg) %4.5lf, speed %lf deg/s\n",
	  toDegrees(azimPos),
	  toDegrees(azimBitsSincePreviousTimerInterrupt*10)  /* ??? 0.1s */
	);
	printf(" Elevation (deg) %4.5lf, speed %lf deg/s\n",
	  toDegrees(elevPos),
	  toDegrees(elevBitsSincePreviousTimerInterrupt*10)  /* ??? 0.1s */
	);
	printf(" Vax azimuth   (deg) %4.5lf\n",
	  toDegrees(azimSet)
	);
	printf(" Vax elevation (deg) %4.5lf\n\n",
	  toDegrees(elevSet)
	);

	owntime = gmtime(&ltime);
	printf(" GMT is %s\n", asctime(owntime));
#endif


	heartbeat();			/* keep motors on		*/

    }

lExit:
  realtimeRestore();

}
