#include "registration.h"
RCSID(cqdr_cpp, "$Id: cqdr.cpp,v 1.4 1999/11/20 19:48:27 amn Exp $");
/* ------------------------------------------------------------------ */
// QDR 4-connector quadrature encoder decoding board,
// connected to one connector on ACL-712 6x3x8bit digital I/O board.

// $Log: cqdr.cpp,v $
// Revision 1.4  1999/11/20 19:48:27  amn
// QDR 'reset()', tTime 'getNow()', and CC_brake_standby uplink command.
//
// Revision 1.3  1999/10/12 08:32:46  amn
// Corrected registration identifier.
//
// Revision 1.2  1999/10/12 07:52:52  amn
// Added registration and log.
//

class cQDR : public c8255 {
public:
  virtual void init(
	int baseAddr,
	int conn,
    int addressing
  );
  virtual unsigned int get(int conn);
  virtual void reset(int conn);
  virtual void cleanup(void);
};

void cQDR::init(
  int baseAddr,  /* of ET1722B card (or any 8255-compatible parallel) */
  int conn,
  int addressing
) {
  c8255::init(baseAddr, conn, addressing);

  // Port A is 8-bit input data bus,
  // port B is used to toggle control lines to HCTL-2016 chips,
  // port C is in practice unused.
  this->setPortABCMode(
	/*A=>*/ IN, /*B=>*/ OUT, /*Cupper=>*/ OUT, /*Clower=>*/ OUT
  );
}

void cQDR::cleanup() {
  c8255::cleanup();
}

unsigned int
cQDR::get(int conn) {
  unsigned char v;
  unsigned int t;

  // Get high byte.
  v = (conn << 1) & 0xFE;
  this->out(/*offset=>*/ 1, 0x08 | v | 0);  // \oe to zero, sel == 0
  t = this->in(/*offset=>*/ 0);

  // Get low byte.
  this->out(1, 0x08 | v | 1);  // \oe still zero, sel == 1
  t = (t << 8) | this->in(0);

  // HCTL 8-bit data bus back to tri-state.
  this->out(1, 0);  // \oe back to 1
  //this->out(1, 0x2f);  // auto-index-pulse-reset-mode
  return (t);
}  // get

void
cQDR::reset(int conn) {
  unsigned char v;

  v = (conn << 1) & 0xFE;
  // Reset to 1, \oe to zero, sel == 0
  this->out(/*offset=>*/ 1, 0x18 | v | 0);
  // HCTL 8-bit data bus back to tri-state.
  this->out(1, 0);  // \oe back to 1
  //this->out(1, 0x2f);  // auto-index-pulse-reset-mode
}  // get
