#include "registration.h"
RCSID(ctime_cpp, "$Id: ctime.cpp,v 1.5 1999/11/20 19:48:27 amn Exp $");
/* ------------------------------------------------------------------ */
// Time handling.

// $Log: ctime.cpp,v $
// Revision 1.5  1999/11/20 19:48:27  amn
// QDR 'reset()', tTime 'getNow()', and CC_brake_standby uplink command.
//
// Revision 1.4  1999/11/06 22:33:00  amn
// Added getting 'tTime' delta in 'double' fractional seconds.
//
// Revision 1.3  1999/11/01 17:57:40  amn
// New xntpd time and interlock type.
//
// Revision 1.2  1999/10/28 18:14:11  amn
// Debugged cROC4xx repeats (7 not enough...).
// Added ignoring values >10% off from last 4 in cSensorChannel.
// Added 'tTime' alias struct timeval (2x longs) into ctime.cpp.
//
// Revision 1.1  1999/10/28 08:29:40  amn
// Separated cTime class to a separate file.
//

#if LINUX_TIME_AVAIL
#include <sys/time.h>
typedef struct timeval tTime;
// xxx: Calculations on tTime rely on subfields being signed integers.
#else
typedef struct timeval {
  long int tv_sec;
  long int tv_usec;
} tTime;
#endif

class cTime {
private:
  tTime tv;  // the current time
  int firstNowCall;  // ==1 first call since init; delta forced to zero
  tTime prevtv;  // previous time for delta calculation
  long deltaUsecs;  // between current and previous
public:
  // Constructor initializes delta calculation.
  cTime::cTime();
#if 0
  // Gives direct access to 'struct timeval'.
  virtual tTime * operator&(void);
#endif
  // Getting a new current time.
  virtual void now(void);
  // Getting the latest delta in usecs between two calls of 'now()'.
  virtual long delta(void);
};  // cTime

cTime::cTime() {
  // First call of 'now()' will reinit delta calculation variables.
  firstNowCall = 1;
}  // cTime constructor

#if 0
tTime *
cTime::operator&(void) {
  return( &(this->tv) );
}  // operator&
#endif

void
addTime(
  const tTime& a,
  const tTime& b,
  tTime& c  // c=a+b
) {
  long u;

  c.tv_sec = a.tv_sec + b.tv_sec;  // add seconds...
  if ((u = (a.tv_usec + b.tv_usec)) >= 1000000) {  // if usec overflow...
    u -= 1000000;
    c.tv_sec++;  // ...normalize by incrementing full seconds
  }
  c.tv_usec = u;
}  // addTime

void
subTime(
  const tTime& a,
  const tTime& b,
  tTime& c  // c=a-b
) {
  long u;

  c.tv_sec = a.tv_sec - b.tv_sec;  // subtract seconds
  if ((u = (a.tv_usec - b.tv_usec)) < 0) {  // if usec underflow...
    u += 1000000;
    c.tv_sec--;  // ...normalize by decrementing full seconds
  }
  c.tv_usec = u;
}  // subTime

// Convert (delta) timeval into microseconds (if within +/- 2000 secs).
long
usecs(const tTime& d) {
  // 32-bit long cannot represent more than +/-2000x1000000 microseconds.
  assert( (d.tv_sec >= -2000 ) && (d.tv_sec <= 2000) );
  return (1000000*d.tv_sec + d.tv_usec);
}  // usecs

// Convert (delta) timeval into (double floating) seconds.
double
dsecs(const tTime& d) {
  // 64-bit double rounds off LS microseconds if seconds difference is large.
  return ((double)d.tv_sec + (double)d.tv_usec/1000000.0);
}  // usecs

// Get current time.
void
getNow(tTime& d) {
#if LINUX_TIME_AVAIL
  assert( gettimeofday(&d, NULL) == 0 );
#endif
}  // getNow

void
cTime::now(void) {
  tTime d;

  getNow(d);
  if (this->firstNowCall) {
    // The very first 'now()' call has no delta when compared to the previous.
    this->prevtv = this->tv = d;
    this->deltaUsecs = 0;
    this->firstNowCall = 0;
  } else {
    // All subsequent 'now()'s have delta.
    this->prevtv = this->tv;
    this->tv = d;
    subTime(this->tv, /* - */ this->prevtv, /* = */ d);
#if 0
    // Test for Linux time being monotonic:
    assert( (d.tv_sec == 0) && (d.tv_usec > 0) );
#endif
    // 32-bit long cannot represent more than +/-2000x1000000 microseconds.
    assert( (d.tv_sec >= -2000 ) && (d.tv_sec <= 2000) );
    this->deltaUsecs = usecs(d);
  }
}  // now

long
cTime::delta(void) {
  return( (this->firstNowCall) ? 0 : this->deltaUsecs );
}  // delta
