/* -------------------------------------------------------------------- */
/* The wire-wrapped prototype antenna board for VAX commanded az-el input. */
typedef void interrupt far ( *ptIRQHandler)(...);
class cProtoAnt {
private:
  unsigned int baseAddr;
  int picControlRegister;  /* 8259 PIC chip reg. */
  int vaxIRQAddr;  /* 8086 interrupt vector offset 0--255. */
  ptIRQHandler oldIRQHandler;
  unsigned int azLow;  /* azimuth low 11 bits		*/
  unsigned int azHigh;  /* azimuth high 11 bits		*/
  long int azSet;  /* combined azimuth 22 bits */
  unsigned int elLow;  /* elevation low 11 bits	*/
  unsigned int elHigh;  /* elevation high 10 bits	*/
  long int elSet;  /* combined elevation 21 bits */
public:
  virtual void init(int baseAddr, int vaxIRQ, ptIRQHandler handler);
  virtual void cleanup(void);
  /* virtual??? */ void handleVaxIRQ(void);
  virtual double getAz(void);
  virtual double getEl(void);
};  /* cProtoAnt */

void cProtoAnt::init(int baseAddr, int vaxIRQ, ptIRQHandler handler) {
  this->baseAddr = baseAddr;
  if (vaxIRQ > 7) {
    this->vaxIRQAddr = 0x70 + (vaxIRQ - 8);
    this->picControlRegister = 0xA1;  /* 2nd 8259 chip */
  } else {
    /* xxx: what are vector numbers of 1st int.contr.? */ this->vaxIRQAddr = 0x72;
    this->picControlRegister = 0x21;  /* 2nd 8259 chip */
  }

  /* Save the old interrupt vector. */
   this->oldIRQHandler = getvect(vaxIRQAddr);

  /* Install the new interrupt handler and enable this interrupt. */
   setvect(vaxIRQAddr, handler);
   outportb(
     this->picControlRegister,
     (inportb(this->picControlRegister) & /* xxx: */ 0xFB)
   );

  this->azSet = this->elSet = 0;
}

void cProtoAnt::cleanup(void) {
  /* Reset the old interrupt handler; xxx: disable this interrupt. */
  outportb(
    this->picControlRegister,
    (inportb(this->picControlRegister) | /* xxx: */ 0x04)
  );
  setvect(this->vaxIRQAddr, this->oldIRQHandler);
}

void cProtoAnt::handleVaxIRQ(void) {
  unsigned int vaxdata;
  unsigned char vaxtype;

  vaxdata = inport(this->baseAddr + 24);
  vaxtype = (vaxdata & 0xf000) >> 12;

  if ( vaxtype == 0) elLow = vaxdata;
  if ( vaxtype == 1) {
    elHigh = vaxdata;
    elSet = (((long)elHigh & 0x03ff) << 11) | (elLow & 0x07ff);
      /* xxx: el always assumed positive; seems to wrap from 0.0 to 360.0 */
  }
  if ( vaxtype == 4) azLow = vaxdata;
  if ( vaxtype == 5) {
    azHigh = vaxdata;
    azSet = (((long)azHigh & 0x07ff) << 11) | (azLow & 0x07ff);
    /* Extend MSB == 1  => negative in 2-complement. */
    if (azSet & 0x00200000) {
      /* MSB sign bit == 1, extend. */
      azSet |= 0xFFE00000;
    }
  }
#if 1
  /* I believe these are not needed with Borland C++ 3.1 libraries. */
  /* xxx: Needs testing again, once freezed the whole machine. */
#define ICW1		0x20		/* Address of primary 8259 ICW1. */
#define ICW2		0xA0		/* Address of secondary 8259 ICW1. */
#define EOI10		0x62		/* Specific EOI for int10 */
#define NS_EOI		0x20		/* Non-specific EOI */
  outp(ICW1,NS_EOI);			/* Send End-of-Interrupt to 8259. */
  outp(ICW2,EOI10);			/* Send End-of-Interrupt to 8259. */
#endif
}

double cProtoAnt::getAz(void) {
  return ((double)this->azSet * 360.0 / 2097152.0);
    /* xxx: interrupt protection when retrieving long value from class struct? */
}

double cProtoAnt::getEl(void) {
  return ((double)this->elSet * 360.0 / 2097152.0);
    /* xxx: int protection ? */
}

class cProtoAnt vProtoAnt;

/* static */ void interrupt far doVaxIRQ(...) {
  vProtoAnt.handleVaxIRQ();
}


