// downf/ant.cpp  --  New PWMISA card plus QDR tacho encoders.

// $Log: ant.cpp,v $
// Revision 1.19  1999/10/11 12:39:58  amn
// Hide cursor.
//
// Revision 1.18  1999/10/05 14:09:29  amn
// Tested 10kHz PWM base frequency.
//
// Revision 1.17  1999/10/05 12:24:44  amn
// Added using 'cQDR.cpp' include.
//
// Revision 1.16  1999/10/04 10:13:03  amn
// Adding tiltmeter.
//
// Revision 1.15  1999/09/30 17:03:02  amn
// Removed cQDR code to a separate cQDR.cpp file.
//
// Revision 1.14  1999/09/30 10:26:43  amn
// Added brake I measurements.
//
// Revision 1.13  1999/09/29 11:15:23  amn
// Added absolute Az/El, tacho/QDR, and interlock in hexadecimal.
//
// Revision 1.12  1999/09/29 07:12:33  amn
// Harmonized c24In and c24Out.
//
// Revision 1.11  1999/09/28 12:04:13  amn
// Changed the location of c8255.cpp and its charset to DOS.
//
// Revision 1.10  1999/09/28 07:28:07  amn
// Just before starting changes for ant/up with the final pedestal PC
// in tower; changing to CR-LF DOS convention for direct CVS use.
//
// Revision 1.9  1999/08/20 10:57:07  amn
// New PWMISA code replaced old wire-wrap PWM board code.
//
// Revision 1.8  1999/05/11 11:36:17  amn
// Added more A/D channels, 's' for servoing.
//
// Revision 1.7  1999/03/18 10:21:26  amn
// Added missing CVS log entries to this file.
//
// Revision 1.6  1999/03/18 09:28:41  amn
// The first super-simple servo with 0.070 degree error/slowness...
//
// Revision 1.5  1999/03/16 08:51:33  amn
// Version used to measure antiphase FET16r board prototypes.
//
// Revision 1.4  1999/02/05 07:53:28  amn
// Changed "stress test" into scanning all motor drive value pairs (Az1, Az2).
//
// Revision 1.3  1999/02/02 08:07:33  amn
// Added speed and acceleration values to display and log.
//
// Revision 1.2  1999/01/28 13:14:14  amn
// Changed automatic RCS version printout from Revision to Id.
//
// Revision 1.1  1999/01/28 13:11:32  amn
// First version with antenna position logging to ant.log.
//
// Revision 1.0  1999/01/28 08:38:22  amn
// First version which duplicates old 'ant.c' FET board tester with cPWMISA.
//
// Revision 0.5  1999/01/27 13:55:23  amn
// First release which compiled with C++ and class cPWMISA.
// Problems with setting PWM timers/counters (stay at "slow" speed).
//
// 27-Jan-1999  amn  Original version with class cPWMISA.


#define AD_CURRENT_MEASUREMENTS 1
#define CATCH_OVER_10AMPS 0

#define REAL_SERIAL_CONN 0
#define REAL_VAXLINK 0
#define TIME_DISPLAY 1

// Log time and value change resolution.
#define MINLOGCHANGE 0.00005  // degrees
#define MINRECALCSPACC 200  // ms

#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>  // for fabs()
#include <assert.h>
#include <dos.h>
#include <time.h>


/*------------------------------------------------------------*/
#include "cPWMISA.cpp"

// Setting up 1pps IRQ of PWMISA card "manually".

class cPWMISA vPWMISA;

#define ICW1		0x20		/* Address of primary 8259 ICW. */
#define INTCON2		0xA1		/* Address of secondary 8259 . */
#define ICW2		0xA0		/* Address of secondary 8259 ICW. */
#define INT11		0x73		/* Interrupt 11 number */
#define EOI11		0x63		/* Specific EOI for int11 */
#define NS_EOI		0x20		/* Non-specific EOI */


#ifdef __cplusplus
    #define __CPPARGS ...
    #define __MUSTBEFAR
#else
    #define __CPPARGS
    #define __MUSTBEFAR far
#endif

void interrupt __MUSTBEFAR (*intsave)(__CPPARGS);
void interrupt __MUSTBEFAR timerInt(__CPPARGS);
time_t ltime;


void realtimeInstall(void) {
	/* Get current local time and start incrementing from there */
	/* with the timer interrupt subroutine. */
    tzset();
	time(&ltime);
	intsave =  getvect(INT11);
	setvect(INT11, timerInt);

  // IRQ11 is not enabled in 8259 interrupt controller chip by default:
  // (Need to put bit 3 (==3+8=11) to low.)
  outportb(INTCON2, (inportb(INTCON2) & ~(0x08)) );
}  // reatimeInstall

void realtimeRestore(void) {
  // Put IRQ11 disable bit back to 1 in controller chip.
  outportb(INTCON2, (inportb(INTCON2) | 0x08) );
  setvect(INT11, intsave);
}  // realtimeRestore


unsigned long microseconds = 0;
unsigned long prevUsecs = 0;
unsigned long ms_since_init = 0;
unsigned long s_since_init = 0;

void interrupt __MUSTBEFAR timerInt(__CPPARGS)
{
  microseconds = vPWMISA.getUsecs();  // get from board
  ms_since_init;  // calculate running milliseconds
  ltime++;			/* increment GMT time 1 second */
  s_since_init;  // ?

  outportb(ICW1,NS_EOI);			/* Send End-of-Interrupt to 8259. */
  outportb(ICW2,EOI11);			/* Send End-of-Interrupt to 8259. */
}  /* timerInt */


/*------------------------------------------------------------*/
// ACL-8113 32-ch 12-bit A/D board, surprisingly similar to
// ET2816 Capax 16-ch 16-bit A/D board.
#include "cSensor.cpp"
#define DEBUG_REPEATS 1
#define NO_ET28xx_PRESENT 0
#include "cET28xx.cpp"

/*------------------------------------------------------------*/
// ACL-712 6x3x8bit digital I/O board, surprisingly similar to
// ET7122 Capax 6x3x8bit DI/O board.  Emulates 6 8255 Intel PIO chips.
#include "c8255.cpp"

/*------------------------------------------------------------*/
// Heidenhain ROC412 12-bit parallel output absolute encoders (Az/El),
// each connected to one connector on ACL-712 6x3x8bit digital I/O board.
#include "cROC4xx.cpp"

/* ------------------------------------------------------------------ */
// QDR 4-connector quadrature encoder decoding board,
// connected to one connector on ACL-712 6x3x8bit digital I/O board.
#include "cQDR.cpp"

/*------------------------------------------------------------*/
// Explain interlock chain hexadecimals.
char *
explainInterlock(long int bits) {
  // Interlock chain which is common to both Az and El.
  if (!(bits & 0x000800))
    return ("main panel 'STOP' switch");
  if (!(bits & 0x000400))
    return ("emergency stop switch");
  if (!(bits & 0x000200))
    return ("downstairs pedestal manual interlock switch");
  if (!(bits & 0x000100))
    return ("Bronto switch");
  if (!(bits & 0x000080))
    return ("upstairs pedestal hoist switch");
  if (!(bits & 0x000040))
    return ("receiver interlock switch");
  if (!(bits & 0x000020))
    return ("upstairs pedestal manual interlock switch");
  if (!(bits & 0x000010))
    return ("(reserved option switch, not yet used)");
  if (!(bits & 0x000008))
  // End of common interlock chain.

  // When common chain is true, the following apply to Az.
    return ("Az1 motor interlock switch");
  if (!(bits & 0x000004))
    return ("CW final limit switch");
  if (!(bits & 0x000002))
    return ("CCW final limit switch");
  if (!(bits & 0x000001))
    return ("'AZ OFF' panel switch");
  if (!(bits & 0x001000))

  // When common chain is true, the following apply to El.
    return ("upstairs pedestal stow pin switch");
  if (!(bits & 0x002000))
    return ("El1 motor interlock switch");
  if (!(bits & 0x004000))
    return ("UP final limit switch");
  if (!(bits & 0x008000))
    return ("DOWN final limit switch");
  if (!(bits & 0x010000))
    return ("'EL OFF' panel switch");

  // "Interlock ok" messages.
  if ((bits & 0x01ffef) == 0x01ffef) {
    if ((bits & 0xf00000) == 0xf00000)
      return ("all interlock switches ok");
    if (!(bits & 0x800000))
      return ("CW slow limit reached, interlock ok");
    if (!(bits & 0x400000))
      return ("CCW slow limit reached, interlock ok");
    if (!(bits & 0x200000))
      return ("UP slow limit reached, interlock ok");
    if (!(bits & 0x100000))
      return ("DOWN slow limit reached, interlock ok");
  }  // if interlock ok but slow limits on

  // Az, El, or both limit switches overridden manually from panel.
  if ((bits & 0x01ffe9) == 0x01ffe9)
    return ("ok with Az limits overridden");
  if ((bits & 0x013fef) == 0x013fef)
    return ("ok with El limits overridden");
  if ((bits & 0x01ffef) == 0x01ffef)
    return ("ok with both Az and El limit switches overridden");

  return ("???");
}  // explainInterlock

/*------------------------------------------------------------*/
// Utilities.
double
toDegrees(
  long actual23BitValue
) {
  return ((double)actual23BitValue
    * (360.0 / 4194304.0 /* == (pow(2.0, 22.0))*/)
  );
}  /* toDegrees */

int getVal(char *prompt) {
  char buff[255];
  int temp;

  printf("%s", prompt);
  gets(buff);
  (void)sscanf(buff, "%d", &temp);
  return (temp);
}

double getDouble(char *prompt) {
  char buff[255];
  double temp;

  printf("%s", prompt);
  gets(buff);
  (void)sscanf(buff, "%lg", &temp);
  return (temp);
}


/*------------------------------------------------------------*/
/* Definition of the struct of all sensor values on the serial line down. */
#include "downlink.h"

typedef union {
  tDownlink downlink;
  unsigned char buff[sizeof(tDownlink)];
} tDLBuff;

#include "serial.h"

/* Reads one character; if timeout, returns 0xff. */
unsigned char getOne(int port) {
  unsigned int i = 50000;
  int c;

  while (i && ((c = ReadSerial(port)) == -1))
    i--;
  return ((unsigned char)c);
}

tDLBuff dl;
unsigned long vSerialReadErrors = 0L;

void getDownlink(void) {
    do {
      int i;

      /* The other end is not supposed to send any responses
         before receiving a command, make sure by: */
      FlushInBuf(COM2);

      /* Ask for all sensor values by sending a command character. */
      WriteSerial(COM2, '*');

      /* Try to read exactly the expected number of reply bytes. */
      for (i=0; i<sizeof(tDownlink); i++) {
        dl.buff[i] = getOne(COM2);
      }
        /* Too few characters: trailing characters will be 0xff;
           extra characters will be flushed before the next read. */

    /* Message must have correct length at end. */
      if (dl.downlink.length != sizeof(dl)) {
        vSerialReadErrors++;
      }
    } while (dl.downlink.length != sizeof(dl));
}  // getDownlink

typedef struct sVaxlink {
  double az;
  double el;
  int length;
} tVaxlink;

typedef union uVLbuff {
  tVaxlink vaxlink;
  unsigned char buff[sizeof(tVaxlink)];
} tVLBuff;

tVLBuff vl;

void getVaxlink(void) {
  tVLBuff savedVl;

  savedVl = vl;  // "backup copy"
  do {
      int i;

      /* The other end is not supposed to send any responses
         before receiving a command, make sure by: */
      FlushInBuf(COM1);

      /* Start comms by sending a command character plus az/el. */
      WriteSerial(COM1, '*');
      WriteStrSerial(COM1, savedVl.buff, sizeof(savedVl));

      /* Try to read exactly the expected number of reply bytes. */
      for (i=0; i<sizeof(tVaxlink); i++) {
        vl.buff[i] = getOne(COM1);
      }
        /* Too few characters: trailing characters will be 0xff;
           extra characters will be flushed before the next read. */

    /* Message must have correct length at end. */
      if (vl.vaxlink.length != sizeof(vl)) {
        vSerialReadErrors++;
      }
  } while (vl.vaxlink.length != sizeof(vl));
}  // getVaxlink


/*------------------------------------------------------------*/
int main()
{
  class cPortROC412 currAbsAz;
  class cPortROC412 currAbsEl;
  class cQDR vTachos;
  class c24In vInterlockChain;
  class cET28xx vACL8113;
  class cSensorChannel vAz1ICurrCh, vAz2ICurrCh, vEl1ICurrCh, vEl2ICurrCh;
  class cSensorChannel vTRackUpperCh,
        vTRackLowerCh,
        vTPCInsideCh,
        vAz1TboardCh,
        vAz2TBoardCh,
        vEl1TBoardCh,
        vEl2TBoardCh,
        vSupply5VCh,
        vSupply12VCh,
        vSupply24VCh,
        vAz1SupplyCh,
        vAz2SupplyCh,
        vEl1SupplyCh,
        vEl2SupplyCh,
        vSupplyN15VCh,
        vSupply15VCh,
        vSupplyGNDCh,
        vIBrakeAzCh,
        vIBrakeElCh;
  class cSensorChannel vTiltYAzCh, vTiltXElCh, vTiltTempCh;


  FILE *vLogFile;

#define AZAXIS 0
#define ELAXIS 1
  int vWhichAxis;  // AZ or EL, active with '>' cursor
  int vAzPulse[3];  /* Too lazy to use 0..n-1; using 1..n and leaving 0 unused. */
  int vWhichAzMotor;
  int vElPulse[3];  /* Too lazy to use 0..n-1; using 1..n and leaving 0 unused. */
  int vWhichElMotor;

  // Antenna position, speed, and acceleration.
  double currAz, currEl;
  double goalAz, goalEl;
  unsigned long currAzElTime = 0;
  double prevAz, prevEl;
  unsigned long prevAzElTime = 0;
  double currAzSp = 0.0, currElSp = 0.0;
  unsigned long currAzElSpTime = 0;
  double prevAzSp, prevElSp;
  unsigned long prevAzElSpTime = 0;
  double currAzAcc = 0.0, currElAcc = 0.0;

  // Motor I currents.
  double vAz1ICurr;
  double vAz2ICurr;
  double vEl1ICurr;
  double vEl2ICurr;
  double vTRackUpper,
        vTRackLower,
        vTPCInside,
        vAz1Tboard,
        vAz2TBoard,
        vEl1TBoard,
        vEl2TBoard,
        vSupply5V,
        vSupply12V,
        vSupply24V,
        vAz1Supply,
        vAz2Supply,
        vEl1Supply,
        vEl2Supply,
        vSupplyN15V,
        vSupply15V,
        vSupplyGND,
        vIBrakeAz,
        vIBrakeEl;
  double vTiltYAz, vTiltXEl, vTiltTemp;

  int vStressTestOn;
  int vServoingIsOn;
  unsigned long vSTNextChangeTime;
#define cSTInterval 2000  /* milliseconds */
#define cSTIncrement 5
/*** #define MAX_TEST MAX_PWM ***/
#define MIN_TEST 075
#define MAX_TEST 175

  // Init one board at default base address.
  vPWMISA.init();
  realtimeInstall();
  vPWMISA.setBasePWMFreq(PWMBASE10KHZ);

  // Init ROC412 absolute encoders via parallel interface board.
  currAbsAz.init(
	/*baseAddr=>*/ 0x220,
	/*conn=>*/ 0,
	/*numOfBitsMask=>*/ 0x0800
  );
  currAbsAz.setCalibration(/*offset=>*/ 540.2062, /*coeff=>*/ -3.000867);
    /* Measured on 04-Sep-1996 with ant. range -360..0. */

  currAbsEl.init(
	/*baseAddr=>*/ 0x220,
	/*conn=>*/ 1,
	/*numOfBitsMask=>*/ 0x0800
  );
  currAbsEl.setCalibration(/*offset=>*/ +185.24, /*coeff=>*/ -1.0);
    /* xxx: not measured yet */

  // Init QDR tachometer interface board 24-bit connector.
  vTachos.init(
    /*baseAddr=>*/ 0x220,
    /*conn=>*/ 2,
    /*addressing=>*/ 0  // normal 4 regs per 8255, not indirect
  );

  // Init interlock chain opto board 24-bit connector.
  vInterlockChain.init(
    /*baseAddr=>*/ 0x220,
    /*conn=>*/ 3,
    /*addressing=>*/ 0  // normal 4 regs per 8255, not indirect
  );

  // Init subreflector motor selector switcher relay board 'antrel8'.
  // xxx
  // /*conn=>*/ 4

  // /*conn=>*/ 5  not used yet.


  // Init 32-ch 12-bit A/D converter.
  vACL8113.init(0x260, /*boardModel=>*/ ACL8113, /*numOfBitsMask=>*/ 0x0fff,
                /*loopLimit=>*/ 70
               );
  vACL8113.setCalibration(/*offset=>*/ 0.0, /*coeff=>*/ 1.0);
    /* calibration of A/D board itself */

  // Init individual A/D channels.
  vAz1ICurrCh.init(&vACL8113);
  vAz1ICurrCh.setCalibration(
    /*offset=>*/ 0.0,
    /*coeff=>*/ 2.5
  );
  vAz2ICurrCh.init(&vACL8113);
  vAz2ICurrCh.setCalibration(
    /*offset=>*/ 0.0,
    /*coeff=>*/ 2.5
  );
  vEl1ICurrCh.init(&vACL8113);
  vEl1ICurrCh.setCalibration(
    /*offset=>*/ 0.0,
    /*coeff=>*/ 2.5
  );
  vEl2ICurrCh.init(&vACL8113);
  vEl2ICurrCh.setCalibration(
    /*offset=>*/ 0.0,
    /*coeff=>*/ 2.5
  );
#define INITCH(ch, off, coeff) \
  ch##Ch.init(&vACL8113); \
  ch##Ch.setCalibration( \
    /*offset=>*/ off, \
    /*coeff=>*/ coeff \
  )
  INITCH(vTRackUpper, 0.0, 100.0);
  INITCH(vTRackLower, 0.0, 100.0);
  INITCH(vTPCInside, 0.0, 100.0);
  INITCH(vAz1Tboard, 0.0, 100.0);
  INITCH(vAz2TBoard, 0.0, 100.0);
  INITCH(vEl1TBoard, 0.0, 100.0);
  INITCH(vEl2TBoard, 0.0, 100.0);
  INITCH(vSupply5V, 0.0, 1.0);
  INITCH(vSupply12V, 0.0, 2.0);
  INITCH(vSupply24V, 0.0, 4.0);
  INITCH(vAz1Supply, 0.0, 4.33);  // (33.3k+10k)/(10k)
  INITCH(vAz2Supply, 0.0, 4.33);
  INITCH(vEl1Supply, 0.0, 4.33);
  INITCH(vEl2Supply, 0.0, 4.33);
  INITCH(vSupplyN15V, 0.0, 2.0);
  INITCH(vSupply15V, 0.0, 2.0);
  INITCH(vSupplyGND, 0.0, 1.0);
  INITCH(vIBrakeAz, 0.0, 1.0);
  INITCH(vIBrakeEl, 0.0, 1.0);

  /* Tiltmeter channels. */
  INITCH(vTiltYAz, /*offset=>*/ 0.0,
    /*coeff=>*/ 0.4833 /* low gain; high=0.05002 uradians / mV */
                 * 180.0 / M_PI  /* millidegrees / V */
                 / 1000.0  /* millidegrees -> degrees */
  );
  INITCH(vTiltXEl, /*offset=>*/ 0.0,
    /*coeff=>*/ 0.4740 /* low gain; high=0.04991 */
                 * 180.0 / M_PI  /* millidegrees / V */
                 / 1000.0  /* millidegrees -> degrees */
  );
  INITCH(vTiltTemp, /*offset=>*/ 0.0, /*coeff=>*/ 100.0);

  /* Serial uplink, VAX-link. */
  OpenSerial(COM2, 115200L, D8S1N, NULL);  // uplink
  OpenSerial(COM1, 115200L, D8S1N, NULL);  // vaxlink

#if REMOTE_POSITION_DISPLAY
#if REAL_SERIAL_CONN
    getDownlink();
#else
    dl.downlink.incr.az = 123.456789;
    dl.downlink.incr.el = 45.45454545;
#endif

    currAz = dl.downlink.incr.az;
    currEl = dl.downlink.incr.el;
    currAzElTime = ms_since_init;
    prevAz = currAz;
    prevEl = currEl;
    prevAzElTime = currAzElTime;
#endif

  /* Open log file always in append mode. */
  assert ( (vLogFile = fopen("ant.log", "at")) != NULL );  // append in text mode
  fprintf(vLogFile,
    "\n\n----------\nGMT is %sMilliseconds since program start %010lu\n---\n",
    asctime(gmtime(&ltime)),
    ms_since_init
  );

#define MIN_PWM_OFF 0
  vWhichAxis = AZAXIS;
  vAzPulse[1] = vAzPulse[2] = MIN_PWM_OFF;
  vWhichAzMotor = 1;
  vElPulse[1] = vElPulse[2] = MIN_PWM_OFF;
  vWhichElMotor = 1;
  vStressTestOn = 0;
  vServoingIsOn = 0;
  vSTNextChangeTime = ms_since_init;
  clrscr();
  _setcursortype(_NOCURSOR);
  printf("$Id: ant.cpp,v 1.19 1999/10/11 12:39:58 amn Exp $");
  for(;;) {
    int recalced;

	if (kbhit()) {
	  switch (getch()) {
	    case '0': {
          vPWMISA.resetStatus(/*motor=>*/0);
          vPWMISA.resetStatus(/*motor=>*/1);
          vPWMISA.resetStatus(/*motor=>*/2);
          vPWMISA.resetStatus(/*motor=>*/3);
          vPWMISA.resetStatus(/*motor=>*/4);
		  break;
        }
	    case '1':
          vWhichAxis = AZAXIS;
	      vWhichAzMotor = 1;
		  break;
	    case '2':
          vWhichAxis = AZAXIS;
	      vWhichAzMotor = 2;
	      break;
	    case '3':
          vWhichAxis = ELAXIS;
	      vWhichElMotor = 1;
		  break;
	    case '4':
          vWhichAxis = ELAXIS;
	      vWhichElMotor = 2;
	      break;

#define PLUSMINUSINCREMENT 5
	    case '+':
          switch (vWhichAxis) {
            case AZAXIS: {
              vAzPulse[vWhichAzMotor] += PLUSMINUSINCREMENT;
 	          if (vAzPulse[vWhichAzMotor] > MAX_PWM) {
                vAzPulse[vWhichAzMotor] = MAX_PWM;
	          }
              break;
            }
            case ELAXIS: {
              vElPulse[vWhichElMotor] += PLUSMINUSINCREMENT;
 	          if (vElPulse[vWhichElMotor] > MAX_PWM) {
                vElPulse[vWhichElMotor] = MAX_PWM;
	          }
              break;
            }
          }  // switch axis
	      break;
	    case '-':
          switch (vWhichAxis) {
            case AZAXIS: {
              vAzPulse[vWhichAzMotor] -= PLUSMINUSINCREMENT;
 	          if (vAzPulse[vWhichAzMotor] < -MAX_PWM) {
                vAzPulse[vWhichAzMotor] = -MAX_PWM;
	          }
              break;
            }
            case ELAXIS: {
              vElPulse[vWhichElMotor] -= PLUSMINUSINCREMENT;
 	          if (vElPulse[vWhichElMotor] < -MAX_PWM) {
                vElPulse[vWhichElMotor] = -MAX_PWM;
	          }
              break;
            }
          }  // switch axis
 	      break;
	    case 'n':
          switch (vWhichAxis) {
            case AZAXIS: {
              gotoxy(30, 3+vWhichAzMotor);
	          vAzPulse[vWhichAzMotor] = getVal(" -> ");
              break;
            }
            case ELAXIS: {
              gotoxy(70, 3+vWhichElMotor);
	          vElPulse[vWhichElMotor] = getVal(" -> ");
              break;
            }
          }  // switch axis
		  break;
	    case 'p':
	    case 'P': {
          int i;

          for (i=0; i<NUMOFMOTORS; i++) {
  	        vPWMISA.setEnable(i, TOGGLE);
  	        vPWMISA.setRelay(i, TOGGLE);
          }
	      break;
        }
	    case 's':
	    case 'S':
	      vServoingIsOn = (vServoingIsOn ? 0 : 1);
	      break;
	    case 't':
	    case 'T':
          vAzPulse[1] = MIN_TEST;
          vAzPulse[2] = MIN_TEST;
   		  vSTNextChangeTime = ms_since_init + cSTInterval;
	      vStressTestOn = (vStressTestOn ? 0 : 1);
	      break;
	    case 'q':
	    case 'Q':
	      goto lExit;
	    default:
	      break;
	  }
	}
	gotoxy(1,3);
	printf( " p)      Az enabled   %d %d                 El enabled   %d %d %d\n",
      vPWMISA.getEnable(0),
      vPWMISA.getEnable(1),
      vPWMISA.getEnable(2),
      vPWMISA.getEnable(3),
      vPWMISA.getEnable(4)
    );
#if 0
	printf( " u,d)    Az direction %d  (%s)  \n",
	  vPWMISA.getAxisDir(AZ),
	  (vPWMISA.getAxisDir(AZ) == UP ? "up" : "down")
	);
#endif
	printf("%c1:+,-)  Az1 pulse    %+05d           %c3) El1 pulse    %+05d\n",
	  (((vWhichAxis == AZAXIS) && (vWhichAzMotor == 1)) ? '>' : ' '),
	  vAzPulse[1],
	  (((vWhichAxis == ELAXIS) && (vWhichElMotor == 1)) ? '>' : ' '),
	  vElPulse[1]
	);
	printf("%c2:+,-)  Az2 pulse    %+05d           %c4) El2 pulse    %+05d\n",
	  (((vWhichAxis == AZAXIS) && (vWhichAzMotor == 2)) ? '>' : ' '),
	  vAzPulse[2],
	  (((vWhichAxis == ELAXIS) && (vWhichElMotor == 2)) ? '>' : ' '),
	  vElPulse[2]
	);
	printf( " t)      Stress test  %d                s) Servoing     %d \n",
      vStressTestOn,
      vServoingIsOn
    );

	/* Stress test. */
	if (vStressTestOn) {
	  if (ms_since_init > vSTNextChangeTime) {
        vAzPulse[1] += cSTIncrement;
        if (vAzPulse[1] > MAX_TEST) {
          vAzPulse[1] = MIN_TEST;
          vAzPulse[2] += cSTIncrement;

          // Stop motors for 2 seconds and then continue with new "jump" sp.
          vPWMISA.setSpeed(0 /*Az1*/, 0);
          vPWMISA.setSpeed(1 /*Az2*/, 0);
          delay(2000);

          if (vAzPulse[2] > MAX_TEST) {
            // To restart:
            // vAzPulse[2] = MIN_TEST;
            vAzPulse[1] = 0;
            vAzPulse[2] = 0;
            vStressTestOn = 0;
          }
        }
		vSTNextChangeTime = ms_since_init + cSTInterval;
	  }
	}

	/* Security checks: */
	{
	  int a1 = vAzPulse[1];
	  int a2 = vAzPulse[2];
	  int e1 = vElPulse[1];
	  int e2 = vElPulse[2];

	  /* In range (-MAX_PWM)..(MAX_PWM): */
	  a1 = (a1 < -MAX_PWM) ? -MAX_PWM : a1;
	  vAzPulse[1] = a1 = (a1 > MAX_PWM) ? MAX_PWM : a1;
	  a2 = (a2 < -MAX_PWM) ? -MAX_PWM : a2;
	  vAzPulse[2] = a2 = (a2 > MAX_PWM) ? MAX_PWM : a2;
      e1 = (e1 < -MAX_PWM) ? -MAX_PWM : e1;
	  vElPulse[1] = e1 = (e1 > MAX_PWM) ? MAX_PWM : e1;
	  e2 = (e2 < -MAX_PWM) ? -MAX_PWM : e2;
	  vElPulse[2] = e2 = (e2 > MAX_PWM) ? MAX_PWM : e2;

#if 0
No more after -250..250 range:
#define MAX_OTHER_DRIVE MAX_PWM
	  if ( (a1>a2 ? a1 : a2) > MAX_OTHER_DRIVE /* 4,5 A */) {
		/* Ease the "other" motor to maximum "safe" drive. */
		int ap;

		ap = vAzPulse[(vWhichAzMotor==1 ? 2 : 1)];
		vAzPulse[(vWhichAzMotor==1 ? 2 : 1)] = (ap > MAX_OTHER_DRIVE ? MAX_OTHER_DRIVE : ap);
	  }
#endif
	}  // of security checks

	vPWMISA.setSpeed(0, vAzPulse[1]);
	vPWMISA.setSpeed(1, vAzPulse[2]);
	vPWMISA.setSpeed(2, vElPulse[1]);
	vPWMISA.setSpeed(3, vElPulse[2]);
	vPWMISA.setSpeed(4, /*subreflector=>*/ 0);


#if AD_CURRENT_MEASUREMENTS
    vACL8113.setChannel(/*channel=>*/ 2, SE, /*gain=>*/ 2);
    vAz1ICurr = vAz1ICurrCh.get();
    vACL8113.setChannel(/*channel=>*/ 0, SE, /*gain=>*/ 2);
    vAz2ICurr = vAz2ICurrCh.get();
    vACL8113.setChannel(/*channel=>*/ 6, SE, /*gain=>*/ 2);
    vEl1ICurr = vEl1ICurrCh.get();
    vACL8113.setChannel(/*channel=>*/ 4, SE, /*gain=>*/ 2);
    vEl2ICurr = vEl2ICurrCh.get();
#define GETCH(name, ch, gain) \
    vACL8113.setChannel(/*channel=>*/ ch, SE, /*gain=>*/ gain); \
    name = name##Ch.get()
    GETCH(vTRackUpper, 27, 2);
    GETCH(vTRackLower, 28, 2);
    GETCH(vTPCInside, 29, 2);
    GETCH(vAz1Tboard, 22, 2);
    GETCH(vAz2TBoard, 23, 2);
    GETCH(vEl1TBoard, 24, 2);
    GETCH(vEl2TBoard, 25, 2);
    GETCH(vSupply5V, 16, 1);
    GETCH(vSupply12V, 17, 1);
    GETCH(vSupply24V, 18, 1);
    GETCH(vAz1Supply, 3, 1);
    GETCH(vAz2Supply, 1, 1);
    GETCH(vEl1Supply, 7, 1);
    GETCH(vEl2Supply, 5, 1);
    GETCH(vSupplyN15V, 31, 1);
    GETCH(vSupply15V, 30, 1);
    GETCH(vSupplyGND, 20, 8);
    GETCH(vIBrakeAz, 14, 1);
    GETCH(vIBrakeEl, 15, 1);
    GETCH(vTiltYAz, 11, 1);
    GETCH(vTiltXEl, 10, 1);
    GETCH(vTiltTemp, 12, 1);
#else
    vAz1ICurr = 0.0;
    vAz2ICurr = 0.0;
#endif

#if REAL_SERIAL_CONN
    getDownlink();
#else
    dl.downlink.incr.az = 123.456789;
    dl.downlink.incr.el = 45.45454545;
#endif

    currAz = dl.downlink.incr.az;
    currEl = dl.downlink.incr.el;
    currAzElTime = ms_since_init;

    // Updating VAX link.
    vl.vaxlink.az = currAz;
    vl.vaxlink.el = currEl;
    vl.vaxlink.length = sizeof(vl);
#if REAL_VAXLINK
    getVaxlink();  // send Heidenhein encoders --> VAX and get VAX el/az back
    goalAz = vl.vaxlink.az;
    goalEl = vl.vaxlink.el;
#else
    goalAz = 123.0;
    goalEl = 45.0;
#endif

    // Instant servo.
    if (vServoingIsOn)
    {
      double posErrAz;
      double posErrEl;

      posErrAz = currAz - goalAz;
      if (posErrAz > 1.0) {
        posErrAz = 1.0;
      }
      if (posErrAz < -1.0) {
        posErrAz = -1.0;
      }
      posErrEl = currEl - goalEl;
      if (posErrEl > 1.0) {
        posErrEl = 1.0;
      }
      if (posErrEl < -1.0) {
        posErrEl = -1.0;
      }
      // "Dead" zero speed area of -15..15 of PWM drive.
      // Map position error -0.001..0.001deg to -15..15 of PWM
#define ZEROERR 0.001
#define ZEROPWM 16
#define MAXUSABLEPWM 175.0
#define TIGHTENGEARBOX 30
      if ((posErrAz >= -ZEROERR) && (posErrAz <= ZEROERR)) {
        posErrAz *= (ZEROPWM / ZEROERR);
      } else if (posErrAz < 0.0) {
        posErrAz = (posErrAz + ZEROERR) * MAXUSABLEPWM - ZEROPWM;
      } else /* if (posErrAz > 0.0) */ {
        posErrAz = (posErrAz - ZEROERR) * MAXUSABLEPWM + ZEROPWM;
      }

      if ((posErrEl >= -ZEROERR) && (posErrEl <= ZEROERR)) {
        posErrEl *= (ZEROPWM / ZEROERR);
      } else if (posErrEl < 0.0) {
        posErrEl = (posErrEl + ZEROERR) * MAXUSABLEPWM - ZEROPWM;
      } else /* if (posErrEl > 0.0) */ {
        posErrEl = (posErrEl - ZEROERR) * MAXUSABLEPWM + ZEROPWM;
      }

      // "Motor against motor" drive difference to tighten gears.
      vAzPulse[1] = (int)posErrAz - TIGHTENGEARBOX;
      vAzPulse[2] = (int)posErrAz + TIGHTENGEARBOX;
      vElPulse[1] = (int)posErrEl - TIGHTENGEARBOX;
      vElPulse[2] = (int)posErrEl + TIGHTENGEARBOX;
    }

    {  // Instant antenna location security checks.
       if ( (currAz < -350.0) || (currAz > +350.0)
            /*** || (currEl < 10.0) || (currEl > 95.0) ***/
            || (currEl < 0.2) || (currEl > 106.0)
       ) {
	     vPWMISA.disableAll();
       }
    }  // if servoing is on

  { // Logging all measured values to a file.
    static double loggedAz, loggedEl;

    recalced = ((currAzElTime - prevAzElTime) > MINRECALCSPACC);
    if ( (fabs(currAz - loggedAz) > MINLOGCHANGE)
          || (fabs(currEl - loggedEl) > MINLOGCHANGE)
          || (recalced) ) {

  {
    unsigned long int tdiffMs;  // between two pos measurements in ms.
    double tdiff;  // and in secs.

    if ( (tdiffMs = (currAzElTime - prevAzElTime)) > 0 ) {
      tdiff = (double)tdiffMs / 1000.0;
      prevAzSp = currAzSp;
      prevElSp = currElSp;
      prevAzElSpTime = currAzElSpTime;
      currAzSp = (currAz - prevAz) / tdiff;
      currElSp = (currEl - prevEl) / tdiff;
      currAzElSpTime = currAzElTime;
    }
    if ( (tdiffMs = (currAzElSpTime - prevAzElSpTime)) > 0 ) {
      tdiff = (double)tdiffMs / 1000.0;
      currAzAcc = (currAzSp - prevAzSp) / tdiff;
      currElAcc = (currElSp - prevElSp) / tdiff;
    }
    prevAz = currAz;
    prevEl = currEl;
    prevAzElTime = currAzElTime;
  }

    if ( (fabs(currAz - loggedAz) > MINLOGCHANGE)
          || (fabs(currEl - loggedEl) > MINLOGCHANGE)
          || (recalced) ) {
      fprintf(vLogFile,
        "%010lu\t%05d\t%5.2f\t%05d\t%5.2f\t%11.6f\t%11.6f\t%11.6f\t%11.6f\t%11.6f\t%05d\t%5.2f\t%05d\t%5.2f\t%11.6f\t%11.6f\t%11.6f\t%11.6f\t%11.6f\t%9.4f\t%9.4f\t%8.4f\t\n",
        currAzElTime,
        vPWMISA.getEnable(0) ? vAzPulse[1] : 0,
        vAz1ICurr,
        vPWMISA.getEnable(1) ? vAzPulse[2] : 0,
        vAz2ICurr,
        currAz,
        currAzSp,
        currAzAcc,
        goalAz,
        currAz - goalAz,
        vPWMISA.getEnable(2) ? vElPulse[1] : 0,
        vEl1ICurr,
        vPWMISA.getEnable(3) ? vElPulse[2] : 0,
        vEl2ICurr,
        currEl,
        currElSp,
        currElAcc,
        goalEl,
        currEl - goalEl,
        vTiltYAz,  // dl.downlink.tilt.az,
        vTiltXEl,  // dl.downlink.tilt.el,
        vTiltTemp  // dl.downlink.tiltTemp,
        // dl.downlink.extTemp
      );
      loggedAz = currAz;
      loggedEl = currEl;
    }  // if needs to be logged
    }  // if needs to calc sp/acc or log a line
  }
      printf(
        "  AbsAz %7.2f         AbsEl %7.2f  \n",
        currAbsAz.get(),
        currAbsEl.get()
      );
      printf(
        "     Az %11.6f        El %11.6f                SerErr %3lu\n",
        currAz,
        currEl,
        vSerialReadErrors
      );
      printf(
        " GoalAz %11.6f    GoalEl %11.6f  \n",
        goalAz,
        goalEl
      );
      printf(
        "ErrorAz %11.6f   ErrorEl %11.6f  \n\n",
        goalAz - currAz,
        goalEl - currEl
      );
      printf(
        "SpeedAz %11.6f   SpeedEl %11.6f  (deg/s)\n",
        currAzSp,
        currElSp
      );
      printf(
        "  AccAz %11.6f     AccEl %11.6f  (deg/s/s)    \n",
        currAzAcc,
        currElAcc
      );
      printf(
	"Tilt-Az %11.6f   Tilt-El %11.6f  Tilt-Temp %5.1f\n",
        vTiltYAz,  // dl.downlink.tilt.az,
        vTiltXEl,  // dl.downlink.tilt.el,
        vTiltTemp  // dl.downlink.tiltTemp
        // dl.downlink.extTemp
      );
      printf(
        "      Az1       Az2         El1       El2\n"
      );
      printf(
        "  %5.2f A   %5.2f A     %5.2f A   %5.2f A  T-rack %4.1f/%4.1f/%4.1f C\n",
        vAz1ICurr,
        vAz2ICurr,
        vEl1ICurr,
        vEl2ICurr,
        vTRackUpper,
        vTRackLower,
        vTPCInside
      );
      printf(
        "   %4.1f C    %4.1f C      %4.1f C    %4.1f C  Vsuppl %4.2f/%4.1f/%4.1f V\n",
        vAz1Tboard,
        vAz2TBoard,
        vEl1TBoard,
        vEl2TBoard,
        vSupply5V,
        vSupply12V,
        vSupply24V
      );
      printf(
        "   %4.1f V    %4.1f V      %4.1f V    %4.1f V  Vsupp %5.1f/%4.1f/%7.4f V\n",
        vAz1Supply,
        vAz2Supply,
        vEl1Supply,
        vEl2Supply,
        vSupplyN15V,
        vSupply15V,
        vSupplyGND
      );
      printf(
        "    %5u     %5u       %5u     %5u  Tachos  IBrakeAz %4.1f  El %4.1f A\n",
        vTachos.get(0),
        vTachos.get(1),
        vTachos.get(2),
        vTachos.get(3),
        vIBrakeAz,
        vIBrakeEl
      );
      {
        char vMs0[10];
        char vMs1[10];
        char vMs2[10];
        char vMs3[10];

        vPWMISA.formatStatus(vPWMISA.getStatus(/*motor=>*/ 0), vMs0);
        vPWMISA.formatStatus(vPWMISA.getStatus(/*motor=>*/ 1), vMs1);
        vPWMISA.formatStatus(vPWMISA.getStatus(/*motor=>*/ 2), vMs2);
        vPWMISA.formatStatus(vPWMISA.getStatus(/*motor=>*/ 3), vMs3);
        printf(
          "  %07s   %07s     %07s   %07s  APH STO FLT CHO REL ENA DIR,  0)rst\n\n",
          vMs0,
          vMs1,
          vMs2,
          vMs3
        );
      }
      printf(
        "Interlock   %06lX, %s",
        vInterlockChain.get(),
        explainInterlock( vInterlockChain.get() )
      );
      clreol(); printf("\n");

#if CATCH_OVER_10AMPS
      if ( (fabs(vAz1ICurr) >= 10.0)
            || (fabs(vAz2ICurr) >= 10.0)
            || (fabs(vEl1ICurr) >= 10.0)
            || (fabs(vEl2ICurr) >= 10.0) ) {
        sound(600); delay(500); nosound();
        while (!kbhit()) ;
        (void)getch();
      }
#endif

    // Generate running 'ltime' (seconds) and 'ms_since_init' (msecs).
    prevUsecs = microseconds;
    microseconds = vPWMISA.getUsecs();
    if (microseconds > 999999) {
      // Board is missing.
      microseconds = 0;
    }
    if (microseconds < prevUsecs) {  // overflow, i.e. next second
      ltime++;
      s_since_init++;
    }
    ms_since_init = s_since_init*1000L + microseconds/1000L;

#if TIME_DISPLAY
 	printf("UTC is %s%06lu %010lu  %1d\n",
      asctime(gmtime(&ltime)),
      microseconds,
      ms_since_init,
      vPWMISA.getSecs()
    );
#endif

	vPWMISA.refreshWatchdog();  /* keep motors on */

    }

lExit:
  _setcursortype(_NORMALCURSOR);

  CloseSerial(COM1);
  CloseSerial(COM2);

  // xxx: Should in principle 'cleanup()' lots of '*Ch' sensor channels.

  vACL8113.cleanup();
  vInterlockChain.cleanup();
  vTachos.cleanup();
  currAbsEl.cleanup();
  currAbsAz.cleanup();

  realtimeRestore();
  vPWMISA.cleanup();

  return (0);
}  // main
