#include "registration.h"
RCSID(downlink_h, "$Id: downlink.h,v 1.15 2000/03/13 08:57:47 amn Exp $");
/* ------------------------------------------------------------------ */
/* A binary message containing all sensor and encoder data. */

// $Log: downlink.h,v $
// Revision 1.15  2000/03/13 08:57:47  amn
// Changed 'CC_vax_position' into 'CC_track_position'.
//
// Revision 1.14  1999/11/20 19:48:27  amn
// QDR 'reset()', tTime 'getNow()', and CC_brake_standby uplink command.
//
// Revision 1.13  1999/11/02 11:14:46  amn
// Added manual az-el-pwm with tighten gearbox uplink command.
//
// Revision 1.12  1999/11/01 17:57:40  amn
// New xntpd time and interlock type.
//
// Revision 1.11  1999/11/01 15:06:21  amn
// Added UTC/xntpd time from antcond in uplink to ant/dos/up/ant.cpp.
//
// Revision 1.10  1999/10/22 12:58:53  amn
// Changed command long integer codes to universal "char-by-char" format.
//
// Revision 1.9  1999/10/19 12:42:15  amn
// Added new-style command long words; debugging serial.c TIMEOUT (50000L?).
//
// Revision 1.8  1999/10/15 12:46:44  amn
// Patched ROC412 to be 10 bits only until wiring is corrected.
// Both downlink and uplink in 'downlink.h'.
// 'serial.c' enhanced to support (and require!) receiver FIFOs.
//
// Revision 1.7  1999/10/14 12:48:29  amn
// Added IK121 fault status detection; both down and uplinks.
//
// Revision 1.6  1999/10/12 07:47:36  amn
// Added registration.
//

typedef struct sCoord {
  double az;
  double el;
} tCoord, *ptCoord;

typedef struct sDownlink {
  /* Timestamp, secs + usecs, when incr encoders were read. */
  tTime time;
  tCoord incr;  /* incremental precision encoders */
  tCoord abs;  /* coarse absolute encoders */
  tInterlockBits interlock;  /* copy of interlock switches for panel/antcond use */
  /* crc? */
  long int length;  /* xxx: instead of crc... Must be last to detect lost bytes. */
  /* 'short int' :16 didn't work in GNU g++, wants to pad structs to 32-bits */
} tDownlink, *ptDownlink;

/* Command 32-bit long words. */
/* (Borland C++ 3.1 had trouble with 4-character 'xyzx'-style constants.) */
/* (And it lays out 'xy' two-character constants in Intel little-endian */
/* fashion whereas gcc uses big-endian order. */
#define fourCharLong(a,b,c,d) ( ((long)a)<<24 | ((long)b)<<16 | ((long)c)<<8 | d )
static long int CC_brake_standby =     fourCharLong('S','t','b','y');
static long int CC_track_position =      fourCharLong('T','P','o','s');
static long int CC_manual_position =   fourCharLong('M','P','o','s');
static long int CC_manual_velocity =   fourCharLong('M','V','e','l');
static long int CC_manual_azeltighten_pwm = fourCharLong('M','A','E','T');
static long int CC_manual_az_pwm =     fourCharLong('M','A','z','P');
static long int CC_manual_el_pwm =     fourCharLong('M','E','l','P');

/* A binary message containing panel/antcond control/command messages */
/* to ant/dos/up/ant.cpp. */
typedef struct sUplink {
  /* Timestamp: checked xntpd/seconds + PWMISA/microseconds. */
  tTime time;
  long int command;  /* what antcond requests up/ant to do */
  tCoord cc;  /* command coordinates */
  /* crc? */
  long int length;  /* xxx: instead of crc... Must be last to detect lost bytes. */
  /* 'short int' :16 didn't work in GNU g++, wants to pad structs to 32-bits */
} tUplink, *ptUplink;
