#include "registration.h"
RCSID(interloc_cpp, "$Id: interloc.cpp,v 1.5 1999/11/09 14:25:01 amn Exp $");
/*------------------------------------------------------------*/
// Explain interlock chain hexadecimals using the 24-bit word
// from parallel interface.

// $Log: interloc.cpp,v $
// Revision 1.5  1999/11/09 14:25:01  amn
// Added spare interlock bits.
//
// Revision 1.4  1999/11/01 17:57:40  amn
// New xntpd time and interlock type.
//
// Revision 1.3  1999/10/28 08:27:59  amn
// Added registration and CVS log.
//

typedef unsigned long int tInterlockBits;

#define IL_az_off 0x000001
#define IL_el_off 0x010000
#define IL_el_reserved_1 0x020000
#define IL_el_reserved_2 0x040000
#define IL_el_reserved_3 0x080000

char *
explainInterlock(tInterlockBits bits) {
  // Interlock chain which is common to both Az and El.
  if (!(bits & 0x000800))
    return ("main panel 'STOP' switch");
  if (!(bits & 0x000400))
    return ("emergency stop switch");
  if (!(bits & 0x000200))
    return ("downstairs pedestal manual interlock switch");
  if (!(bits & 0x000100))
    return ("Bronto switch");
  if (!(bits & 0x000080))
    return ("upstairs pedestal hoist switch");
  if (!(bits & 0x000040))
    return ("receiver interlock switch");
  if (!(bits & 0x000020))
    return ("upstairs pedestal manual interlock switch");
  if (!(bits & 0x000010))
    return ("(reserved option switch, not yet used)");
  // End of common interlock chain.

  // When common chain is true, the following apply to Az.
  if (!(bits & 0x000008))
    return ("Az1 motor interlock switch");
  if (!(bits & 0x000004))
    return ("CW final limit switch");
  if (!(bits & 0x000002))
    return ("CCW final limit switch");
  if (!(bits & IL_az_off))
    return ("'AZ OFF' panel switch");

  // When common chain is true, the following apply to El.
  if (!(bits & 0x001000))
    return ("upstairs pedestal stow pin switch");
  if (!(bits & 0x002000))
    return ("El1 motor interlock switch");
  if (!(bits & 0x004000))
    return ("UP final limit switch");
  if (!(bits & 0x008000))
    return ("DOWN final limit switch");
  if (!(bits & IL_el_off))
    return ("'EL OFF' panel switch");

  // "Interlock ok" messages.
  if ((bits & 0x01ffef) == 0x01ffef) {
    if ((bits & 0xf00000) == 0xf00000)
      return ("all interlock switches ok");
    if (!(bits & 0x800000))
      return ("CW slow limit reached, interlock ok");
    if (!(bits & 0x400000))
      return ("CCW slow limit reached, interlock ok");
    if (!(bits & 0x200000))
      return ("UP slow limit reached, interlock ok");
    if (!(bits & 0x100000))
      return ("DOWN slow limit reached, interlock ok");
  }  // if interlock ok but slow limits on

  // Az, El, or both limit switches overridden manually from panel.
  if ((bits & 0x01ffe9) == 0x01ffe9)
    return ("ok with Az limits overridden");
  if ((bits & 0x013fef) == 0x013fef)
    return ("ok with El limits overridden");
  if ((bits & 0x01ffef) == 0x01ffef)
    return ("ok with both Az and El limit switches overridden");

  return ("???");
}  // explainInterlock
