#define SUNANT 1

#include "registration.h"
RCSID(ant_cpp, "$Id: ant.cpp,v 1.41 2005/11/09 11:49:58 lindi Exp $");
/* ------------------------------------------------------------------ */
// ant/dos/up/ant.cpp  --  Pedestal PC (up) control program.

// $Log: ant.cpp,v $
// Revision 1.41  2005/11/09 11:49:58  lindi
// Update azSpeedPWM struct for new Az motor.
//
// Revision 1.40  2001/09/10 12:27:33  amn
// Hmm, this should be the same as 1.40, why does CVS want to commit this?
//
// Revision 1.39  2001/04/12 06:59:25  amn
// Saved this spring 2001 version when antup1 --> antup3 (stuck CPU fan).
//
// Revision 1.38  2001/01/03 08:21:52  amn
// Corrected El2 motor tacho ticks to antenna degrees coeff bug (was 1.0deg).
//
// Revision 1.37  2000/12/30 13:50:24  amn
// Limited calculated goalVel to maximum antenna hw allowed.
//
// Revision 1.36  2000/10/23 07:08:00  amn
// Tacho augmentation up to one "bit" of absolute encoder, not more.
//
// Revision 1.35  2000/10/06 10:54:53  amn
// Started to use tachometers for finer encoders.
//
// Revision 1.34  2000/08/14 11:09:29  amn
// SunAnt el encoder fine-tuning, tachometer motor speeds (gear reductions).
//
// Revision 1.33  2000/08/10 09:33:06  amn
// First usable SunAnt version.
//
// Revision 1.32  2000/08/01 08:22:34  amn
// Spring 2000 changes (speed table).
//
// Revision 1.31  2000/03/13 08:58:53  amn
// Changed 'CC_vax_position' into 'CC_track_position'.
//
// Revision 1.30  2000/03/03 11:47:52  amn
// Added tachometer "micro-boost" inner servo loop.
//
// Revision 1.29  2000/02/20 13:59:12  amn
// Added "slow limit switches" also in software.
// Limits max speed when antenna is in extreme positions
// so that hw PWM never goes over +/-255 (the hardware limit).
//
// Revision 1.28  2000/02/09 07:46:30  amn
// Measured speed--PWM tables with only 22GHz VLBI RX in antenna.
//
// Revision 1.27  1999/11/22 13:16:07  amn
// Added auto-brakes if downstairs dies, adjusted az gain 0.6-->0.4.
//
// Revision 1.26  1999/11/20 19:49:34  amn
// Corrected unsymmetrical posErr calculation in velocity servo.
//
// Revision 1.25  1999/11/18 09:25:50  amn
// Velocity servo as tested in 1st solar map/track on 17-Nov-1999.
//
// Revision 1.24  1999/11/09 14:25:24  amn
// Added max acceleration speed/PWM object.
//
// Revision 1.23  1999/11/07 21:57:46  amn
// First version with passable velocity servo.
//
// Revision 1.22  1999/11/06 22:33:25  amn
// New 'cVelAcc' class for calculating pos/vel/acc.
// Added getting 'tTime' delta in 'double' fractional seconds.
//
// Revision 1.21  1999/11/05 08:41:29  amn
// "Final" super-instant servo version, before starting to do velocity servo.
//
// Revision 1.20  1999/11/02 11:17:02  amn
// Experimented with instant servo.
//
// Revision 1.19  1999/11/01 18:00:32  amn
// Getting xntpd UTC time from downstairs via uplink.
//
// Revision 1.18  1999/10/28 18:14:37  amn
// Debugged cROC4xx repeats (7 not enough...).
// Added ignoring values >10% off from last 4 in cSensorChannel.
// Added 'tTime' alias struct timeval (2x longs) into ctime.cpp.
// Added PrevLoopUsecs timing.
//
// Revision 1.17  1999/10/28 11:31:47  amn
// Resolved CVS log conflict.
//
// Revision 1.16  1999/10/28 11:01:49  amn
// Tried shortening channel/gain change delay from 1ms to 10usec.
//
// Revision 1.15  1999/10/28 08:29:55  amn
// Separated cTime class to a separate file.
//
// Revision 1.14  1999/10/22 12:59:09  amn
// Initial VAX tracking capability.
//
// Revision 1.13  1999/10/22 10:04:04  amn
// Fixed PWMISA usec timer --> seconds problem,
// started using 'cprintf()' + 'directvideo' for faster screen updates.
//
// Revision 1.12  1999/10/19 12:42:32  amn
// Fine-tuning elSpeed tables.
//
// Revision 1.11  1999/10/15 12:47:19  amn
// Patched ROC412 to be 10 bits only until wiring is corrected.
// Both downlink and uplink in 'downlink.h'.
// 'serial.c' enhanced to support (and require!) receiver FIFOs,
// thus serial link works 99.99% reliably.  (Send timeout too short, still?)
//
// Revision 1.10  1999/10/14 12:48:49  amn
// Added IK121 fault status detection; both down and uplinks.
// Speed to PWM mapping interpolation tables.
//
// Revision 1.9  1999/10/13 12:16:21  amn
// IK121 stopped/refmark search, all pwm variables not in 'sh->'.
//
// Revision 1.8  1999/10/12 14:23:00  jen
// Completely reworked to use 'sh->' global struct for (almost) all variables.
//
// Revision 1.7  1999/10/12 08:31:25  amn
// Added some handling of IK121 and serial line from downstairs.
//
// Revision 1.6  1999/10/12 08:16:18  amn
// Major revision based on 'downf', added IK121.
//
// Revision 1.19  1999/10/11 12:39:58  amn
// Hide cursor.
//
// Revision 1.18  1999/10/05 14:09:29  amn
// Tested 10kHz PWM base frequency.
//
// Revision 1.17  1999/10/05 12:24:44  amn
// Added using 'cQDR.cpp' include.
//
// Revision 1.16  1999/10/04 10:13:03  amn
// Adding tiltmeter.
//
// Revision 1.15  1999/09/30 17:03:02  amn
// Removed cQDR code to a separate cQDR.cpp file.
//
// Revision 1.14  1999/09/30 10:26:43  amn
// Added brake I measurements.
//
// Revision 1.13  1999/09/29 11:15:23  amn
// Added absolute Az/El, tacho/QDR, and interlock in hexadecimal.
//
// Revision 1.12  1999/09/29 07:12:33  amn
// Harmonized c24In and c24Out.
//
// Revision 1.11  1999/09/28 12:04:13  amn
// Changed the location of c8255.cpp and its charset to DOS.
//
// Revision 1.10  1999/09/28 07:28:07  amn
// Just before starting changes for ant/up with the final pedestal PC
// in tower; changing to CR-LF DOS convention for direct CVS use.
//
// Revision 1.9  1999/08/20 10:57:07  amn
// New PWMISA code replaced old wire-wrap PWM board code.
//
// Revision 1.8  1999/05/11 11:36:17  amn
// Added more A/D channels, 's' for servoing.
//
// Revision 1.7  1999/03/18 10:21:26  amn
// Added missing CVS log entries to this file.
//
// Revision 1.6  1999/03/18 09:28:41  amn
// The first super-simple servo with 0.070 degree error/slowness...
//
// Revision 1.5  1999/03/16 08:51:33  amn
// Version used to measure antiphase FET16r board prototypes.
//
// Revision 1.4  1999/02/05 07:53:28  amn
// Changed "stress test" into scanning all motor drive value pairs (Az1, Az2).
//
// Revision 1.3  1999/02/02 08:07:33  amn
// Added speed and acceleration values to display and log.
//
// Revision 1.2  1999/01/28 13:14:14  amn
// Changed automatic RCS version printout from Revision to Id.
//
// Revision 1.1  1999/01/28 13:11:32  amn
// First version with antenna position logging to ant.log.
//
// Revision 1.0  1999/01/28 08:38:22  amn
// First version which duplicates old 'ant.c' FET board tester with cPWMISA.
//
// Revision 0.5  1999/01/27 13:55:23  amn
// First release which compiled with C++ and class cPWMISA.
// Problems with setting PWM timers/counters (stay at "slow" speed).
//
// 27-Jan-1999  amn  Original version with class cPWMISA.


#define AD_CURRENT_MEASUREMENTS 1
#define SCREEN_PRINTF_UPDATES 1
#define SCREEN_BARE_MINIMUM_PRINTFS 1
#define SCREEN_TIME_DISPLAY 1
#define CATCH_OVER_10AMPS 0

// Servo parameters.
#if SUNANT
 #define MAXACC 0.06
 #define MAXVEL 0.09
 #define AZGAINCOEFF 0.4
 #define ELGAINCOEFF 0.6
 #define TACHOBOOSTGAIN 0.25
 #define MAXTACHOBOOSTVEL 0.030  // deg/s
  // tacho speed calc range about -0.045..+0.045 deg/s (elevation more limiting)
#else
 // Normal large antenna.
 #define MAXACC 0.3
 #define MAXVEL 1.2
 #define AZGAINCOEFF 0.4
 // Was in Jan/Feb-2000: #define ELGAINCOEFF 0.7
 // Now on 10-Apr-2000:
 #define ELGAINCOEFF 0.6
 #define TACHOBOOSTGAIN 0.25
 // The speed limit up to which tachometer deliver accurate speed info.
 #define MAXTACHOBOOSTVEL 0.200  // deg/s
  // tacho speed calc range about -0.285..+0.285 deg/s
#endif

// Uplink maximum delays, after MAXDELAYFROMDOWNSTAIRS we start putting
// speeds to zero.
#define MAXDELAYFROMDOWNSTAIRS 2000
// An additional delay after which the motor relays are turned off.
#define SHUTDOWNDELAY 4000

// How many PWM units to "drive against" motor 1 vs motor 2?
// (Tried 18-Nov-1999 '60'.)
// (For Nov-1999 until Jan-2000 this was 40.)
#define TIGHTENGEARBOX 50

// Forced "slow speed" limit switches.
#if SUNANT
 // SunAnt az limit switches activate at
 //  az -164.476 .. +165.860, el -0.3 .. +98.437.
 // Deactivation requires only about 0.2--0.3degrees "back".
 // 09-Oct-2000: el downlimit seems to activate at +0.45 (about),
 //  changed MINSLOWEL from +0.2 to +1.0.
 #define MINSLOWAZ (-163.0)
 #define MAXSLOWAZ (+163.0)
 #define MINSLOWEL (+1.0)
 #define MAXSLOWEL (+98.0)
#else
 // xxx: check these az limits using actual switch hardware
 #define MINSLOWAZ (-240.0)
 #define MAXSLOWAZ (+240.0)
 #define MINSLOWEL (4.0)
  // Min El has been checked: coming down at 1.3deg/s the antenna
  // slows down nicely at +4.0deg so that at 1.3deg it's <0.1deg/s.
  // (The slow limit switch activates at about +2.0deg.)
 #define MAXSLOWEL (100.0)
#endif
// Log time and value change resolution.
#define MINLOGCHANGE 0.00005  // degrees
#define MINLOGUPDTIME 100  // ms

#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>  // for fabs()
#include <assert.h>
#include <dos.h>
#include <time.h>


/*------------------------------------------------------------*/
// ACL-712 6x3x8bit digital I/O board, surprisingly similar to
// ET7122 Capax 6x3x8bit DI/O board.  Emulates 6 8255 Intel PIO chips.
#include "c8255.cpp"

/*------------------------------------------------------------*/
// Both absolute encoder paraller cROC4xx and
// A/D converter cET28XX need offset/coeffiecient class cSensor.
#include "cSensor.cpp"

/*------------------------------------------------------------*/
// ACL-8113 32-ch 12-bit A/D board, surprisingly similar to
// ET2816 Capax 16-ch 16-bit A/D board.
#define DEBUG_REPEATS 0
#define NO_ET28xx_PRESENT 0
#include "cET28xx.cpp"

/* ------------------------------------------------------------------ */
// QDR 4-connector quadrature encoder decoding board,
// connected to one connector on ACL-712 6x3x8bit digital I/O board.
// Heidenhain ERN420 digital quadrature tachometers are being used;
// they have max. 12000 rev/minute rating and a maximum change frequency
// of A/B signals of 300kHz.  We have 2500 slots/rev in their optical disk
// so we'll reach 300kHz at 120 rev/sec, that is, 7200 rev/minute.
// HCTL-2016 on QDR board has 10MHz clock and it requires stable A/B signal
// for three clock periods => max. A/B frequency is 3.3MHz >> 300kHz.
// 26LS32 RS422 receivers should work in the region of 30ns i.e. 33MHz.
// HCTL-2016 16-bit counter overflows with ERN420/2500 after
// 65536/(4*2500) == 6.5 revs.  At ERN420 max this is only 54.2msec.
// The observed max speed of antenna, about 1.2deg/sec translates
// to 1.2*(18187..18295.2)deg/sec == 21824.4..21954.24deg/sec at motor,
// i.e. 60.62..60.98 rev/sec (3637..3659rev/minute), roughly 150kHz output.
#include "cQDR.cpp"

#if SUNANT
// SunAnt encoders.

/*------------------------------------------------------------*/
// Heidenhain ECN113 13-bit SSI output absolute encoders (Az/El),
// each connected to one connector on ACL-712 6x3x8bit digital I/O board.
#define DEBUG_REPEATS 1
#include "cSSI.cpp"

/*------------------------------------------------------------*/
// SunAnt virtual incremental encoders.
// (Combining absolute encoder values and tachometer values.)
class cIncrTacho {
private:
  double prevAbs;  // for value change detection
  unsigned int prevTacho;  // for tacho count difference
  double cumuTacho;  // cumulative degrees
  double toDegCoeff;  // from tacho counts to incr degrees
  double maxAugDeg;  // maximum tacho augmentation, +/-
public:
  virtual void init(double toDegCoeff, double maxAugDeg);
  virtual void cleanup(void);
  // Set absolute and tacho values.
  virtual void set(double abso, unsigned int tacho);
  // Get back interpolated value.
  virtual double get(void);
};  // cIncrTacho

void
cIncrTacho::init(double toDegCoeff, double maxAugDeg)
{
  this->prevAbs = 0.0;
  this->prevTacho = 0;
  this->cumuTacho = 0.0;
  this->toDegCoeff = toDegCoeff;
  this->maxAugDeg = maxAugDeg;
}  // init

void
cIncrTacho::cleanup(void)
{
}  // cleanup

// Detect absolute encoder change, its direction(?), and augment with tacho.
void
cIncrTacho::set(double abso, unsigned int tacho)
{
  if (abso == this->prevAbs) {
    // Augment with tacho.
    double diff = (double)tacho - (double)this->prevTacho;

    // Handle tacho 16-bit counter wrap-around.
    if (diff > 32767.0) {
      diff -= 65536.0;
    } else if (diff < -32768.0) {
      diff += 65536.0;
    }
    this->cumuTacho += (this->toDegCoeff * diff);
    // Clip to a maximum of absolute "real" encoder".
    if (this->cumuTacho > this->maxAugDeg) {
      this->cumuTacho = this->maxAugDeg;
    } else if (this->cumuTacho < -(this->maxAugDeg)) {
      this->cumuTacho = -(this->maxAugDeg);
    }
  } else {
    // Fresh, new absolute value, reset tacho augmentation.
    this->cumuTacho = 0.0;
    // Store the new absolute value.
    this->prevAbs = abso;
  }
  this->prevTacho = tacho;
}  // set

double
cIncrTacho::get(void)
{
  return (this->prevAbs + this->cumuTacho);
}  // get

#else
// Normal large antenna encoders.

/*------------------------------------------------------------*/
// Heidenhain ROC412 12-bit parallel output absolute encoders (Az/El),
// each connected to one connector on ACL-712 6x3x8bit digital I/O board.
#define DEBUG_REPEATS 1
#include "cROC4xx.cpp"

/* ------------------------------------------------------------------ */
/* Heidenhain IK121/RON806 incremental encoders. */
#include "cIK121.cpp"
#endif

/*------------------------------------------------------------*/
#include "cPWMISA.cpp"

// Setting up 1pps IRQ of PWMISA card "manually".

class cPWMISA vPWMISA;

#define ICW1		0x20		/* Address of primary 8259 ICW. */
#define INTCON2		0xA1		/* Address of secondary 8259 . */
#define ICW2		0xA0		/* Address of secondary 8259 ICW. */
#define INT11		0x73		/* Interrupt 11 number */
#define EOI11		0x63		/* Specific EOI for int11 */
#define NS_EOI		0x20		/* Non-specific EOI */


#ifdef __cplusplus
    #define __CPPARGS ...
    #define __MUSTBEFAR
#else
    #define __CPPARGS
    #define __MUSTBEFAR far
#endif

void interrupt __MUSTBEFAR (*intsave)(__CPPARGS);
void interrupt __MUSTBEFAR timerInt(__CPPARGS);


void realtimeInstall(void) {
	/* Get current local time and start incrementing from there */
	/* with the timer interrupt subroutine. */
#if 0
    // No more interrupt-incremented time.
    tzset();
	time(&ltime);
#endif
	intsave =  getvect(INT11);
	setvect(INT11, timerInt);

  // IRQ11 is not enabled in 8259 interrupt controller chip by default:
  // (Need to put bit 3 (==3+8=11) to low.)
  outportb(INTCON2, (inportb(INTCON2) & ~(0x08)) );
}  // reatimeInstall

void realtimeRestore(void) {
  // Put IRQ11 disable bit back to 1 in controller chip.
  outportb(INTCON2, (inportb(INTCON2) | 0x08) );
  setvect(INT11, intsave);
}  // realtimeRestore


void interrupt __MUSTBEFAR timerInt(__CPPARGS)
{
#if 0
  Could check vPWMISA.getUsecs() to be close to zero.
#endif

  outportb(ICW1,NS_EOI);			/* Send End-of-Interrupt to 8259. */
  outportb(ICW2,EOI11);			/* Send End-of-Interrupt to 8259. */
}  /* timerInt */

/* ----------- */
// Motor PWM tables on top of PWMISA "bare" PWM control.
typedef struct sSpeedPWM {
  double speed;  // degrees/s?
  int mPWM;
  int m1tight;
  int m1antiph;
  int m2tight;
  int m2antiph;
} tSpeedPWM, *ptSpeedPWM;

tSpeedPWM azSpeedPWM[] = {
#if 0
  { -9.0000, +1023, 0, +1000, 0 },
  { -1.2000,  +785, 0,  +723, 0 },
  { -0.5000,  +475, 0,  +380, 0 },
  { -0.1000,  +187, 0,   +83, 0 },
  { -0.0100,  +102, 0,   -32, 0 },
  { -0.0010,  +107, 0,   -37, 0 },
  { -0.0001,   +72, 0,   -17, 0 },
  {  0.0000,   +50, 0,   -50, 0 },
  {  0.0001,   +17, 0,   -82, 0 },
  {  0.0010,   +34, 0,  -100, 0 },
  {  0.0100,    +3, 0,   -96, 0 },
  {  0.1000,   -81, 0,  -181, 0 },
  {  0.5000,  -390, 0,  -490, 0 },
  { +1.2000,  -713, 0,  -785, 0 },
  { +9.0000, -1000, 0, -1023, 0 }
#endif
#if 0
// Coarsely and manually measured on 15-Oct-1999 by amn:
// Swapping m1 and m2 columns swaps gearbox tightening direction.
  { -9.0000, -1023, 0, -1000, 0 },
  { -1.2000,  -785, 0,  -723, 0 },
  { -0.5000,  -475, 0,  -380, 0 },
  { -0.1000,  -187, 0,   -83, 0 },
  { -0.0100,  -102, 0,   +32, 0 },
  { -0.0010,  -107, 0,   +37, 0 },
  { -0.0001,   -72, 0,   +17, 0 },
  {  0.0000,   -50, 0,   +50, 0 },
  {  0.0001,   -17, 0,   +82, 0 },
  {  0.0010,   -34, 0,  +100, 0 },
  {  0.0100,    -3, 0,   +96, 0 },
  {  0.1000,   +81, 0,  +181, 0 },
  {  0.5000,  +390, 0,  +490, 0 },
  { +1.2000,  +713, 0,  +785, 0 },
  { +9.0000, +1000, 0, +1023, 0 }
#endif
#if 0
// Measured 07-Nov-1999 with "sawtooth" PWM -75..75 with TIGHTENGEARBOX 40.
// File ant-1999110705.log (for -0.030..0.030)
// and ant-1999110706.log (for -0.5..0.5).
  { -9.0000,  -983, +40, 0, -40, 0 },
  { -1.2000,  -980, +40, 0, -40, 0 },
  { -0.8000,  -600, +40, 0, -40, 0 },
  { -0.5000,  -420, +40, 0, -40, 0 },
  { -0.1000,  -142, +40, 0, -40, 0 },
  { -0.0500,  -105, +40, 0, -40, 0 },
  { -0.0300,   -72, +40, 0, -40, 0 },
  { -0.0200,   -61, +40, 0, -40, 0 },
  { -0.0100,   -44, +40, 0, -40, 0 },
  { -0.0050,   -41, +40, 0, -40, 0 },
  { -0.0030,   -39, +40, 0, -40, 0 },
  { -0.0020,   -37, +40, 0, -40, 0 },
  { -0.0010,   -32, +40, 0, -40, 0 },
  {  0.0000,    +0, +40, 0, -40, 0 },
  { +0.0010,   +30, +40, 0, -40, 0 },
  { +0.0020,   +36, +40, 0, -40, 0 },
  { +0.0030,   +38, +40, 0, -40, 0 },
  { +0.0050,   +39, +40, 0, -40, 0 },
  { +0.0100,   +44, +40, 0, -40, 0 },
  { +0.0200,   +60, +40, 0, -40, 0 },
  { +0.0300,   +78, +40, 0, -40, 0 },
  { +0.0500,   +80, +40, 0, -40, 0 },
  { +0.1000,  +130, +40, 0, -40, 0 },
  { +0.5000,  +425, +40, 0, -40, 0 },
  { +0.8000,  +600, +40, 0, -40, 0 },
  { +1.2000,  +980, +40, 0, -40, 0 },
  { +9.0000,  +983, +40, 0, -40, 0 }
#endif
#if 0
// Measured 08-Nov-1999 with "sawtooth" PWM -700..-100..0..100..700
// with TIGHTENGEARBOX 40.
// File ant-1999110802.log.
  { -9.0000,  -983, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -1.2000,  -980, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.8400,  -700, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.1000,  -130, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0500,   -85, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
#if 0
  { -0.0100,   -49, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0005,   -36, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  {  0.0000,    -3, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0005,   +33, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0100,   +46, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
#endif
#if 0
  // Fine-tuning based on ant-1999111801--03.log.
  // (Had only 37GHz continuum RX on; this was used with both 22&37GHz cont.)
  { -0.0200,   -61, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0100,   -44, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0005,   -39, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  {  0.0000,    -3, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0005,   +36, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0100,   +41, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0200,   +59, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
#endif
  { +0.0500,   +85, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.1000,  +130, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.8400,  +700, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +1.2000,  +980, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +9.0000,  +983, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 }
#endif

#if 1
 #if SUNANT
  #if 0
  // Measured with manual PWM values and observed tachometer speeds
  // on 05-Oct-2000.
  { -9.0000,  -1023, 0, 0, 0, 0 },
  { -0.0970,  -1020, 0, 0, 0, 0 },
  { -0.0716,  -750,  0, 0, 0, 0 },
  { -0.0473,  -500,  0, 0, 0, 0 },
  { -0.0230,  -250,  0, 0, 0, 0 },
#define MINSLOWAZVEL (-0.022)
  { -0.0181,  -200,  0, 0, 0, 0 },
  { -0.0122,  -140,  0, 0, 0, 0 },
  { -0.0064,   -80,  0, 0, 0, 0 },
  { -0.0044,   -60,  0, 0, 0, 0 },
  { -0.0018,   -35,  0, 0, 0, 0 },
  { -0.0002,   -20,  0, 0, 0, 0 },
  { +0.0002,   +20,  0, 0, 0, 0 },
  { +0.0008,   +25,  0, 0, 0, 0 },
  { +0.0013,   +30,  0, 0, 0, 0 },
  { +0.0019,   +35,  0, 0, 0, 0 },
  { +0.0024,   +40,  0, 0, 0, 0 },
  { +0.0044,   +60,  0, 0, 0, 0 },
  { +0.0063,   +80,  0, 0, 0, 0 },
  { +0.0092,  +110,  0, 0, 0, 0 },
  { +0.0121,  +140,  0, 0, 0, 0 },
  { +0.0179,  +200,  0, 0, 0, 0 },
#define MAXSLOWAZVEL (+0.022)
  { +0.0226,  +250,  0, 0, 0, 0 },
  { +0.0467,  +500,  0, 0, 0, 0 },
  { +0.0714,  +750,  0, 0, 0, 0 },
  { +0.0962, +1020,  0, 0, 0, 0 },
  { +9.0000, +1023,  0, 0, 0, 0 }
  #endif
  #if 1
  // Measured with manual PWM values and observed tachometer speeds
  // on 09-Nov-2005.
  { -9.0000,  -512, 0, 0, 0, 0 },
  { -0.1060,  -500, 0, 0, 0, 0 },
  { -0.1000,  -480,  0, 0, 0, 0 },
  { -0.0930,  -440,  0, 0, 0, 0 },
  { -0.0800,  -380,  0, 0, 0, 0 },
#define MINSLOWAZVEL (-0.022)
  { -0.0700,  -340,  0, 0, 0, 0 },
  { -0.0620,  -300,  0, 0, 0, 0 },
  { -0.0534,  -260,  0, 0, 0, 0 },
  { -0.0483,  -240,  0, 0, 0, 0 },
  { -0.0400,  -200,  0, 0, 0, 0 },
  { -0.0300,  -160,  0, 0, 0, 0 },
  { -0.0265,  -140,  0, 0, 0, 0 },
  { -0.0220,  -120,  0, 0, 0, 0 },
  { -0.0183,  -100,  0, 0, 0, 0 },
  { -0.0139,   -80,  0, 0, 0, 0 },
  { -0.0117,   -70,  0, 0, 0, 0 },
  { -0.0095,   -60,  0, 0, 0, 0 },
  { -0.0080,   -55,  0, 0, 0, 0 },
  { -0.0073,   -50,  0, 0, 0, 0 },
  { -0.0064,   -45,  0, 0, 0, 0 },
  { -0.0050,   -40,  0, 0, 0, 0 },
  { -0.0039,   -35,  0, 0, 0, 0 },
  { -0.0025,   -30,  0, 0, 0, 0 },
  { -0.0016,   -25,  0, 0, 0, 0 },
  { -0.0005,   -20,  0, 0, 0, 0 },
  { +0.0013,   +25,  0, 0, 0, 0 },
  { +0.0026,   +30,  0, 0, 0, 0 },
  { +0.0036,   +35,  0, 0, 0, 0 },
  { +0.0047,   +40,  0, 0, 0, 0 },
  { +0.0059,   +45,  0, 0, 0, 0 },
  { +0.0070,   +50,  0, 0, 0, 0 },
  { +0.0080,   +55,  0, 0, 0, 0 },
  { +0.0092,   +60,  0, 0, 0, 0 },
  { +0.0115,   +70,  0, 0, 0, 0 },
  { +0.0135,   +80,  0, 0, 0, 0 },
  { +0.0179,  +100,  0, 0, 0, 0 },
  { +0.0120,  +120,  0, 0, 0, 0 },
  { +0.0264,  +140,  0, 0, 0, 0 },
  { +0.0350,  +180,  0, 0, 0, 0 },
  { +0.0440,  +220,  0, 0, 0, 0 },
#define MAXSLOWAZVEL (+0.022)
  { +0.0577,  +280,  0, 0, 0, 0 },
  { +0.0665,  +320,  0, 0, 0, 0 },
  { +0.0800,  +380,  0, 0, 0, 0 },
  { +0.0890,  +420,  0, 0, 0, 0 },
  { +0.1070,  +500,  0, 0, 0, 0 },
  { +9.0000,  +512,  0, 0, 0, 0 }
  #endif
 #else
  // Fine-tuning based on antnn.log files on 2000-01-27, TIGHTEN==50.
  // (Elevations 15/30/45/60/85, with only 22GHz VLBI RX in tower.)
  { -9.0000,  -983, +TIGHTENGEARBOX-30, 0, -TIGHTENGEARBOX+30, 0 },
  { -1.2000,  -980, +TIGHTENGEARBOX-30, 0, -TIGHTENGEARBOX+30, 0 },
  { -0.8400,  -700, +TIGHTENGEARBOX-20, 0, -TIGHTENGEARBOX+20, 0 },
#define MINSLOWAZVEL (-0.1)
  { -0.1000,  -115, +TIGHTENGEARBOX-10, 0, -TIGHTENGEARBOX+10, 0 },
  { -0.0500,   -75, +TIGHTENGEARBOX-5,  0, -TIGHTENGEARBOX+5, 0 },

  { -0.0200,   -49, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0150,   -42, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0100,   -36, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0050,   -28, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0010,   -25, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0005,   -21, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  {  0.0000,    +2, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0005,   +24, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0010,   +31, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0050,   +34, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0100,   +40, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0200,   +51, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },

  { +0.0500,   +78, +TIGHTENGEARBOX-5,  0, -TIGHTENGEARBOX+5, 0 },
  { +0.1000,  +115, +TIGHTENGEARBOX-10, 0, -TIGHTENGEARBOX+10, 0 },
#define MAXSLOWAZVEL (+0.1)
  { +0.8400,  +700, +TIGHTENGEARBOX-20, 0, -TIGHTENGEARBOX+20, 0 },
  { +1.2000,  +980, +TIGHTENGEARBOX-30, 0, -TIGHTENGEARBOX+30, 0 },
  { +9.0000,  +983, +TIGHTENGEARBOX-30, 0, -TIGHTENGEARBOX+30, 0 }
 #endif
#endif
};  // azSpeedPWM

tSpeedPWM elSpeedPWM[] = {
#if 0
  { -9.0000, -1000, 0, -1023, 0 },
  { -1.0000,  -600, 0,  -700, 0 },
  {  0.0000,   +50, 0,   -50, 0 },
  { +1.0000,  +700, 0,  +600, 0 },
  { +9.0000, -1023, 0, -1000, 0 }
#endif
#if 0
  // Just copied on 19-Oct-199 from az,
  // cannot be right because the slight trend of antenna
  // to "lift up", in + el direction.
  // 19-Oct-1999: fine-tuned with "zero speed" measurements
  // of El1=+30..+67..+105|El2=-100; El1=+105..+172..+240|El2=-200
  { -9.0000, -1000, 0, -1023, 0 },
  { -1.2000,  -723, 0,  -785, 0 },
  { -0.5000,  -380, 0,  -475, 0 },
  { -0.1000,   -83, 0,  -187, 0 },
  // { -0.0100,   +32, 0,  -102, 0 },
  // { -0.0010,   +37, 0,  -107, 0 },
  { -0.0001,   +25, 0,  -100, 0 },
  {  0.0000,  +172, 0,  -200, 0 },
  {  0.0001,   +82, 0,   -17, 0 },
  {  0.0010,  +100, 0,   -34, 0 },
  {  0.0100,   +96, 0,    -3, 0 },
  {  0.1000,  +181, 0,   +81, 0 },
  {  0.5000,  +490, 0,  +390, 0 },
  { +1.2000,  +785, 0,  +713, 0 },
  { +9.0000, +1023, 0, +1000, 0 }
#endif
#if 0
// Copied directly from azimuth.
// Measured 07-Nov-1999 with "sawtooth" PWM -75..75 with TIGHTENGEARBOX 40.
// File ant-1999110705.log (for -0.030..0.030)
// and ant-1999110706.log (for -0.5..0.5).
  { -9.0000,  -983, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -1.2000,  -980, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.8000,  -600, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.5000,  -420, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.1000,  -142, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0500,  -105, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0300,   -72, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0200,   -61, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0100,   -44, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0050,   -41, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0030,   -39, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0020,   -37, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0010,   -32, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  {  0.0000,    +0, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0010,   +30, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0020,   +36, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0030,   +38, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0050,   +39, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0100,   +44, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0200,   +60, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0300,   +78, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0500,   +80, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.1000,  +130, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.5000,  +425, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.8000,  +600, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +1.2000,  +980, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +9.0000,  +983, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 }
#endif
#if 0
// Measured 08-Nov-1999 with "sawtooth" PWM -700..-100..0..100..700
// with TIGHTENGEARBOX 40.
// File ant-1999110802.log.
  { -9.0000,  -983, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -1.2000,  -980, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.9000,  -700, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.1000,  -135, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0500,   -90, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
#if 0
// Measured 08-Nov-1999 with "sawtooth" PWM -700..-100..0..100..700
// with TIGHTENGEARBOX 40.
// File ant-1999110802.log.
  { -0.0100,   -57, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0050,   -52, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0005,   -46, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  {  0.0000,   -22, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0005,    +9, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0050,   +14, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0100,   +17, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
#endif
  { +0.0500,   +56, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.1000,   +95, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.8600,  +700, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +1.2000,  +980, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +9.0000,  +983, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 }
#endif

#if 0
// Measured 27-Jan-2000 with "sawtooth" PWM -120..-100..-70..0..70..100..120
// with TIGHTENGEARBOX 50.
// Files antnn.log, nn=15/30/45/60/85, elevation used.
  { -9.0000,  -983, +TIGHTENGEARBOX-30, 0, -TIGHTENGEARBOX+30, 0 },
  { -1.2000,  -980, +TIGHTENGEARBOX-30, 0, -TIGHTENGEARBOX+30, 0 },
  { -0.9000,  -700, +TIGHTENGEARBOX-20, 0, -TIGHTENGEARBOX-20, 0 },
  { -0.1000,  -110, +TIGHTENGEARBOX-10, 0, -TIGHTENGEARBOX-10, 0 },
#define MINSLOWELVEL (-0.1)
  { -0.0500,   -70, +TIGHTENGEARBOX-5,  0, -TIGHTENGEARBOX-5, 0 },

  { -0.0200,   -43, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0150,   -40, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0100,   -36, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0050,   -29, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0010,   -24, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0005,   -20, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  {  0.0000,    +0, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0005,   +21, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0010,   +25, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0050,   +28, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0100,   +33, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0150,   +38, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0200,   +41, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },

  { +0.0500,   +70, +TIGHTENGEARBOX-5,  0, -TIGHTENGEARBOX+5, 0 },
  { +0.1000,  +110, +TIGHTENGEARBOX-10, 0, -TIGHTENGEARBOX+10, 0 },
#define MAXSLOWELVEL (+0.1)
  { +0.8600,  +700, +TIGHTENGEARBOX-20, 0, -TIGHTENGEARBOX+20, 0 },
  { +1.2000,  +980, +TIGHTENGEARBOX-30, 0, -TIGHTENGEARBOX+30, 0 },
  { +9.0000,  +983, +TIGHTENGEARBOX-30, 0, -TIGHTENGEARBOX+30, 0 }
#endif

#if 1
 #if SUNANT
  #if 1
  // Measured with manual PWM values and observed tachometer speeds
  // on 05-Oct-2000.
  { -9.0000,  -1023, 0, 0, 0, 0 },
  { -0.0450,  -903,  0, 0, 0, 0 },
  { -0.0367,  -600,  0, 0, 0, 0 },
  { -0.0242,  -400,  0, 0, 0, 0 },
  { -0.0118,  -200,  0, 0, 0, 0 },
#define MINSLOWELVEL (-0.012)
  { -0.0054,  -100,  0, 0, 0, 0 },
  { -0.0036,   -70,  0, 0, 0, 0 },
  { -0.0024,   -50,  0, 0, 0, 0 },
  { -0.0012,   -30,  0, 0, 0, 0 },
  { -0.0006,   -20,  0, 0, 0, 0 },
  { -0.0002,   -15,  0, 0, 0, 0 },
  { +0.0002,   +35,  0, 0, 0, 0 },
  { +0.0005,   +40,  0, 0, 0, 0 },
  { +0.0011,   +50,  0, 0, 0, 0 },
  { +0.0027,   +70,  0, 0, 0, 0 },
  { +0.0047,  +100,  0, 0, 0, 0 },
  { +0.0109,  +200,  0, 0, 0, 0 },
#define MAXSLOWELVEL (+0.011)
  { +0.0233,  +400,  0, 0, 0, 0 },
  { +0.0359,  +600,  0, 0, 0, 0 },
  { +0.0450, +1020,  0, 0, 0, 0 },
  { +9.0000, +1023,  0, 0, 0, 0 }
  #endif
 #else
// Measured 10-Apr-2000 with "sawtooth" PWM -120..-100..-70..0..70..100..120
// with TIGHTENGEARBOX 50.
// File ant45.log, elevation 45deg.
  { -9.0000,  -983, +TIGHTENGEARBOX-30, 0, -TIGHTENGEARBOX+30, 0 },
  { -1.2000,  -980, +TIGHTENGEARBOX-30, 0, -TIGHTENGEARBOX+30, 0 },
  { -0.9000,  -700, +TIGHTENGEARBOX-20, 0, -TIGHTENGEARBOX-20, 0 },
  { -0.1000,   -80, +TIGHTENGEARBOX-10, 0, -TIGHTENGEARBOX-10, 0 },
#define MINSLOWELVEL (-0.1)
  { -0.0500,   -49, +TIGHTENGEARBOX-5,  0, -TIGHTENGEARBOX-5, 0 },

  { -0.0200,   -30, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0150,   -40, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0100,   -36, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0050,   -29, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0010,   -24, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { -0.0005,   -23, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  {  0.0000,   +10, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0005,   +40, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0010,   +41, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0050,   +43, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0100,   +49, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0150,   +50, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },
  { +0.0200,   +55, +TIGHTENGEARBOX, 0, -TIGHTENGEARBOX, 0 },

  { +0.0500,   +80, +TIGHTENGEARBOX-5,  0, -TIGHTENGEARBOX+5, 0 },
  { +0.1000,  +115, +TIGHTENGEARBOX-10, 0, -TIGHTENGEARBOX+10, 0 },
#define MAXSLOWELVEL (+0.1)
  { +0.8600,  +700, +TIGHTENGEARBOX-20, 0, -TIGHTENGEARBOX+20, 0 },
  { +1.2000,  +980, +TIGHTENGEARBOX-30, 0, -TIGHTENGEARBOX+30, 0 },
  { +9.0000,  +983, +TIGHTENGEARBOX-30, 0, -TIGHTENGEARBOX+30, 0 }
 #endif
#endif
};  // elSpeedPWM

/*------------------------------------------------------------*/
// Microsecond time handling.
#define LINUX_TIME_AVAIL 0
#include "ctime.cpp"



// Facilitating log file writing from classes, too.
FILE *vLogFile;



/*------------------------------------------------------------*/
// Speed-to-PWM conversion class with max acceleration.
class cSpeedPWM {
private:
  ptSpeedPWM st;  // (constant) speed table
  int nst;  // number of entries in speed table
  // Last interpolation results.
  int m1PWM;
  int m2PWM;
  // The place for a copy of newly-set PWM values.
  // Maximum acceleration management for speeds changed with this class.
  double desiredSpeed;
  double speedAtPWMs;  // latest speed setting at PWMs/motors
  tTime prevPWMChange;  // when we changed PWMs the last time
  // The place for a copy of newly-set PWM values.
  int *pm1PWM;
  int *pm2PWM;
protected:
  // This is used to get current (microsecond) timestamps.
  virtual void getTime(tTime& t);
  // Interpolate first with this...
  virtual void interpolatePWM(double speed);
  // ...and get access to PWM values with these after interpolating.
  virtual int getm1PWM(void);
  virtual int getm2PWM(void);
public:
  virtual void init(
    ptSpeedPWM speedTable,
    int numOfSpeeds,
    int *pPWM1,
    int *pPWM2
  );
  // Schedule changing PWMs not violating the maximum acceleration.
  virtual void setSpeed(double speed);
  // Is there a need to run 'run()'?
  virtual int isSpeedChangePending(void);
  // Actually change PWMs by one step of (remaining) speed change.
  virtual void run(double tacho1Speed, double tacho2Speed);
};  // cSpeedPWM

void cSpeedPWM::init(
  ptSpeedPWM speedTable,
  int numOfSpeeds,
  int *pPWM1,
  int *pPWM2
) {
  this->st = speedTable;
  this->nst = numOfSpeeds;
  this->m1PWM = this->m2PWM = 0;
  this->desiredSpeed = 0.0;
  this->speedAtPWMs = 0.0;
  this->getTime(prevPWMChange);
  this->pm1PWM = pPWM1;
  this->pm2PWM = pPWM2;
}  // init

void cSpeedPWM::interpolatePWM(double speed) {
  int i;
  int fi = -1;
  int se = -1;

  // Search for two lines in speed table,
  // the first <= speed, and the second >= speed.
  for (i=0; i < this->nst; i++) {
    if (this->st[i].speed >= speed) {
      se = i;
      break;
    }
  }  // for

  // The second line must be in table lines 1..(nst-1) (not 0..(nst-2)),
  // because we need the first line, too.
  if (!( (se >= 1) && (se < this->nst) )) {
    fprintf(stderr, "interpolatePWM(): speed = %f not found at se = %d\n",
      speed, se);
    abort();
  }
  fi = se - 1;
  assert( this->st[fi].speed <= speed );

  // Interpolate the PWM values between those two table entries.
  {
    double coeff;
    int fim1PWM, sem1PWM;
    int fim2PWM, sem2PWM;

    coeff = (speed - this->st[fi].speed)
      / (this->st[se].speed - this->st[fi].speed);
    fim1PWM = this->st[fi].mPWM + this->st[fi].m1tight;
    fim2PWM = this->st[fi].mPWM + this->st[fi].m2tight;
    sem1PWM = this->st[se].mPWM + this->st[se].m1tight;
    sem2PWM = this->st[se].mPWM + this->st[se].m2tight;
    this->m1PWM = fim1PWM + coeff * (double)(sem1PWM - fim1PWM);
    this->m2PWM = fim2PWM + coeff * (double)(sem2PWM - fim2PWM);
  }
  // xxx: antiphase?
}  // interpolatePWM

int cSpeedPWM::getm1PWM(void) {
  return (this->m1PWM);
}

int cSpeedPWM::getm2PWM(void) {
  return (this->m2PWM);
}

// Schedule changing PWMs not violating the maximum acceleration
// and not violating the "forced slow speed" areas ("slow limit switches").
void cSpeedPWM::setSpeed(double speed) {
  if ((speed < -MAXVEL) || (speed > MAXVEL)) {
    fprintf(vLogFile, "cSpeedPWM::setSpeed(): outside hw range = %f\n", speed);
    speed = (speed < -MAXVEL) ? (-MAXVEL) : MAXVEL;
  }
  this->desiredSpeed = speed;
}

// Is there a need to run 'run()'?
int cSpeedPWM::isSpeedChangePending(void) {
  return (this->desiredSpeed != this->speedAtPWMs);
}

// Actually change PWMs by one step of (remaining) speed change.
void cSpeedPWM::run(double tacho1Speed, double tacho2Speed) {
  tTime now;
  tTime tdiff;
  double td;

  // Update "latest PWM change time" and get the time difference
  // how long ago PWMs were last changed (or attempted to change).
  this->getTime(now);
  subTime(now, this->prevPWMChange, tdiff);
  this->prevPWMChange = now;
  td = dsecs(tdiff);
  if ((td <= 0.0) || (td >= 1.0)) {
    fprintf(vLogFile, "cSpeedPWM::run(): time difference td outside 0..1: %f\n",
      td);
    // xxx:abort();  fix-fix
    td = 0.030;
  }

  // Determine the maximum "velocity change slice" allowed
  // by max acceleration.
  td = MAXACC * td;
  if (this->desiredSpeed > this->speedAtPWMs) {
    this->speedAtPWMs += td;
    if (this->desiredSpeed < this->speedAtPWMs) {  // went over
      this->speedAtPWMs = this->desiredSpeed;
    }
  } else if (this->desiredSpeed < this->speedAtPWMs) {
    this->speedAtPWMs -= td;
    if (this->desiredSpeed > this->speedAtPWMs) {  // went over
      this->speedAtPWMs = this->desiredSpeed;
    }
  } else {
    // Nothing, do not change PWMs.
    return;
  }

  // Give "boost" to "lagging" motors based on tachometer
  // information.
  {
    double boosted;

    boosted = this->speedAtPWMs;
    if (fabs(boosted) < MAXTACHOBOOSTVEL) {
      boosted += TACHOBOOSTGAIN * (this->speedAtPWMs - tacho1Speed);
    }
    this->interpolatePWM(boosted);
    if (this->pm1PWM) *(this->pm1PWM) = this->getm1PWM();

    boosted = this->speedAtPWMs;
    if (fabs(boosted) < MAXTACHOBOOSTVEL) {
      boosted += TACHOBOOSTGAIN * (this->speedAtPWMs - tacho2Speed);
    }
    this->interpolatePWM(boosted);
    if (this->pm2PWM) *(this->pm2PWM) = this->getm2PWM();
  }  // boosting
}  // run


/*------------------------------------------------------------*/
// Interpretation of 24-bit interlock switch chain.
#include "interloc.cpp"


/*------------------------------------------------------------*/
// Utilities.

int getVal(char *prompt) {
  char buff[255];
  int temp;

  printf("%s", prompt);
  gets(buff);
  (void)sscanf(buff, "%d", &temp);
  return (temp);
}

double getDouble(char *prompt) {
  char buff[255];
  double temp;

  printf("%s", prompt);
  gets(buff);
  (void)sscanf(buff, "%lg", &temp);
  return (temp);
}


/*------------------------------------------------------------*/
/* Definition of the struct of all sensor values on the serial line down. */

#include "downlink.h"

/*------------------------------------------------------------*/
// Class for any (double type) position which changes in time.
// (This results in measured velocity and acceleration.)
class cVelAcc {
private:
  // Minimum number of microseconds before new vel/acc estimates are
  // calculated.
  double minVelAccTime;

  int firstCall;  // 1==not inited before
  double pos;  // the latest position
  double *ppos;  // and a pointer to where update 'pos' whenever it changes
  tTime posTime;  // the timepoint when the latest position was updated
  double prevPos;  // the previous position
  tTime prevPosTime;  // the time when previous position was updated
    // (if position stayed the same, previous time stays too)
  double posTimeDiff;

  double vel;  // estimated most recent velocity
  double *pvel;  // and its pointer
//  tTime velTime;  // when the speed was the estimated
  double prevVel;  // the previous estimated velocity
//  tTime prevVelTime;  // when the previous velocity estimate was updated
//  double velTimeDiff;

  double acc;  // acceleration based on vel and prevVel
  double *pacc;  // and its pointer
protected:
  // This is used to get current (microsecond) timestamps.
  virtual void getTime(tTime& t);
  // "Replaceable" position difference routine.
  // (For QDR wrap-around figures.)
  virtual double posDiff(double p1, double p2);
public:
  // Constructor initializes speed/acceleration calculation.
  cVelAcc::cVelAcc();

  // Set pointers to where copy pos/vel/acc whenever they change.
  virtual void setPointers(
    double *pp, double *pv, double *pa
  );
  // Changing position coordinates stores the current timestamp
  // and recalculates velocity and acceleration.
  virtual void set(double p);

  virtual void resetSpeeds(void);

  virtual double get(void);
  virtual double getVel(void);
  virtual double getAcc(void);

  // Set minimum time between new velocity and acceleration estimates.
  virtual void setTime(double newMinTime);
};  // class cVelAcc

cVelAcc::cVelAcc() {
  this->firstCall = 1;
  this->minVelAccTime = 0.035;  // default 35msec, just > one loop
  this->ppos = this->pvel = this->pacc = NULL;
}  // constructor

void
cVelAcc::setPointers(
  double *pp, double *pv, double *pa
) {
  this->ppos = pp;
  this->pvel = pv;
  this->pacc = pa;
}  // constructor

// The default position difference function, plain subtraction.
double
cVelAcc::posDiff(double p1, double p2) {
  return (p1 - p2);
}  // posDiff

void
cVelAcc::set(double p) {
  tTime now;
  tTime td;
  double ptd;

  this->getTime(now);
  // First position set call.
  if (firstCall) {
    this->pos = this->prevPos = p;
    this->vel = this->prevVel = 0.0;
    this->acc = 0.0;
    this->posTime = this->prevPosTime = now;
    this->firstCall = 0;
    goto lPointerUpdate;
  }
  // At least one previous position measurement before.
  // Tentatively compare 'p' to 'pos', or actually only elapsed time.
  subTime(now, /* - */ this->posTime, td);
  ptd = dsecs(td);
  if (ptd <= this->minVelAccTime) {
    // Ignore this new point until later.
    // Hmm, should update '*ppos' anyhow, try it...
    if (this->ppos) *(this->ppos) = p;
    return;
  }
  // Long enough has elapsed, update new position.
  this->prevPos = this->pos;
  this->prevPosTime = this->posTime;
  this->pos = p;
  this->posTime = now;
  this->posTimeDiff = ptd;

  // Calculate new velocity and acceleration.
  this->prevVel = this->vel;
  this->vel = this->posDiff(this->pos, this->prevPos) / this->posTimeDiff;
  this->acc = (this->vel - this->prevVel) / this->posTimeDiff;

lPointerUpdate:
  if (this->ppos) *(this->ppos) = this->pos;
  if (this->pvel) *(this->pvel) = this->vel;
  if (this->pacc) *(this->pacc) = this->acc;
}  // set

void
cVelAcc::setTime(double newMinTime) {
  this->minVelAccTime = newMinTime;
}

void
cVelAcc::resetSpeeds(void) {
  this->firstCall = 1;
  // Next 'set()' will result in zero speeds etc.
}  // resetSpeeds

double cVelAcc::get(void) { return (this->pos); }
double cVelAcc::getVel(void) { return (this->vel); }
double cVelAcc::getAcc(void) { return (this->acc); }


class cQDRVelAcc : public cVelAcc {
protected:
  double toDegCoeff;
public:
  // Constructor initializes toDegCoeff.
  cQDRVelAcc::cQDRVelAcc();
  // Setting the tacho count --> degrees coefficient;
  virtual void setCoeff(double coeff);
protected:
  // "Replaceable" position difference routine.
  // For QDR wrap-around figures, wraps in range 0..65535.
  virtual double posDiff(double p1, double p2);
};  // cQDRVelAcc

cQDRVelAcc::cQDRVelAcc() {
  //// cVelAcc::cVelAcc();
  this->toDegCoeff = 1.0;
}  // constructor

void
cQDRVelAcc::setCoeff(double coeff) {
  this->toDegCoeff = coeff;
}  // setCoeff

// The QDR 0..65535 position difference function.
// We map to degrees, i.e. 360deg == 10000 tacho counts via 18187:1 gear
// to main axis.
double
cQDRVelAcc::posDiff(double p1, double p2) {
  double d = p1 - p2;

  if (d > 32767.0) {
    d -= 65536.0;
  } else if (d < -32768.0) {
    d += 65536.0;
  }
  return (this->toDegCoeff * d);
}  // posDiff


/*------------------------------------------------------------*/
// The complete and common shared struct
// across ant/dos/up/ant.cpp -- antcond -- xantcon.
//
// ant.cpp <--> antcond is transmitted field-by-field using COM2 serial,
// antcond <--> xantcon uses shared memory and the same struct layout.

// Per motor / FET board settings and measurements.
typedef struct sMotor {
  int pwm;  // as in cPWMISA, -1023..1023
  int antiph;  // as in cPWMISA, -1023..1023, only 16 real steps
  unsigned int qdr;  // latest QDR tacho value
  double qdrVel;  // calculated QDR velocity, counts/s
  double qdrAcc;  // acceleration, counts/s
  double vsupply;  // power supply of this motor
  double icurr;  // latest current measurement of this motor
  double temp;  // temperature of FET board
} tMotor, *ptMotor;

/* ---------- */
// The struct :-)
typedef struct sSh {
  // Current time, UTC in Unix 2xlong format, and m/usecs since start.
  tTime utc;
  unsigned long msecs;
  // xxx: sidereal time?

  // ----- Pedestal PC hardware-related:
  // Interlock chain bits.
  tInterlockBits interlock;

  // Angle encoders, all in degrees, az=-180..180 "old" convention.
  tCoord abs;
  tCoord incr;
  tCoord tilt;

  // Per motor / FET board settings and measurements.
  union {
    tMotor motors[5];
    struct {
      tMotor az1;
      tMotor az2;
      tMotor el1;
      tMotor el2;
      tMotor subr;
    } motor;
  };

  // Brakes (brake Ibrake currents).
  struct {
    double az;
    double el;
  } ibrake;

  // Supply voltages from power supplies.
  struct {
    double v5;
    double v12;
    double v24;
    double vn15;
    double v15;
    double vgnd;
  } volt;

  // Temperatures.
  struct {
    double tilt;
    double racklow;
    double rackmiddle;
    double rackpc;
  } temp;

  // ----- ant/dos/up/ant.cpp -related:
  // Azimuth system used, 0.0 for "traditional" Metsahovi -180..180
  // and 180.0 for international 0..360 azimuth.
  int azel;  // 1== Az first, El then; 0==El first, Az then
  double azsys;  // 0.0 or 180.0
  // Add 'azsys' to all internal azimuths before displaying to the user,
  // and subtract it before accepting azimuths from the user.

  // Current antenna position (==incr), angular velocity, and acceleration.
  tCoord curr;
  tCoord currSp;
  tCoord currAcc;

  // The latest uplink message time from downstairs 'antcond',
  // i.e. xntpd time.
  tTime xntpdTime;

  // Current "momentary" (==for this loop run) target position.
  tCoord goalPos;
  tCoord goalVel;
  tCoord goalAcc;

  // Manual/keyed in main antenna angular speed.
  tCoord goalSpeed;

} tSh, *ptSh;

// Global shared struct.
tSh vSh;
// And a pointer to it, for shared memory references in Linux with 'sh->'.
ptSh sh = &vSh;

// Time handling; candidates for 'sh' struct.
time_t ltime;
long v1ppsExtraPulses = 0L;
long v1ppsMissingPulses = 0L;
unsigned long microseconds = 0;
unsigned long prevUsecs = 0;
unsigned long ms_since_init = 0;
unsigned long s_since_init = 0;
unsigned long prevLoopUsecs = 0;

// For tracking uplink performance:
// shut down if nothing is heard from downstairs.
unsigned long vLatestSerialActivity;
unsigned long vLatestDownstairsDelay;


void
cVelAcc::getTime(tTime& t) {
  t.tv_sec = ltime;
  t.tv_usec = microseconds;
}  // getTime

void
cSpeedPWM::getTime(tTime& t) {
  t.tv_sec = ltime;
  t.tv_usec = microseconds;
}  // getTime


// Serial port handling.
#include "serial.h"
unsigned long vSerialReadErrors = 0L;


/*------------------------------------------------------------*/
int main(void)
{
  // Four 50-pin parallel connectors.
#if SUNANT
  class cPortSSI currAbsAz;
  class cPortSSI currAbsEl;
#else
  class cPortROC412 currAbsAz;
  class cPortROC412 currAbsEl;
#endif
  class cQDR vTachos;
  class c24In vInterlockChain;

#if SUNANT
  // Virtual abs+tacho combining "incremental" encoders for SunAnt.
  class cIncrTacho vAzIncrTacho;
  class cIncrTacho vElIncrTacho;
#else
  // Heidenhain incremental encoder board.
  class cIK121 vIK121;
#endif

  // 32-ch A/D converter with its channels.
  class cET28xx vACL8113;
  class cSensorChannel
        vAz1ICurrCh, vAz2ICurrCh, vEl1ICurrCh, vEl2ICurrCh, vSubrICurrCh;
  class cSensorChannel
        vTRackUpperCh,
        vTRackLowerCh,
        vTPCInsideCh,
        vAz1TBoardCh,
        vAz2TBoardCh,
        vEl1TBoardCh,
        vEl2TBoardCh,
        vSubrTBoardCh,
        vSupply5VCh,
        vSupply12VCh,
        vSupply24VCh,
        vAz1SupplyCh,
        vAz2SupplyCh,
        vEl1SupplyCh,
        vEl2SupplyCh,
        vSubrSupplyCh,
        vSupplyN15VCh,
        vSupply15VCh,
        vSupplyGNDCh,
        vIBrakeAzCh,
        vIBrakeElCh;
  class cSensorChannel
        vTiltYAzCh, vTiltXElCh, vTiltTempCh;

  // Tables for angular speed --> PWM settings.
  class cSpeedPWM vSpAz;
  class cSpeedPWM vSpEl;



#define AZAXIS 0
#define ELAXIS 1
#define SUBRAXIS 2
#define SPEEDAXIS 3
  int vWhichAxis;  // AZ or EL or subreflector, active with '>' cursor
  int vWhichMotor;

  // Antenna position, speed, and acceleration calculator objects.
  class cVelAcc vAz, vEl;
  unsigned long currAzElTime = 0;
  unsigned long prevAzElTime = 0;
  class cQDRVelAcc vQDRTachos[5];  // tachometer pos/vel/acc calc objs.
  class cVelAcc vGoalAz, vGoalEl;

  // "Stress test" variables.
  int vStressTestOn;
  int vServoingIsOn;
  int vLoggingEnabled;
  unsigned long vSTNextChangeTime;
#if 0
// "Old" az-el-"mapping" scanning of angular speeds.
#define cSTInterval 2000  /* milliseconds */
#define cSTIncrement 10
#define MIN_TEST 075
#define MAX_TEST 175
#endif

#if 1
#define cSTInterval 1000  /* milliseconds */
// Test goes 0..hi..0..-hi..0... (plus TIGHTENGEARBOX).
#define cSTSlowPWM 70
#define cSTSlowIncr 1
#define cSTMedPWM 100
#define cSTMedIncr 2
#define cSTHiPWM 120
#define cSTHiIncr 10
int vSTPhase;  // +1 incr, -1 decr
int vSTPWM;
#endif

  // Init one board at default base address.
  vPWMISA.init();
  tzset();
  time(&ltime);
  realtimeInstall();
  vPWMISA.setBasePWMFreq(PWMBASE10KHZ);

#if SUNANT
  // Init ECN113 absolute SSI encoders via parallel interface board.
  currAbsAz.init(
	/*baseAddr=>*/ 0x220,
	/*conn=>*/ 0,
	/*numOfBitsMask=>*/ 0x1000
  );
  currAbsAz.setCalibration(/*offset=>*/
    /* Measured 03-Aug-2000 with a compass> */
     -173.0 - 6.154297
    , /*coeff=>*/ 1.0);

  currAbsEl.init(
	/*baseAddr=>*/ 0x220,
	/*conn=>*/ 1,
	/*numOfBitsMask=>*/ 0x1000
  );
  currAbsEl.setCalibration(/*offset=>*/
    /* Measured 03-Aug-2000 with a level: */
     -30.0 - 5.815430

     + 0.65  /* measured 14-Aug-2000 with manual elevation offset */

     /* 05-Jan-2001 noticed that the following line was in
        large antenna part of code; moved here for doc. purposes... */
     /* - 0.65  measured 10-Aug-2000 */

    , /*coeff=>*/ 1.0);
#else
  // Init ROC412 absolute encoders via parallel interface board.
  currAbsAz.init(
	/*baseAddr=>*/ 0x220,
	/*conn=>*/ 0,
	/*numOfBitsMask=>*/ 0x0800
  );
  currAbsAz.setCalibration(/*offset=>*/ 540.2062, /*coeff=>*/ -3.000867);
    /* Measured on 04-Sep-1996 with ant. range -360..0. */

  currAbsEl.init(
	/*baseAddr=>*/ 0x220,
	/*conn=>*/ 1,
	/*numOfBitsMask=>*/ 0x0800
  );
  currAbsEl.setCalibration(/*offset=>*/ +185.24, /*coeff=>*/ -1.0);
    /* xxx: not measured yet */
#endif

  // Init QDR tachometer interface board 24-bit connector.
  vTachos.init(
    /*baseAddr=>*/ 0x220,
    /*conn=>*/ 2,
    /*addressing=>*/ 0  // normal 4 regs per 8255, not indirect
  );

  // Init interlock chain opto board 24-bit connector.
  vInterlockChain.init(
    /*baseAddr=>*/ 0x220,
    /*conn=>*/ 3,
    /*addressing=>*/ 0  // normal 4 regs per 8255, not indirect
  );

  // Init subreflector motor selector switcher relay board 'antrel8'.
  // xxx
  // /*conn=>*/ 4

  // /*conn=>*/ 5  not used yet.

#if SUNANT
  //// 1 turn == 159(az motor) * 3045(cyclo) * 500slots * 4pulses/slot
  // #define AZTACHOCOEFF (360.0 / (159.0 * 3045.0) / 2000.0)
  // 1 turn == 66(az motor) * 3045(cyclo) * 500slots * 4pulses/slot
  #define AZTACHOCOEFF (360.0 / (66.0 * 3045.0) / 2000.0)
  // 1 turn == 246(el motor) * 3045(cyclo) * 500slots * 4pulses/slot
  #define ELTACHOCOEFF (360.0 / (246.0 * 3045.0) / 2000.0)
  // Tachometers provide more resolution to absolute encoders,
  // but not more than one absolute encoder bit/slot.
  vAzIncrTacho.init(AZTACHOCOEFF, currAbsAz.getBinToDegreesCoeff());
  vElIncrTacho.init(ELTACHOCOEFF, currAbsEl.getBinToDegreesCoeff());
#else
  // Incremental encoders.
  vIK121.init(/*baseAddr=>*/ 0x320);
#endif

  // Init 32-ch 12-bit A/D converter.
  vACL8113.init(0x260, /*boardModel=>*/ ACL8113, /*numOfBitsMask=>*/ 0x0fff,
                /*loopLimit=>*/ 70
               );
  vACL8113.setCalibration(/*offset=>*/ 0.0, /*coeff=>*/ 1.0);
    /* calibration of A/D board itself */

  // Init individual A/D channels.
  vAz1ICurrCh.init(&vACL8113);
  vAz1ICurrCh.setCalibration(
    /*offset=>*/ 0.0,
    /*coeff=>*/ 2.5
  );
  vAz2ICurrCh.init(&vACL8113);
  vAz2ICurrCh.setCalibration(
    /*offset=>*/ 0.0,
    /*coeff=>*/ 2.5
  );
  vEl1ICurrCh.init(&vACL8113);
  vEl1ICurrCh.setCalibration(
    /*offset=>*/ 0.0,
    /*coeff=>*/ 2.5
  );
  vEl2ICurrCh.init(&vACL8113);
  vEl2ICurrCh.setCalibration(
    /*offset=>*/ 0.0,
    /*coeff=>*/ 2.5
  );
  vSubrICurrCh.init(&vACL8113);
  vSubrICurrCh.setCalibration(
    /*offset=>*/ 0.0,
    /*coeff=>*/ 2.5
  );
#define INITCH(ch, off, coeff) \
  ch##Ch.init(&vACL8113); \
  ch##Ch.setCalibration( \
    /*offset=>*/ off, \
    /*coeff=>*/ coeff \
  )
  INITCH(vTRackUpper, 0.0, 100.0);
  INITCH(vTRackLower, 0.0, 100.0);
  INITCH(vTPCInside, 0.0, 100.0);
  INITCH(vAz1TBoard, 0.0, 100.0);
  INITCH(vAz2TBoard, 0.0, 100.0);
  INITCH(vEl1TBoard, 0.0, 100.0);
  INITCH(vEl2TBoard, 0.0, 100.0);
  INITCH(vSubrTBoard, 0.0, 100.0);
  INITCH(vSupply5V, 0.0, 1.0);
  INITCH(vSupply12V, 0.0, 2.0);
#if SUNANT
  INITCH(vSupply24V, 0.0, 3.0);
#else
  INITCH(vSupply24V, 0.0, 4.0);
#endif
  INITCH(vAz1Supply, 0.0, 4.33);  // (33.3k+10k)/(10k)
  INITCH(vAz2Supply, 0.0, 4.33);
  INITCH(vEl1Supply, 0.0, 4.33);
  INITCH(vEl2Supply, 0.0, 4.33);
  INITCH(vSubrSupply, 0.0, 4.33);
  INITCH(vSupplyN15V, 0.0, 2.0);
  INITCH(vSupply15V, 0.0, 2.0);
  INITCH(vSupplyGND, 0.0, 1.0);
  INITCH(vIBrakeAz, 0.0, 1.0);
  INITCH(vIBrakeEl, 0.0, 1.0);

  /* Tiltmeter channels. */
  /* With low gain the calibrated range is +/-2000uradians */
  /* i.e. +/-0.115degrees (full scale about +/-4000urad, +/-0.23degrees). */
  INITCH(vTiltYAz, /*offset=>*/ 0.0,
    /*coeff=>*/ 0.4833 /* low gain; high=0.05002 uradians / mV */
                 * 180.0 / M_PI  /* millidegrees / V */
                 / 1000.0  /* millidegrees -> degrees */
  );
  INITCH(vTiltXEl, /*offset=>*/ 0.0,
    /*coeff=>*/ 0.4740 /* low gain; high=0.04991 */
                 * 180.0 / M_PI  /* millidegrees / V */
                 / 1000.0  /* millidegrees -> degrees */
  );
  INITCH(vTiltTemp, /*offset=>*/ 0.0, /*coeff=>*/ 100.0);

  // Init motor speed tables / max accel PWM changers.
  vSpAz.init(azSpeedPWM, (sizeof(azSpeedPWM) / sizeof(tSpeedPWM)),
    &(sh->motor.az1.pwm), &(sh->motor.az2.pwm)
  );
  vSpEl.init(elSpeedPWM, (sizeof(elSpeedPWM) / sizeof(tSpeedPWM)),
    &(sh->motor.el1.pwm), &(sh->motor.el2.pwm)
  );

  /* Serial link from pedestal up PC to control room antcon. */
  OpenSerial(COM2, 115200L, D8S1N, NULL);
  // OpenSerial(COM2, 9600L, D8S1N, NULL);

  // Get incremental encoder values. xxx
  sh->curr.az = 0.0;
  sh->curr.el = 0.0;
  vAz.setPointers(&(sh->curr.az), &(sh->currSp.az), &(sh->currAcc.az));
  vEl.setPointers(&(sh->curr.el), &(sh->currSp.el), &(sh->currAcc.el));
  currAzElTime = ms_since_init;
  prevAzElTime = currAzElTime;

  // Set pos/vel/acc calculator objects.
#if SUNANT
  vQDRTachos[0].setCoeff(AZTACHOCOEFF);
  vQDRTachos[1].setCoeff(1.0);
  vQDRTachos[2].setCoeff(ELTACHOCOEFF);
  vQDRTachos[3].setCoeff(1.0);
  vQDRTachos[4].setCoeff(1.0);
#else
  // 1 turn == 18187(gear) * 2500slots * 4pulses/slot
  vQDRTachos[0].setCoeff(360.0 / 18187.0 / 10000.0 );
  vQDRTachos[1].setCoeff(360.0 / 18187.0 / 10000.0 );
  vQDRTachos[2].setCoeff(360.0 / 18187.0 / 10000.0 );
  vQDRTachos[3].setCoeff(360.0 / 18187.0 / 10000.0 );
  vQDRTachos[4].setCoeff(1.0);
#endif
  vQDRTachos[0].setPointers(NULL, &(sh->motors[0].qdrVel), &(sh->motors[0].qdrAcc));
  vQDRTachos[1].setPointers(NULL, &(sh->motors[1].qdrVel), &(sh->motors[1].qdrAcc));
  vQDRTachos[2].setPointers(NULL, &(sh->motors[2].qdrVel), &(sh->motors[2].qdrAcc));
  vQDRTachos[3].setPointers(NULL, &(sh->motors[3].qdrVel), &(sh->motors[3].qdrAcc));
  vQDRTachos[4].setPointers(NULL, &(sh->motors[4].qdrVel), &(sh->motors[4].qdrAcc));
  vGoalAz.setPointers(&(sh->goalPos.az), &(sh->goalVel.az), &(sh->goalAcc.az));
  vGoalEl.setPointers(&(sh->goalPos.el), &(sh->goalVel.el), &(sh->goalAcc.el));
  // Slow down goal/source/target coordinate velocity estimates.
  vGoalAz.setTime(0.21);  // VAX compatibility...
  vGoalEl.setTime(0.21);

#if 0
  // Test pos/vel/acc class.
  {
    int i;
    for (i=0; i<10; i++) {
      vAz.set((double)(i/3));
    }
  }
goto lExit;
#endif

  /* Open log file always in append mode. */
  assert ( (vLogFile = fopen("ant.log", "at")) != NULL );  // append in text mode
  fprintf(vLogFile,
    "\n\n----------\nGMT is %sMilliseconds since program start %010lu\n---\n",
    asctime(gmtime(&ltime)),
    ms_since_init
  );

#define MIN_PWM_OFF 0
  sh->motor.az1.pwm = sh->motor.az2.pwm = MIN_PWM_OFF;
  sh->motor.el1.pwm = sh->motor.el2.pwm = MIN_PWM_OFF;
  sh->motor.subr.pwm = MIN_PWM_OFF;
  sh->xntpdTime.tv_sec = sh->xntpdTime.tv_usec = 0;
  sh->goalSpeed.az = sh->goalSpeed.el = 0.0;
  vLatestSerialActivity = ms_since_init;

  vWhichAxis = AZAXIS;
  vWhichMotor = 0;
  vStressTestOn = 0;  vSTPhase = 0;
  vServoingIsOn = 0;
  vLoggingEnabled = 0;
  vSTNextChangeTime = ms_since_init;

  clrscr();
  _setcursortype(_NOCURSOR);
  directvideo = 1;
#define SCRPRINTF cprintf
  SCRPRINTF("$Id: ant.cpp,v 1.41 2005/11/09 11:49:58 lindi Exp $");

  /* ---------- main for-ever loop */
  for(;;) {

    // Measure sensors.
#if AD_CURRENT_MEASUREMENTS
#define GETCHN(var, name, ch, gain) \
    vACL8113.setChannel(/*channel=>*/ ch, SE, /*gain=>*/ gain); \
    var = name##Ch.get()
#define GETCH(name, ch, gain) GETCHN(name, name, ch, gain)
#define GETCHS(field, name, ch, gain) GETCHN(sh->field, name, ch, gain)

    // Process all A/D converter channels.
    GETCHS(motor.az2.icurr,   vAz2ICurr,  0, 2);
    GETCHS(motor.az2.vsupply, vAz2Supply, 1, 1);
    GETCHS(motor.az1.icurr,   vAz1ICurr,  2, 2);
    GETCHS(motor.az1.vsupply, vAz1Supply, 3, 1);
    GETCHS(motor.el2.icurr,   vEl2ICurr,  4, 2);
    GETCHS(motor.el2.vsupply, vEl2Supply, 5, 1);
    GETCHS(motor.el1.icurr,   vEl1ICurr,  6, 2);
    GETCHS(motor.el1.vsupply, vEl1Supply, 7, 1);
    GETCHS(motor.subr.icurr,  vSubrICurr,  8, 2);
    GETCHS(motor.subr.vsupply,vSubrSupply, 9, 1);
    GETCHS(tilt.el,     vTiltXEl,   10, 1);
    GETCHS(tilt.az,     vTiltYAz,   11, 1);
    GETCHS(temp.tilt,   vTiltTemp,  12, 1);
    GETCHS(ibrake.az,   vIBrakeAz,  14, 1);
    GETCHS(ibrake.el,   vIBrakeEl,  15, 1);
    GETCHS(volt.v5,     vSupply5V,  16, 1);
    GETCHS(volt.v12,    vSupply12V, 17, 1);
    GETCHS(volt.v24,    vSupply24V, 18, 1);
    GETCHS(volt.vgnd,   vSupplyGND, 20, 8);
    GETCHS(motor.az1.temp,    vAz1TBoard, 22, 2);
    GETCHS(motor.az2.temp,    vAz2TBoard, 23, 2);
    GETCHS(motor.el1.temp,    vEl1TBoard, 24, 2);
    GETCHS(motor.el2.temp,    vEl2TBoard, 25, 2);
    GETCHS(motor.subr.temp, vSubrTBoard, 26, 2);
    GETCHS(temp.rackmiddle, vTRackUpper, 27, 2);
    GETCHS(temp.racklow,    vTRackLower, 28, 2);
    GETCHS(temp.rackpc,     vTPCInside,  29, 2);
    GETCHS(volt.v15,        vSupply15V,  30, 1);
    GETCHS(volt.vn15,       vSupplyN15V, 31, 1);
#endif

    // Get incremental and absolute encoder values.
    sh->interlock = vInterlockChain.get();

    sh->abs.az = currAbsAz.get();
    sh->abs.el = currAbsEl.get();
#if DEBUG_REPEATS
    if ( (sh->abs.az < MINSLOWAZ) || (sh->abs.az > MAXSLOWAZ) ) {
      fprintf(vLogFile, " - bad abs az = %.6f\n", sh->abs.az);
    }
    // (MINSLOWEL - 4) is ok for both large and SunAnt:
    if ( (sh->abs.el < (MINSLOWEL - 4.0)) || (sh->abs.el > MAXSLOWEL) ) {
      fprintf(vLogFile, " - bad abs el = %.6f\n", sh->abs.el);
    }
#endif

    // Get motor tachometers.
#if SUNANT
    // xxx: Invert el tacho (mechanically opposite).
    vQDRTachos[0].set( (sh->motor.az1.qdr = vTachos.get(0)) );
    vQDRTachos[1].set( (sh->motor.az2.qdr = 0) );
    vQDRTachos[2].set( (sh->motor.el1.qdr = vTachos.get(2)) );
    vQDRTachos[3].set( (sh->motor.el2.qdr = 0) );
#else
    // (Invert az tachos, as mechanically az gears turn in opposite dir.)
    vQDRTachos[0].set( (sh->motor.az1.qdr = -vTachos.get(0)) );
    vQDRTachos[1].set( (sh->motor.az2.qdr = -vTachos.get(1)) );
    vQDRTachos[2].set( (sh->motor.el1.qdr = vTachos.get(2)) );
    vQDRTachos[3].set( (sh->motor.el2.qdr = vTachos.get(3)) );
#endif
    vQDRTachos[4].set( (sh->motor.subr.qdr = 0) );
      // no tacho yet for subreflector motor

    // For SunAnt, incremental values are fabricated
    // by combining absolute encoders and motor tachometers.
    // For main antenna the incremental Heidenhain encoders are used.
#if SUNANT
    vAzIncrTacho.set(sh->abs.az, sh->motor.az1.qdr);
    vElIncrTacho.set(sh->abs.el, sh->motor.el1.qdr);
 #if 1
    sh->incr.az = vAzIncrTacho.get();
    sh->incr.el = vElIncrTacho.get();
 #else
    sh->incr.az = sh->abs.az;
    sh->incr.el = sh->abs.el;
 #endif
#else
    vIK121.trigger();
    sh->incr.az = (double)vIK121.get(/*axis=>*/ 0) / 102400.0
                  + 1.11 - 0.022 /* measured 29-Sep-1996 */;
    sh->incr.el = -(double)vIK121.get(/*axis=>*/ 1) / 102400.0
                  + +42.516475 + 0.035 /* measured 29-Sep-1996 */
                  ;
#endif
    /* xxx: sh->time = */ currAzElTime = ms_since_init;
    vAz.set(sh->incr.az);  // updates pos/vel/acc vars via pointers
    vEl.set(sh->incr.el);  // updates pos/vel/acc vars via pointers
    // xxx: get time of this measurement as timestamp for this loop pass.

    /* Serial line command actions. ----- */
    /* Test for incoming serial characters. */
    if (BufLen(COM2) > 0) {
      unsigned char c = (unsigned char)ReadSerial(COM2);
      union {
        char line[81];
        tDownlink downlink;
        tUplink uplink;
      };
      int len;

      // Record the timestamp when the latest serial message arrived.
      // ('ms_since_init' keeps incrementing regardless of
      //  how we get donwstairs xntpd time.)
      vLatestSerialActivity = ms_since_init;

      // Command '*' arrived, means we need to send data down
      // to control room 'antcond', "downlink".
      switch (c) {
        case '*': {
          // xxx: send down our conception of the UTC timepoint
          // when the incremental encoders were read.
          // downlink.time = sh->incrTime;
          downlink.abs.az = sh->abs.az;
          downlink.abs.el = sh->abs.el;
          downlink.incr.az = sh->incr.az;
          downlink.incr.el = sh->incr.el;
          downlink.interlock = sh->interlock;
          /* 'downlink' is layered on top of 'line'. */
          len = downlink.length = sizeof(downlink);
          break;
        }

        // Command '+' arrived, we are about to receive
        // command & coordinates (or other data values) from
        // downstairs 'antcond', "uplink".
        case '+': {
          int i;

          for (i=0; i<sizeof(uplink); i++) {
            unsigned int timeout = 50000;

            while ((BufLen(COM2) <= 0) && (--timeout)) {
              // Nothing.
            }
            if (timeout == 0) {
              vSerialReadErrors++;
            }
            line[i] = ReadSerial(COM2);
          }  // for
          if (uplink.length == sizeof(uplink)) {
            int m;

            // Check UTC time, especially seconds.
            sh->xntpdTime = uplink.time;

            // Enable motors by default.
            for (m=0; m<NUMOFMOTORS; m++) {
  	          vPWMISA.setEnable(m, 1);
  	          vPWMISA.setRelay(m, 1);
            }

            // Interpret command.
            if (uplink.command == CC_track_position) {
              // Force servo mode.
              vServoingIsOn = 1;
              vGoalAz.set(uplink.cc.az);  // sets 'sh->goalPos' via ptr
              vGoalEl.set(uplink.cc.el);
            } else if (uplink.command == CC_manual_velocity) {
              // Force non-servo mode.
              vServoingIsOn = 0;
              // Update PWMs directly.
              sh->goalSpeed.az = uplink.cc.az;
              if ((sh->incr.az < MINSLOWAZ) || (sh->incr.az > MAXSLOWAZ)) {
                 if (sh->goalSpeed.az < MINSLOWAZVEL) {
                    sh->goalSpeed.az = MINSLOWAZVEL;
                 }
                 if (sh->goalSpeed.az > MAXSLOWAZVEL) {
                    sh->goalSpeed.az = MAXSLOWAZVEL;
                 }
              }
              vSpAz.setSpeed(sh->goalSpeed.az);

              sh->goalSpeed.el = uplink.cc.el;
              if ((sh->incr.el < MINSLOWEL) || (sh->incr.el > MAXSLOWEL)) {
                 if (sh->goalSpeed.el < MINSLOWELVEL) {
                    sh->goalSpeed.el = MINSLOWELVEL;
                 }
                 if (sh->goalSpeed.el > MAXSLOWELVEL) {
                    sh->goalSpeed.el = MAXSLOWELVEL;
                 }
              }
              vSpEl.setSpeed(sh->goalSpeed.el);
            } else if (uplink.command == CC_manual_azeltighten_pwm) {
              // Force non-servo mode.
              vServoingIsOn = 0;
              sh->motor.az1.pwm = (int)uplink.cc.az + TIGHTENGEARBOX;
              sh->motor.az2.pwm = (int)uplink.cc.az - TIGHTENGEARBOX;
              sh->motor.el1.pwm = (int)uplink.cc.el + TIGHTENGEARBOX;
              sh->motor.el2.pwm = (int)uplink.cc.el - TIGHTENGEARBOX;
            } else if (uplink.command == CC_manual_az_pwm) {
              // Force non-servo mode.
              vServoingIsOn = 0;
              sh->motor.az1.pwm = (int)uplink.cc.az;
              sh->motor.az2.pwm = (int)uplink.cc.el;
            } else if (uplink.command == CC_manual_el_pwm) {
              // Force non-servo mode.
              vServoingIsOn = 0;
              sh->motor.el1.pwm = (int)uplink.cc.az;
              sh->motor.el2.pwm = (int)uplink.cc.el;
#if SUNANT
            } else if (uplink.command == CC_brake_standby) {
              // xxx: temporarily do nothing
              // this allows tachometer speed measurements
              // with downstairs providing the clock...
#endif
            } else {
              // if (uplink.command == CC_brake_standby) {
              // xxx: command not recognized,
              // it is formally correct but unknown => standby

              // Disable everything, including servo.
              vServoingIsOn = 0;
              vPWMISA.disableAll();
            }
          } else { // if uplink receive error
            vSerialReadErrors++;
          }
          len = 0;
          break;
        }  // case '+'
#if SUNANT
#else
        case 'r': {
          //sound(600);
          //delay(500);
          //nosound();
          vIK121.startCountingFromZeroAtRefMark();
          len = 0;
          break;
        }
        case 'l': {
          vIK121.latchAtNextRefMark();
          len = 0;
          break;
        }
#endif
        default: {
          vSerialReadErrors++;
          // len = sprintf(line, "%c -> unrecognized command letter, try 'a/0/1/2/3'\r\n", c);
          len = 0;
          break;
        }
      }  /* switch */

      /* Send a response down. */
      if (len > 0) {
        WriteStrSerial(COM2, (unsigned char *)line, len);
      }
      /* The other end is not supposed to send new commands
         before receiving a previous response, make sure by: */
      // xxx: Hmm, now with '+' and '*', a '*' may follow
      // quite soon after '+'...
      // FlushInBuf(COM2);
    }  /* if serial characters have arrived */
    /* ----- */

    // Keyboard actions.
	if (kbhit()) {
	  switch (getch()) {
	    case '0': {
#if SUNANT
#else
          vIK121.resetAmplStatus(0);
          vIK121.resetAmplStatus(1);
#endif
          vPWMISA.resetStatus(/*motor=>*/0);
          vPWMISA.resetStatus(/*motor=>*/1);
          vPWMISA.resetStatus(/*motor=>*/2);
          vPWMISA.resetStatus(/*motor=>*/3);
          vPWMISA.resetStatus(/*motor=>*/4);
		  break;
        }
	    case '1':
          vWhichAxis = AZAXIS;
	      vWhichMotor = 0;
		  break;
	    case '2':
          vWhichAxis = AZAXIS;
	      vWhichMotor = 1;
	      break;
	    case '3':
          vWhichAxis = ELAXIS;
	      vWhichMotor = 2;
		  break;
	    case '4':
          vWhichAxis = ELAXIS;
	      vWhichMotor = 3;
	      break;
	    case '5':
          vWhichAxis = SUBRAXIS;
	      vWhichMotor = 4;
	      break;
	    case '6':
          vWhichAxis = SPEEDAXIS;
	      vWhichMotor = 6;
	      break;
	    case '7':
          vWhichAxis = SPEEDAXIS;
	      vWhichMotor = 7;
	      break;

#define PLUSMINUSINCREMENT 5
#define PLUSMINUSSPEEDINCREMENT 0.1
	    case '+': {
          if (vWhichMotor <= NUMOFMOTORS) {
            sh->motors[vWhichMotor].pwm += PLUSMINUSINCREMENT;
    	    if (sh->motors[vWhichMotor].pwm > MAX_PWM) {
              sh->motors[vWhichMotor].pwm = MAX_PWM;
	        }
          } else if (vWhichMotor == 6) {  // az speed
            sh->goalSpeed.az += PLUSMINUSSPEEDINCREMENT;
            vSpAz.setSpeed(sh->goalSpeed.az);
          } else if (vWhichMotor == 7) {  // el speed
            sh->goalSpeed.el += PLUSMINUSSPEEDINCREMENT;
            vSpEl.setSpeed(sh->goalSpeed.el);
          }
          break;
        }
	    case '-': {
          if (vWhichMotor <= NUMOFMOTORS) {
            sh->motors[vWhichMotor].pwm -= PLUSMINUSINCREMENT;
            if (sh->motors[vWhichMotor].pwm < -MAX_PWM) {
              sh->motors[vWhichMotor].pwm = -MAX_PWM;
	        }
          } else if (vWhichMotor == 6) {  // az speed
            sh->goalSpeed.az -= PLUSMINUSSPEEDINCREMENT;
            vSpAz.setSpeed(sh->goalSpeed.az);
          } else if (vWhichMotor == 7) {  // el speed
            sh->goalSpeed.el -= PLUSMINUSSPEEDINCREMENT;
            vSpEl.setSpeed(sh->goalSpeed.el);
          }
          break;
        }
	    case 'n':
          switch (vWhichAxis) {
            case AZAXIS: {
              gotoxy(30, 4+vWhichMotor);
	          sh->motors[vWhichMotor].pwm = getVal(" -> ");
              break;
            }
            case ELAXIS: {
              gotoxy(70, 4-2+vWhichMotor);
	          sh->motors[vWhichMotor].pwm = getVal(" -> ");
              break;
            }
            case SUBRAXIS: {
              break;
            }
            default: {
              break;
            }
          }  // switch axis
		  break;
	    case 'p':
	    case 'P': {
          int i;

          for (i=0; i<NUMOFMOTORS; i++) {
  	        vPWMISA.setEnable(i, TOGGLE);
  	        vPWMISA.setRelay(i, TOGGLE);
          }
	      break;
        }
	    case 's':
	    case 'S':
	      vServoingIsOn = (vServoingIsOn ? 0 : 1);
	      break;
	    case 'l':
	    case 'L':
	      vLoggingEnabled = (vLoggingEnabled ? 0 : 1);
	      break;
	    case 't':
	    case 'T':
#if 0
          sh->motor.az1.pwm = MIN_TEST;
          sh->motor.az2.pwm = MIN_TEST;
#endif
   		  vSTNextChangeTime = ms_since_init + cSTInterval;
	      vStressTestOn = (vStressTestOn ? 0 : 1);
          vSTPhase = +3;  // increasing PWM (after 3..2..1)
          vSTPWM = 0;
	      break;
	    case 'q':
	    case 'Q':
	      goto lExit;
	    default:
	      break;
	  }
	}

	/* Stress test. */
	if (vStressTestOn) {
	  if (ms_since_init > vSTNextChangeTime) {
#if 0
        sh->motor.az1.pwm += cSTIncrement;
        if (sh->motor.az1.pwm > MAX_TEST) {
          sh->motor.az1.pwm = MIN_TEST;
          sh->motor.az2.pwm += cSTIncrement;

          // Stop motors for 2 seconds and then continue with new "jump" sp.
          vPWMISA.setSpeed(0 /*Az1*/, 0);
          vPWMISA.setSpeed(1 /*Az2*/, 0);
          delay(2000);

          if (sh->motor.az2.pwm > MAX_TEST) {
            // To restart:
            // sh->motor.az2.pwm = MIN_TEST;
            sh->motor.az1.pwm = 0;
            sh->motor.az2.pwm = 0;
            vStressTestOn = 0;
          }
        }  // if az1 > max
#endif
#if 1
        if (vSTPhase < -1) {
          vSTPhase++;
        } else if (vSTPhase == -1) {
          if (abs(vSTPWM) < cSTSlowPWM) {
            vSTPWM -= cSTSlowIncr;
          } else if (abs(vSTPWM) < cSTMedPWM) {
            vSTPWM -= cSTMedIncr;
          } else {
            vSTPWM -= cSTHiIncr;
          }
          if (vSTPWM < -(cSTHiPWM)) {
            vSTPWM = -(cSTHiPWM);
            vSTPhase = +4;
          }
        } else if (vSTPhase == 0) {
        } else if (vSTPhase == +1) {
          if (abs(vSTPWM) < cSTSlowPWM) {
            vSTPWM += cSTSlowIncr;
          } else if (abs(vSTPWM) < cSTMedPWM) {
            vSTPWM += cSTMedIncr;
          } else {
            vSTPWM += cSTHiIncr;
          }
          if (vSTPWM > cSTHiPWM) {
            vSTPWM = cSTHiPWM;
            vSTPhase = -4;
          }
        } else if (vSTPhase > +1) {
          vSTPhase--;
        } else {
          // xxx: cannot be...
        }

        sh->motor.az1.pwm = vSTPWM + TIGHTENGEARBOX;
        sh->motor.az2.pwm = vSTPWM - TIGHTENGEARBOX;
        sh->motor.el1.pwm = vSTPWM + TIGHTENGEARBOX;
        sh->motor.el2.pwm = vSTPWM - TIGHTENGEARBOX;
#endif
		vSTNextChangeTime = ms_since_init + cSTInterval;
	  }  // if STInterval elapsed
	}  // if stress test

    // Instant servo.
    if (vServoingIsOn)
    {
      // Max. position error where we can decelerate to zero.
      static const double cMaxPosErr = (MAXVEL * MAXVEL) / MAXACC;
      double posErrAz;
      double posErrEl;

      posErrAz = sh->goalPos.az - sh->incr.az;
      posErrEl = sh->goalPos.el - sh->incr.el;

      // Clip large position errors to the maximum from which
      // we can decelerate to zero.
#if 1
      if (posErrAz > cMaxPosErr) {
        posErrAz = cMaxPosErr;
      }
      if (posErrAz < -cMaxPosErr) {
        posErrAz = -cMaxPosErr;
      }
      if (posErrEl > cMaxPosErr) {
        posErrEl = cMaxPosErr;
      }
      if (posErrEl < -cMaxPosErr) {
        posErrEl = -cMaxPosErr;
      }
#endif

    {
      // Velocity servo.
      double azVelCorr, elVelCorr;

      // Translate position error to speed correction.
#if 1
      // Speed at which we can correct and still have time
      // to accelerate/decelerate.
      azVelCorr = ((fabs(posErrAz) < 0.1) ? AZGAINCOEFF : 1.0)
        * sqrt(fabs(posErrAz) * MAXACC);
      if (posErrAz < 0.0) azVelCorr = -azVelCorr;

      elVelCorr = ((fabs(posErrEl) < 0.1) ? ELGAINCOEFF : 1.0)
        * sqrt(fabs(posErrEl) * MAXACC);
      if (posErrEl < 0.0) elVelCorr = -elVelCorr;
#endif
      // Take goal speed and adjust it with speed correction,
      // but don't change real antenna speed so that it would
      // go beyond acceleration.
#if 1
      // Limit calculated/tracking goal speed.
      {
        double gaz, gel;

        gaz = sh->goalVel.az;
        gel = sh->goalVel.el;
        if (gaz < -MAXVEL) gaz = -MAXVEL;
        if (gaz > MAXVEL) gaz = MAXVEL;
        if (gel < -MAXVEL) gel = -MAXVEL;
        if (gel > MAXVEL) gel = MAXVEL;
        // Sum (already limited) position error --> speed
        // plus newly-limited calculated tracking goal speed.
        azVelCorr += gaz;
        elVelCorr += gel;
      }
#endif

      // Extra boost for "lagging" motors (according to tachometer).
      // xxx: Hmm, should be done separately per motor 1/2.
      // This is currently done in 'setSpeed()' "accelerator" system.

      // Limit resulting speed, cannot exceed 1.2 deg/s.
      if (azVelCorr < -MAXVEL) {
        azVelCorr = -MAXVEL;
      }
      if (azVelCorr > MAXVEL) {
        azVelCorr = MAXVEL;
      }
      if (elVelCorr < -MAXVEL) {
        elVelCorr = -MAXVEL;
      }
      if (elVelCorr > MAXVEL) {
        elVelCorr = MAXVEL;
      }
      // Set new PWMs directly based on velocity.
      // (The 'setSpeed()' routine limits acceleration of antenna.)
      if ((sh->incr.az < MINSLOWAZ) || (sh->incr.az > MAXSLOWAZ)) {
        if (azVelCorr < MINSLOWAZVEL) {
          azVelCorr = MINSLOWAZVEL;
        }
        if (azVelCorr > MAXSLOWAZVEL) {
          azVelCorr = MAXSLOWAZVEL;
        }
      }
      vSpAz.setSpeed(azVelCorr);

      if ((sh->incr.el < MINSLOWEL) || (sh->incr.el > MAXSLOWEL)) {
        if (elVelCorr < MINSLOWELVEL) {
          elVelCorr = MINSLOWELVEL;
        }
        if (elVelCorr > MAXSLOWELVEL) {
          elVelCorr = MAXSLOWELVEL;
        }
      }
      vSpEl.setSpeed(elVelCorr);
    }
    }  // if servoing is on

    {  // Instant antenna location security checks.
       // These should be before mechanical limits,
       // _but_ after final limit switches---otherwise cannot
       // drive manually away from final limits.
       if ( (sh->incr.az < -350.0) || (sh->incr.az > +350.0)
            /*** || (currEl < 10.0) || (currEl > 95.0) ***/
//            || (currEl < 0.2) || (currEl > 106.0)
#if SUNANT
            || (sh->incr.el < -10.0) || (sh->incr.el > 120.0)
#else
// 13-Oct-1999: measured 0.2405 down mechanical limits,
// no activation of down slow nor final limit switches
// up-slow switch activates at +107.2018, deactivates back at +106.9512
// up-final activates at +110.3241, deactivates back at +110.01
            || (sh->incr.el < 0.22) || (sh->incr.el > 110.5)
#endif
       ) {
	     vPWMISA.disableAll();
       }
    }  // antenna location security checks

    // Have we heard anything from downstairs within a reasonable time?
    vLatestDownstairsDelay = ms_since_init - vLatestSerialActivity;
    if (vLatestDownstairsDelay > MAXDELAYFROMDOWNSTAIRS) {
#if SUNANT
#else
      // If not, start decelerating to zero speed.
      vSpAz.setSpeed(0.0);
      vSpEl.setSpeed(0.0);
#endif
    }

    /* Direct PWM security checks: */
	{
      int i;

	  /* In range (-MAX_PWM)..(MAX_PWM): */
      for (i=0; i<5; i++) {
        int p = sh->motors[i].pwm;

	    p = (p < -MAX_PWM) ? -MAX_PWM : p;
	    sh->motors[i].pwm = (p > MAX_PWM) ? MAX_PWM : p;
      }  // for
	}  // of security checks

    // Recalculate (with maximum acceleration) all PWM speeds.
    // (Supply motor tachometer velocities for fine-tuning
    //  according to "true" motor movement/speed.)
    vSpAz.run(sh->motor.az1.qdrVel, sh->motor.az2.qdrVel);
    vSpEl.run(sh->motor.el1.qdrVel, sh->motor.el2.qdrVel);

    // If we haven't heard anything from downstairs and we have
    // already had zero speeds for SHUTDOWNDELAY milliseconds,
    // then we'll shut down all motors; normally update PWM speeds.
#if SUNANT
#else
    if (vLatestDownstairsDelay > (MAXDELAYFROMDOWNSTAIRS + SHUTDOWNDELAY) ) {
      vPWMISA.disableAll();
    } else
#endif
    {
      vPWMISA.setSpeed(0, sh->motors[0].pwm);
      vPWMISA.setSpeed(1, sh->motors[1].pwm);
	  vPWMISA.setSpeed(2, sh->motors[2].pwm);
	  vPWMISA.setSpeed(3, sh->motors[3].pwm);
	  vPWMISA.setSpeed(4, sh->motors[4].pwm);
    }

    // Screen updates.
	gotoxy(1,3);
#if SCREEN_PRINTF_UPDATES
	SCRPRINTF( " p)      Az enabled  %d %d   El  %d %d    %c5) Subrefl  %d   %+05d\n",
      vPWMISA.getEnable(0),
      vPWMISA.getEnable(1),
      vPWMISA.getEnable(2),
      vPWMISA.getEnable(3),
	  (((vWhichAxis == SUBRAXIS) && (vWhichMotor == 4)) ? '>' : ' '),
      vPWMISA.getEnable(4),
	  sh->motor.subr.pwm
    );
    gotoxy(1,4);
	SCRPRINTF("%c1:+,-)  Az1 pulse    %+05d           %c3) El1 pulse    %+05d\n",
	  (((vWhichAxis == AZAXIS) && (vWhichMotor == 0)) ? '>' : ' '),
	  sh->motor.az1.pwm,
	  (((vWhichAxis == ELAXIS) && (vWhichMotor == 2)) ? '>' : ' '),
	  sh->motor.el1.pwm
	);
    gotoxy(1,5);
	SCRPRINTF("%c2:+,-)  Az2 pulse    %+05d           %c4) El2 pulse    %+05d\n",
	  (((vWhichAxis == AZAXIS) && (vWhichMotor == 1)) ? '>' : ' '),
	  sh->motor.az2.pwm,
	  (((vWhichAxis == ELAXIS) && (vWhichMotor == 3)) ? '>' : ' '),
	  sh->motor.el2.pwm
	);
    gotoxy(1,6);
	SCRPRINTF( " t)      Stress test  %d  (%3d)         s) Servoing     %d     l) Log %d \n",
      vStressTestOn,
      vSTPhase,
      vServoingIsOn,
      vLoggingEnabled
    );
#endif

  { // Logging all measured values to a file.
    static double loggedAz, loggedEl;

    if ( (fabs(sh->curr.az - loggedAz) > MINLOGCHANGE)
          || (fabs(sh->curr.el - loggedEl) > MINLOGCHANGE)
          || ((currAzElTime - prevAzElTime) > MINLOGUPDTIME) ) {
      if (vLoggingEnabled) {
      fprintf(vLogFile,
        "%010lu\t"  // time, 1
        "%05d\t%05u\t%10.6f\t%5.2f\t"  //az1, 2-5
        "%05d\t%05u\t%10.6f\t%5.2f\t"  //az2, 6-9
        "%11.6f\t%10.6f\t%10.6f\t%11.6f\t%10.6f\t%10.6f\t"  //az, 10-15
        "%05d\t%05u\t%10.6f\t%5.2f\t"  // el1, 16-19
        "%05d\t%05u\t%10.6f\t%5.2f\t"  // el2, 20-23
        "%11.6f\t%10.6f\t%10.6f\t%11.6f\t%10.6f\t%10.6f\t"  //el, 24-29
        "%9.4f\t%9.4f\t%8.4f\t\n",  // tilt, 30-32
        currAzElTime,
        vPWMISA.getEnable(0) ? sh->motor.az1.pwm : 0,
        sh->motor.az1.qdr,  // tacho
        sh->motor.az1.qdrVel,
        sh->motor.az1.icurr,
        vPWMISA.getEnable(1) ? sh->motor.az2.pwm : 0,
        sh->motor.az2.qdr,
        sh->motor.az2.qdrVel,
        sh->motor.az2.icurr,
        sh->curr.az,
        sh->currSp.az,
        sh->currAcc.az,
        sh->goalPos.az,
        sh->goalVel.az,
        sh->goalAcc.az,
        vPWMISA.getEnable(2) ? sh->motor.el1.pwm : 0,
        sh->motor.el1.qdr,
        sh->motor.el1.qdrVel,
        sh->motor.el1.icurr,
        vPWMISA.getEnable(3) ? sh->motor.el2.pwm : 0,
        sh->motor.el2.qdr,
        sh->motor.el2.qdrVel,
        sh->motor.el2.icurr,
        sh->curr.el,
        sh->currSp.el,
        sh->currAcc.el,
        sh->goalPos.el,
        sh->goalVel.el,
        sh->goalAcc.el,
        sh->tilt.az,
        sh->tilt.el,
        sh->temp.tilt
      );
      }  // if logging enabled
      loggedAz = sh->curr.az;
      loggedEl = sh->curr.el;
      prevAzElTime = currAzElTime;
    }  // if needs to be logged
  }

#if SCREEN_PRINTF_UPDATES
    gotoxy(1,7);
    SCRPRINTF(
      "  AbsAz %7.2f  %04X   AbsEl %7.2f  %04X \n",
      sh->abs.az = currAbsAz.get(),
      currAbsAz.getStableBin(),  // or Gray()
      sh->abs.el = currAbsEl.get(),
      currAbsEl.getStableBin()
    );
#endif

#if SCREEN_BARE_MINIMUM_PRINTFS
    gotoxy(1,8);
#if SUNANT
      SCRPRINTF(
        "     Az %11.6f        El %11.6f      SerErr %3lu\n",
        sh->incr.az,
        sh->incr.el,
        vSerialReadErrors
      );
#else
      SCRPRINTF(
        "  %c%c Az %11.6f     %c%c El %11.6f  %s  %s  SerErr %3lu\n",
        vIK121.isStopped(0) ? 'S' : ' ',
        vIK121.isSearchingForRefMark(0) ? 'R' : ' ',
        sh->incr.az,
        vIK121.isStopped(1) ? 'S' : ' ',
        vIK121.isSearchingForRefMark(1) ? 'R' : ' ',
        sh->incr.el,
        vIK121.amplStatus(0),
        vIK121.amplStatus(1),
        vSerialReadErrors
      );
#endif
#endif

#if SCREEN_PRINTF_UPDATES
    gotoxy(1,9);
      SCRPRINTF(
        " GoalAz %11.6f    GoalEl %11.6f    %11.6f  %11.6f\n",
        sh->goalPos.az,
        sh->goalPos.el,
        sh->goalVel.az,
        sh->goalVel.el
      );
    gotoxy(1,10);
      SCRPRINTF(
        "ErrorAz %11.6f   ErrorEl %11.6f  Downstairs %s (%5lu)\n\n",
        sh->goalPos.az - sh->incr.az,
        sh->goalPos.el - sh->incr.el,
        (vLatestDownstairsDelay <= MAXDELAYFROMDOWNSTAIRS)
          ? "ok" : (
          (vLatestDownstairsDelay <= (MAXDELAYFROMDOWNSTAIRS + SHUTDOWNDELAY))
            ? "??" : "NA"
          ),
        vLatestDownstairsDelay
      );
    gotoxy(1,12);
      SCRPRINTF(
        "  %cSpAz %11.6f     %cSpEl %11.6f  (deg/s)\n",
   	    (((vWhichAxis == SPEEDAXIS) && (vWhichMotor == 6)) ? '>' : ' '),
#if 0
        sh->goalSpeed.az,
#else
        sh->currSp.az,
#endif
   	    (((vWhichAxis == SPEEDAXIS) && (vWhichMotor == 7)) ? '>' : ' '),
#if 0
        sh->goalSpeed.el
#else
        sh->currSp.el
#endif
      );
    gotoxy(1,13);
      SCRPRINTF(
        "  AccAz %11.6f     AccEl %11.6f  (deg/s/s)    \n",
        sh->currAcc.az,
        sh->currAcc.el
      );
    gotoxy(1,14);
      SCRPRINTF(
	"Tilt-Az %11.6f   Tilt-El %11.6f  Tilt-Temp %5.1f\n",
        sh->tilt.az,
        sh->tilt.el,
        sh->temp.tilt
      );
    gotoxy(1,15);
      SCRPRINTF(
        "      Az1       Az2         El1       El2\n"
      );
    gotoxy(1,16);
      SCRPRINTF(
        " %6.2f A  %6.2f A    %6.2f A  %6.2f A  T-rack %4.1f/%4.1f/%4.1f C\n",
        sh->motor.az1.icurr,
        sh->motor.az2.icurr,
        sh->motor.el1.icurr,
        sh->motor.el2.icurr,
        sh->temp.rackmiddle,
        sh->temp.racklow,
        sh->temp.rackpc
      );
    gotoxy(1,17);
      SCRPRINTF(
        "  %5.1f C   %5.1f C     %5.1f C   %5.1f C  Vsuppl %4.2f/%4.1f/%4.1f V\n",
        sh->motor.az1.temp,
        sh->motor.az2.temp,
        sh->motor.el1.temp,
        sh->motor.el2.temp,
        sh->volt.v5,
        sh->volt.v12,
        sh->volt.v24
      );
    gotoxy(1,18);
      SCRPRINTF(
        "   %4.1f V    %4.1f V      %4.1f V    %4.1f V  Vsupp %5.1f/%4.1f/%7.4f V\n",
        sh->motor.az1.vsupply,
        sh->motor.az2.vsupply,
        sh->motor.el1.vsupply,
        sh->motor.el2.vsupply,
        sh->volt.vn15,
        sh->volt.v15,
        sh->volt.vgnd
      );
    gotoxy(1,19);
      SCRPRINTF(
        "    %5u     %5u       %5u     %5u  Tachos  IBrakeAz %4.1f  El %4.1f A\n",
        sh->motor.az1.qdr,
        sh->motor.az2.qdr,
        sh->motor.el1.qdr,
        sh->motor.el2.qdr,
        sh->ibrake.az,
        sh->ibrake.el
      );
    gotoxy(1,24);
      SCRPRINTF(
        "%10.6f  %10.6f    %10.6f  %10.6f  ",
        sh->motor.az1.qdrVel,
        sh->motor.az2.qdrVel,
        sh->motor.el1.qdrVel,
        sh->motor.el2.qdrVel
      );
      {
        char vMs0[10];
        char vMs1[10];
        char vMs2[10];
        char vMs3[10];

        vPWMISA.formatStatus(vPWMISA.getStatus(/*motor=>*/ 0), vMs0);
        vPWMISA.formatStatus(vPWMISA.getStatus(/*motor=>*/ 1), vMs1);
        vPWMISA.formatStatus(vPWMISA.getStatus(/*motor=>*/ 2), vMs2);
        vPWMISA.formatStatus(vPWMISA.getStatus(/*motor=>*/ 3), vMs3);
    gotoxy(1,20);
        SCRPRINTF(
          "  %07s   %07s     %07s   %07s  APH STO FLT CHO REL ENA DIR,  0)rst\n\n",
          vMs0,
          vMs1,
          vMs2,
          vMs3
        );
      }
#endif

#if SCREEN_BARE_MINIMUM_PRINTFS
    gotoxy(1,21);
      SCRPRINTF(
        "Interlock   %06lX, %s",
        sh->interlock,
        explainInterlock(sh->interlock)
      );
      clreol(); SCRPRINTF("\n");
#endif

#if CATCH_OVER_10AMPS
      if ( (fabs(vAz1ICurr) >= 10.0)
            || (fabs(vAz2ICurr) >= 10.0)
            || (fabs(vEl1ICurr) >= 10.0)
            || (fabs(vEl2ICurr) >= 10.0) ) {
        sound(600); delay(500); nosound();
        while (!kbhit()) ;
        (void)getch();
      }
#endif

    // Generate running 'ltime' (seconds) and 'ms_since_init' (msecs).
    prevUsecs = microseconds;
    microseconds = vPWMISA.getUsecs();
    if (microseconds > 999999) {
      // Board is missing.
      microseconds = 0;
      // xxx: Consider using microseconds from downstairs,
      // sh->xntpdTime.tv_usec
    }
    if (microseconds < prevUsecs) {  // overflow, i.e. next second
      ltime++;
      s_since_init++;
      prevLoopUsecs = 1000000L - prevUsecs + microseconds;
    } else {
      prevLoopUsecs = microseconds - prevUsecs;
    }
    ms_since_init = s_since_init*1000L + microseconds/1000L;

#if SCREEN_TIME_DISPLAY
    gotoxy(1,22);
 	SCRPRINTF("UTC is %s",
      asctime(gmtime(&ltime))
    );
    gotoxy(36,22);
 	SCRPRINTF("%06lu %010lu  %1d  PrevLoop = %6lu us\n",
      microseconds,
      ms_since_init,
      vPWMISA.getSecs(),
      prevLoopUsecs
    );

    gotoxy(1,23);
 	SCRPRINTF("xntpd  %s",
      asctime(gmtime(&(sh->xntpdTime.tv_sec)))
    );
    gotoxy(36,23);
    {
      signed long int t;
      long int calcSec;

      // xxx: Must check that tv_usec is incrementing,
      // just like PWMISA usecs should be incrementing.
      // If it is not we must not believe tv_secs to adjust
      // our clock 'ltime' but keep the old seconds value.

      if (sh->xntpdTime.tv_usec < microseconds) {
        // Normal case, "old" xntpd time is earlier.
        t = microseconds - sh->xntpdTime.tv_usec;
        calcSec = sh->xntpdTime.tv_sec;
      } else {
        // Microseconds has wrapped back to near zero.
        t = 1000000L + microseconds - sh->xntpdTime.tv_usec;
        calcSec = sh->xntpdTime.tv_sec + 1;
      }
      // Have we slipped a second?  (I.e. 1pps arrived to PWMISA
      // a different number of times times than to xntpd downstairs.)
      if (t < 100000) {  // Time difference less than 100ms.
        // Measurement is reliable.
        if ((ltime < calcSec) && (ltime > 0)) {
          v1ppsExtraPulses++;
        } else if (ltime > calcSec) {
          v1ppsMissingPulses++;
        } else {  // calc == xntpd
          // Ok.
        }
        ltime = calcSec;
      } else {
        // xntpd time is too far off when compared to PWMISA.
        // Retain 'ltime' and increment 'lost xntpd time' error counter.
        // xxx: vxxx++;
      }

   	  SCRPRINTF("%06lu %2s %7ld  Extra = %ld  Miss = %ld\n",
        sh->xntpdTime.tv_usec,
        (t < 100000) ? "Ok" : "!!",  // usec diff should be max 100msec
        t,
        v1ppsExtraPulses,
        v1ppsMissingPulses
      );
    }
#endif

	vPWMISA.refreshWatchdog();  /* keep motors on */

  }  // for ever
  /* ---------- main for-ever loop */

lExit:
  _setcursortype(_NORMALCURSOR);
  printf("\n");

  CloseSerial(COM2);

  // xxx: Should in principle 'cleanup()' lots of '*Ch' sensor channels.

#if SUNANT
  vAzIncrTacho.cleanup();
  vElIncrTacho.cleanup();
#else
  vIK121.cleanup();
#endif
  vACL8113.cleanup();
  vInterlockChain.cleanup();
  vTachos.cleanup();
  currAbsEl.cleanup();
  currAbsAz.cleanup();

  realtimeRestore();
  vPWMISA.cleanup();

#if 1
  printf("dl = %u, ul = %u\n", sizeof(tDownlink), sizeof(tUplink));
#endif
  return (0);
}  // main
